(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global.attadux = {})));
}(this, (function (exports) { 'use strict';

function _isPlaceholder(a) {
       return a != null && typeof a === 'object' && a['@@functional/placeholder'] === true;
}
var _isPlaceholder_1 = _isPlaceholder;

/**
 * Optimized internal one-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */


function _curry1(fn) {
  return function f1(a) {
    if (arguments.length === 0 || _isPlaceholder_1(a)) {
      return f1;
    } else {
      return fn.apply(this, arguments);
    }
  };
}
var _curry1_1 = _curry1;

/**
 * Optimized internal two-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */


function _curry2(fn) {
  return function f2(a, b) {
    switch (arguments.length) {
      case 0:
        return f2;
      case 1:
        return _isPlaceholder_1(a) ? f2 : _curry1_1(function (_b) {
          return fn(a, _b);
        });
      default:
        return _isPlaceholder_1(a) && _isPlaceholder_1(b) ? f2 : _isPlaceholder_1(a) ? _curry1_1(function (_a) {
          return fn(_a, b);
        }) : _isPlaceholder_1(b) ? _curry1_1(function (_b) {
          return fn(a, _b);
        }) : fn(a, b);
    }
  };
}
var _curry2_1 = _curry2;

/**
 * Applies function `fn` to the argument list `args`. This is useful for
 * creating a fixed-arity function from a variadic function. `fn` should be a
 * bound function if context is significant.
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category Function
 * @sig (*... -> a) -> [*] -> a
 * @param {Function} fn The function which will be called with `args`
 * @param {Array} args The arguments to call `fn` with
 * @return {*} result The result, equivalent to `fn(...args)`
 * @see R.call, R.unapply
 * @example
 *
 *      var nums = [1, 2, 3, -99, 42, 6, 7];
 *      R.apply(Math.max, nums); //=> 42
 * @symb R.apply(f, [a, b, c]) = f(a, b, c)
 */


var apply = /*#__PURE__*/_curry2_1(function apply(fn, args) {
  return fn.apply(this, args);
});
var apply_1 = apply;

function _arity(n, fn) {
  /* eslint-disable no-unused-vars */
  switch (n) {
    case 0:
      return function () {
        return fn.apply(this, arguments);
      };
    case 1:
      return function (a0) {
        return fn.apply(this, arguments);
      };
    case 2:
      return function (a0, a1) {
        return fn.apply(this, arguments);
      };
    case 3:
      return function (a0, a1, a2) {
        return fn.apply(this, arguments);
      };
    case 4:
      return function (a0, a1, a2, a3) {
        return fn.apply(this, arguments);
      };
    case 5:
      return function (a0, a1, a2, a3, a4) {
        return fn.apply(this, arguments);
      };
    case 6:
      return function (a0, a1, a2, a3, a4, a5) {
        return fn.apply(this, arguments);
      };
    case 7:
      return function (a0, a1, a2, a3, a4, a5, a6) {
        return fn.apply(this, arguments);
      };
    case 8:
      return function (a0, a1, a2, a3, a4, a5, a6, a7) {
        return fn.apply(this, arguments);
      };
    case 9:
      return function (a0, a1, a2, a3, a4, a5, a6, a7, a8) {
        return fn.apply(this, arguments);
      };
    case 10:
      return function (a0, a1, a2, a3, a4, a5, a6, a7, a8, a9) {
        return fn.apply(this, arguments);
      };
    default:
      throw new Error('First argument to _arity must be a non-negative integer no greater than ten');
  }
}
var _arity_1 = _arity;

/**
 * Internal curryN function.
 *
 * @private
 * @category Function
 * @param {Number} length The arity of the curried function.
 * @param {Array} received An array of arguments received thus far.
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */


function _curryN(length, received, fn) {
  return function () {
    var combined = [];
    var argsIdx = 0;
    var left = length;
    var combinedIdx = 0;
    while (combinedIdx < received.length || argsIdx < arguments.length) {
      var result;
      if (combinedIdx < received.length && (!_isPlaceholder_1(received[combinedIdx]) || argsIdx >= arguments.length)) {
        result = received[combinedIdx];
      } else {
        result = arguments[argsIdx];
        argsIdx += 1;
      }
      combined[combinedIdx] = result;
      if (!_isPlaceholder_1(result)) {
        left -= 1;
      }
      combinedIdx += 1;
    }
    return left <= 0 ? fn.apply(this, combined) : _arity_1(left, _curryN(length, combined, fn));
  };
}
var _curryN_1 = _curryN;

/**
 * Returns a curried equivalent of the provided function, with the specified
 * arity. The curried function has two unusual capabilities. First, its
 * arguments needn't be provided one at a time. If `g` is `R.curryN(3, f)`, the
 * following are equivalent:
 *
 *   - `g(1)(2)(3)`
 *   - `g(1)(2, 3)`
 *   - `g(1, 2)(3)`
 *   - `g(1, 2, 3)`
 *
 * Secondly, the special placeholder value [`R.__`](#__) may be used to specify
 * "gaps", allowing partial application of any combination of arguments,
 * regardless of their positions. If `g` is as above and `_` is [`R.__`](#__),
 * the following are equivalent:
 *
 *   - `g(1, 2, 3)`
 *   - `g(_, 2, 3)(1)`
 *   - `g(_, _, 3)(1)(2)`
 *   - `g(_, _, 3)(1, 2)`
 *   - `g(_, 2)(1)(3)`
 *   - `g(_, 2)(1, 3)`
 *   - `g(_, 2)(_, 3)(1)`
 *
 * @func
 * @memberOf R
 * @since v0.5.0
 * @category Function
 * @sig Number -> (* -> a) -> (* -> a)
 * @param {Number} length The arity for the returned function.
 * @param {Function} fn The function to curry.
 * @return {Function} A new, curried function.
 * @see R.curry
 * @example
 *
 *      var sumArgs = (...args) => R.sum(args);
 *
 *      var curriedAddFourNumbers = R.curryN(4, sumArgs);
 *      var f = curriedAddFourNumbers(1, 2);
 *      var g = f(3);
 *      g(4); //=> 10
 */


var curryN = /*#__PURE__*/_curry2_1(function curryN(length, fn) {
  if (length === 1) {
    return _curry1_1(fn);
  }
  return _arity_1(length, _curryN_1(length, [], fn));
});
var curryN_1 = curryN;

/**
 * Tests whether or not an object is an array.
 *
 * @private
 * @param {*} val The object to test.
 * @return {Boolean} `true` if `val` is an array, `false` otherwise.
 * @example
 *
 *      _isArray([]); //=> true
 *      _isArray(null); //=> false
 *      _isArray({}); //=> false
 */
var _isArray = Array.isArray || function _isArray(val) {
  return val != null && val.length >= 0 && Object.prototype.toString.call(val) === '[object Array]';
};

function _isTransformer(obj) {
  return typeof obj['@@transducer/step'] === 'function';
}
var _isTransformer_1 = _isTransformer;

/**
 * Returns a function that dispatches with different strategies based on the
 * object in list position (last argument). If it is an array, executes [fn].
 * Otherwise, if it has a function with one of the given method names, it will
 * execute that function (functor case). Otherwise, if it is a transformer,
 * uses transducer [xf] to return a new transformer (transducer case).
 * Otherwise, it will default to executing [fn].
 *
 * @private
 * @param {Array} methodNames properties to check for a custom implementation
 * @param {Function} xf transducer to initialize if object is transformer
 * @param {Function} fn default ramda implementation
 * @return {Function} A function that dispatches on object in list position
 */


function _dispatchable(methodNames, xf, fn) {
  return function () {
    if (arguments.length === 0) {
      return fn();
    }
    var args = Array.prototype.slice.call(arguments, 0);
    var obj = args.pop();
    if (!_isArray(obj)) {
      var idx = 0;
      while (idx < methodNames.length) {
        if (typeof obj[methodNames[idx]] === 'function') {
          return obj[methodNames[idx]].apply(obj, args);
        }
        idx += 1;
      }
      if (_isTransformer_1(obj)) {
        var transducer = xf.apply(null, args);
        return transducer(obj);
      }
    }
    return fn.apply(this, arguments);
  };
}
var _dispatchable_1 = _dispatchable;

function _map(fn, functor) {
  var idx = 0;
  var len = functor.length;
  var result = Array(len);
  while (idx < len) {
    result[idx] = fn(functor[idx]);
    idx += 1;
  }
  return result;
}
var _map_1 = _map;

function _isString(x) {
  return Object.prototype.toString.call(x) === '[object String]';
}
var _isString_1 = _isString;

/**
 * Tests whether or not an object is similar to an array.
 *
 * @private
 * @category Type
 * @category List
 * @sig * -> Boolean
 * @param {*} x The object to test.
 * @return {Boolean} `true` if `x` has a numeric length property and extreme indices defined; `false` otherwise.
 * @example
 *
 *      _isArrayLike([]); //=> true
 *      _isArrayLike(true); //=> false
 *      _isArrayLike({}); //=> false
 *      _isArrayLike({length: 10}); //=> false
 *      _isArrayLike({0: 'zero', 9: 'nine', length: 10}); //=> true
 */


var _isArrayLike = /*#__PURE__*/_curry1_1(function isArrayLike(x) {
  if (_isArray(x)) {
    return true;
  }
  if (!x) {
    return false;
  }
  if (typeof x !== 'object') {
    return false;
  }
  if (_isString_1(x)) {
    return false;
  }
  if (x.nodeType === 1) {
    return !!x.length;
  }
  if (x.length === 0) {
    return true;
  }
  if (x.length > 0) {
    return x.hasOwnProperty(0) && x.hasOwnProperty(x.length - 1);
  }
  return false;
});
var _isArrayLike_1 = _isArrayLike;

var XWrap = /*#__PURE__*/function () {
  function XWrap(fn) {
    this.f = fn;
  }
  XWrap.prototype['@@transducer/init'] = function () {
    throw new Error('init not implemented on XWrap');
  };
  XWrap.prototype['@@transducer/result'] = function (acc) {
    return acc;
  };
  XWrap.prototype['@@transducer/step'] = function (acc, x) {
    return this.f(acc, x);
  };

  return XWrap;
}();

function _xwrap(fn) {
  return new XWrap(fn);
}
var _xwrap_1 = _xwrap;

/**
 * Creates a function that is bound to a context.
 * Note: `R.bind` does not provide the additional argument-binding capabilities of
 * [Function.prototype.bind](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Function/bind).
 *
 * @func
 * @memberOf R
 * @since v0.6.0
 * @category Function
 * @category Object
 * @sig (* -> *) -> {*} -> (* -> *)
 * @param {Function} fn The function to bind to context
 * @param {Object} thisObj The context to bind `fn` to
 * @return {Function} A function that will execute in the context of `thisObj`.
 * @see R.partial
 * @example
 *
 *      var log = R.bind(console.log, console);
 *      R.pipe(R.assoc('a', 2), R.tap(log), R.assoc('a', 3))({a: 1}); //=> {a: 3}
 *      // logs {a: 2}
 * @symb R.bind(f, o)(a, b) = f.call(o, a, b)
 */


var bind = /*#__PURE__*/_curry2_1(function bind(fn, thisObj) {
  return _arity_1(fn.length, function () {
    return fn.apply(thisObj, arguments);
  });
});
var bind_1 = bind;

function _arrayReduce(xf, acc, list) {
  var idx = 0;
  var len = list.length;
  while (idx < len) {
    acc = xf['@@transducer/step'](acc, list[idx]);
    if (acc && acc['@@transducer/reduced']) {
      acc = acc['@@transducer/value'];
      break;
    }
    idx += 1;
  }
  return xf['@@transducer/result'](acc);
}

function _iterableReduce(xf, acc, iter) {
  var step = iter.next();
  while (!step.done) {
    acc = xf['@@transducer/step'](acc, step.value);
    if (acc && acc['@@transducer/reduced']) {
      acc = acc['@@transducer/value'];
      break;
    }
    step = iter.next();
  }
  return xf['@@transducer/result'](acc);
}

function _methodReduce(xf, acc, obj, methodName) {
  return xf['@@transducer/result'](obj[methodName](bind_1(xf['@@transducer/step'], xf), acc));
}

var symIterator = typeof Symbol !== 'undefined' ? Symbol.iterator : '@@iterator';

function _reduce(fn, acc, list) {
  if (typeof fn === 'function') {
    fn = _xwrap_1(fn);
  }
  if (_isArrayLike_1(list)) {
    return _arrayReduce(fn, acc, list);
  }
  if (typeof list['fantasy-land/reduce'] === 'function') {
    return _methodReduce(fn, acc, list, 'fantasy-land/reduce');
  }
  if (list[symIterator] != null) {
    return _iterableReduce(fn, acc, list[symIterator]());
  }
  if (typeof list.next === 'function') {
    return _iterableReduce(fn, acc, list);
  }
  if (typeof list.reduce === 'function') {
    return _methodReduce(fn, acc, list, 'reduce');
  }

  throw new TypeError('reduce: list must be array or iterable');
}
var _reduce_1 = _reduce;

var _xfBase = {
  init: function () {
    return this.xf['@@transducer/init']();
  },
  result: function (result) {
    return this.xf['@@transducer/result'](result);
  }
};

var XMap = /*#__PURE__*/function () {

  function XMap(f, xf) {
    this.xf = xf;
    this.f = f;
  }
  XMap.prototype['@@transducer/init'] = _xfBase.init;
  XMap.prototype['@@transducer/result'] = _xfBase.result;
  XMap.prototype['@@transducer/step'] = function (result, input) {
    return this.xf['@@transducer/step'](result, this.f(input));
  };

  return XMap;
}();

var _xmap = /*#__PURE__*/_curry2_1(function _xmap(f, xf) {
  return new XMap(f, xf);
});
var _xmap_1 = _xmap;

function _has(prop, obj) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
}
var _has_1 = _has;

var toString$1 = Object.prototype.toString;
var _isArguments = function () {
  return toString$1.call(arguments) === '[object Arguments]' ? function _isArguments(x) {
    return toString$1.call(x) === '[object Arguments]';
  } : function _isArguments(x) {
    return _has_1('callee', x);
  };
};

var _isArguments_1 = _isArguments;

// cover IE < 9 keys issues


var hasEnumBug = ! /*#__PURE__*/{ toString: null }.propertyIsEnumerable('toString');
var nonEnumerableProps = ['constructor', 'valueOf', 'isPrototypeOf', 'toString', 'propertyIsEnumerable', 'hasOwnProperty', 'toLocaleString'];
// Safari bug
var hasArgsEnumBug = /*#__PURE__*/function () {
  'use strict';

  return arguments.propertyIsEnumerable('length');
}();

var contains = function contains(list, item) {
  var idx = 0;
  while (idx < list.length) {
    if (list[idx] === item) {
      return true;
    }
    idx += 1;
  }
  return false;
};

/**
 * Returns a list containing the names of all the enumerable own properties of
 * the supplied object.
 * Note that the order of the output array is not guaranteed to be consistent
 * across different JS platforms.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig {k: v} -> [k]
 * @param {Object} obj The object to extract properties from
 * @return {Array} An array of the object's own properties.
 * @see R.keysIn, R.values
 * @example
 *
 *      R.keys({a: 1, b: 2, c: 3}); //=> ['a', 'b', 'c']
 */
var _keys = typeof Object.keys === 'function' && !hasArgsEnumBug ? function keys(obj) {
  return Object(obj) !== obj ? [] : Object.keys(obj);
} : function keys(obj) {
  if (Object(obj) !== obj) {
    return [];
  }
  var prop, nIdx;
  var ks = [];
  var checkArgsLength = hasArgsEnumBug && _isArguments_1(obj);
  for (prop in obj) {
    if (_has_1(prop, obj) && (!checkArgsLength || prop !== 'length')) {
      ks[ks.length] = prop;
    }
  }
  if (hasEnumBug) {
    nIdx = nonEnumerableProps.length - 1;
    while (nIdx >= 0) {
      prop = nonEnumerableProps[nIdx];
      if (_has_1(prop, obj) && !contains(ks, prop)) {
        ks[ks.length] = prop;
      }
      nIdx -= 1;
    }
  }
  return ks;
};
var keys = /*#__PURE__*/_curry1_1(_keys);
var keys_1 = keys;

/**
 * Takes a function and
 * a [functor](https://github.com/fantasyland/fantasy-land#functor),
 * applies the function to each of the functor's values, and returns
 * a functor of the same shape.
 *
 * Ramda provides suitable `map` implementations for `Array` and `Object`,
 * so this function may be applied to `[1, 2, 3]` or `{x: 1, y: 2, z: 3}`.
 *
 * Dispatches to the `map` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * Also treats functions as functors and will compose them together.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Functor f => (a -> b) -> f a -> f b
 * @param {Function} fn The function to be called on every element of the input `list`.
 * @param {Array} list The list to be iterated over.
 * @return {Array} The new list.
 * @see R.transduce, R.addIndex
 * @example
 *
 *      var double = x => x * 2;
 *
 *      R.map(double, [1, 2, 3]); //=> [2, 4, 6]
 *
 *      R.map(double, {x: 1, y: 2, z: 3}); //=> {x: 2, y: 4, z: 6}
 * @symb R.map(f, [a, b]) = [f(a), f(b)]
 * @symb R.map(f, { x: a, y: b }) = { x: f(a), y: f(b) }
 * @symb R.map(f, functor_o) = functor_o.map(f)
 */


var map = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['fantasy-land/map', 'map'], _xmap_1, function map(fn, functor) {
  switch (Object.prototype.toString.call(functor)) {
    case '[object Function]':
      return curryN_1(functor.length, function () {
        return fn.call(this, functor.apply(this, arguments));
      });
    case '[object Object]':
      return _reduce_1(function (acc, key) {
        acc[key] = fn(functor[key]);
        return acc;
      }, {}, keys_1(functor));
    default:
      return _map_1(fn, functor);
  }
}));
var map_1 = map;

/**
 * Returns the larger of its two arguments.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Relation
 * @sig Ord a => a -> a -> a
 * @param {*} a
 * @param {*} b
 * @return {*}
 * @see R.maxBy, R.min
 * @example
 *
 *      R.max(789, 123); //=> 789
 *      R.max('a', 'b'); //=> 'b'
 */


var max = /*#__PURE__*/_curry2_1(function max(a, b) {
  return b > a ? b : a;
});
var max_1 = max;

/**
 * Retrieve the value at a given path.
 *
 * @func
 * @memberOf R
 * @since v0.2.0
 * @category Object
 * @typedefn Idx = String | Int
 * @sig [Idx] -> {a} -> a | Undefined
 * @param {Array} path The path to use.
 * @param {Object} obj The object to retrieve the nested property from.
 * @return {*} The data at `path`.
 * @see R.prop
 * @example
 *
 *      R.path(['a', 'b'], {a: {b: 2}}); //=> 2
 *      R.path(['a', 'b'], {c: {b: 2}}); //=> undefined
 */


var path = /*#__PURE__*/_curry2_1(function path(paths, obj) {
  var val = obj;
  var idx = 0;
  while (idx < paths.length) {
    if (val == null) {
      return;
    }
    val = val[paths[idx]];
    idx += 1;
  }
  return val;
});
var path_1 = path;

/**
 * Returns a function that when supplied an object returns the indicated
 * property of that object, if it exists.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig s -> {s: a} -> a | Undefined
 * @param {String} p The property name
 * @param {Object} obj The object to query
 * @return {*} The value at `obj.p`.
 * @see R.path
 * @example
 *
 *      R.prop('x', {x: 100}); //=> 100
 *      R.prop('x', {}); //=> undefined
 */

var prop = /*#__PURE__*/_curry2_1(function prop(p, obj) {
  return path_1([p], obj);
});
var prop_1 = prop;

/**
 * Returns a new list by plucking the same named property off all objects in
 * the list supplied.
 *
 * `pluck` will work on
 * any [functor](https://github.com/fantasyland/fantasy-land#functor) in
 * addition to arrays, as it is equivalent to `R.map(R.prop(k), f)`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Functor f => k -> f {k: v} -> f v
 * @param {Number|String} key The key name to pluck off of each object.
 * @param {Array} f The array or functor to consider.
 * @return {Array} The list of values for the given key.
 * @see R.props
 * @example
 *
 *      R.pluck('a')([{a: 1}, {a: 2}]); //=> [1, 2]
 *      R.pluck(0)([[1, 2], [3, 4]]);   //=> [1, 3]
 *      R.pluck('val', {a: {val: 3}, b: {val: 5}}); //=> {a: 3, b: 5}
 * @symb R.pluck('x', [{x: 1, y: 2}, {x: 3, y: 4}, {x: 5, y: 6}]) = [1, 3, 5]
 * @symb R.pluck(0, [[1, 2], [3, 4], [5, 6]]) = [1, 3, 5]
 */


var pluck = /*#__PURE__*/_curry2_1(function pluck(p, list) {
  return map_1(prop_1(p), list);
});
var pluck_1 = pluck;

/**
 * Optimized internal three-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */


function _curry3(fn) {
  return function f3(a, b, c) {
    switch (arguments.length) {
      case 0:
        return f3;
      case 1:
        return _isPlaceholder_1(a) ? f3 : _curry2_1(function (_b, _c) {
          return fn(a, _b, _c);
        });
      case 2:
        return _isPlaceholder_1(a) && _isPlaceholder_1(b) ? f3 : _isPlaceholder_1(a) ? _curry2_1(function (_a, _c) {
          return fn(_a, b, _c);
        }) : _isPlaceholder_1(b) ? _curry2_1(function (_b, _c) {
          return fn(a, _b, _c);
        }) : _curry1_1(function (_c) {
          return fn(a, b, _c);
        });
      default:
        return _isPlaceholder_1(a) && _isPlaceholder_1(b) && _isPlaceholder_1(c) ? f3 : _isPlaceholder_1(a) && _isPlaceholder_1(b) ? _curry2_1(function (_a, _b) {
          return fn(_a, _b, c);
        }) : _isPlaceholder_1(a) && _isPlaceholder_1(c) ? _curry2_1(function (_a, _c) {
          return fn(_a, b, _c);
        }) : _isPlaceholder_1(b) && _isPlaceholder_1(c) ? _curry2_1(function (_b, _c) {
          return fn(a, _b, _c);
        }) : _isPlaceholder_1(a) ? _curry1_1(function (_a) {
          return fn(_a, b, c);
        }) : _isPlaceholder_1(b) ? _curry1_1(function (_b) {
          return fn(a, _b, c);
        }) : _isPlaceholder_1(c) ? _curry1_1(function (_c) {
          return fn(a, b, _c);
        }) : fn(a, b, c);
    }
  };
}
var _curry3_1 = _curry3;

/**
 * Returns a single item by iterating through the list, successively calling
 * the iterator function and passing it an accumulator value and the current
 * value from the array, and then passing the result to the next call.
 *
 * The iterator function receives two values: *(acc, value)*. It may use
 * [`R.reduced`](#reduced) to shortcut the iteration.
 *
 * The arguments' order of [`reduceRight`](#reduceRight)'s iterator function
 * is *(value, acc)*.
 *
 * Note: `R.reduce` does not skip deleted or unassigned indices (sparse
 * arrays), unlike the native `Array.prototype.reduce` method. For more details
 * on this behavior, see:
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/reduce#Description
 *
 * Dispatches to the `reduce` method of the third argument, if present. When
 * doing so, it is up to the user to handle the [`R.reduced`](#reduced)
 * shortcuting, as this is not implemented by `reduce`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig ((a, b) -> a) -> a -> [b] -> a
 * @param {Function} fn The iterator function. Receives two values, the accumulator and the
 *        current element from the array.
 * @param {*} acc The accumulator value.
 * @param {Array} list The list to iterate over.
 * @return {*} The final, accumulated value.
 * @see R.reduced, R.addIndex, R.reduceRight
 * @example
 *
 *      R.reduce(R.subtract, 0, [1, 2, 3, 4]) // => ((((0 - 1) - 2) - 3) - 4) = -10
 *      //          -               -10
 *      //         / \              / \
 *      //        -   4           -6   4
 *      //       / \              / \
 *      //      -   3   ==>     -3   3
 *      //     / \              / \
 *      //    -   2           -1   2
 *      //   / \              / \
 *      //  0   1            0   1
 *
 * @symb R.reduce(f, a, [b, c, d]) = f(f(f(a, b), c), d)
 */


var reduce = /*#__PURE__*/_curry3_1(_reduce_1);
var reduce_1 = reduce;

/**
 * Returns a list of all the enumerable own properties of the supplied object.
 * Note that the order of the output array is not guaranteed across different
 * JS platforms.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig {k: v} -> [v]
 * @param {Object} obj The object to extract values from
 * @return {Array} An array of the values of the object's own properties.
 * @see R.valuesIn, R.keys
 * @example
 *
 *      R.values({a: 1, b: 2, c: 3}); //=> [1, 2, 3]
 */


var values = /*#__PURE__*/_curry1_1(function values(obj) {
  var props = keys_1(obj);
  var len = props.length;
  var vals = [];
  var idx = 0;
  while (idx < len) {
    vals[idx] = obj[props[idx]];
    idx += 1;
  }
  return vals;
});
var values_1 = values;

/**
 * Given a spec object recursively mapping properties to functions, creates a
 * function producing an object of the same structure, by mapping each property
 * to the result of calling its associated function with the supplied arguments.
 *
 * @func
 * @memberOf R
 * @since v0.20.0
 * @category Function
 * @sig {k: ((a, b, ..., m) -> v)} -> ((a, b, ..., m) -> {k: v})
 * @param {Object} spec an object recursively mapping properties to functions for
 *        producing the values for these properties.
 * @return {Function} A function that returns an object of the same structure
 * as `spec', with each property set to the value returned by calling its
 * associated function with the supplied arguments.
 * @see R.converge, R.juxt
 * @example
 *
 *      var getMetrics = R.applySpec({
 *        sum: R.add,
 *        nested: { mul: R.multiply }
 *      });
 *      getMetrics(2, 4); // => { sum: 6, nested: { mul: 8 } }
 * @symb R.applySpec({ x: f, y: { z: g } })(a, b) = { x: f(a, b), y: { z: g(a, b) } }
 */


var applySpec = /*#__PURE__*/_curry1_1(function applySpec(spec) {
  spec = map_1(function (v) {
    return typeof v == 'function' ? v : applySpec(v);
  }, spec);
  return curryN_1(reduce_1(max_1, 0, pluck_1('length', values_1(spec))), function () {
    var args = arguments;
    return map_1(function (f) {
      return apply_1(f, args);
    }, spec);
  });
});
var applySpec_1 = applySpec;

/**
 * Accepts a converging function and a list of branching functions and returns
 * a new function. When invoked, this new function is applied to some
 * arguments, each branching function is applied to those same arguments. The
 * results of each branching function are passed as arguments to the converging
 * function to produce the return value.
 *
 * @func
 * @memberOf R
 * @since v0.4.2
 * @category Function
 * @sig ((x1, x2, ...) -> z) -> [((a, b, ...) -> x1), ((a, b, ...) -> x2), ...] -> (a -> b -> ... -> z)
 * @param {Function} after A function. `after` will be invoked with the return values of
 *        `fn1` and `fn2` as its arguments.
 * @param {Array} functions A list of functions.
 * @return {Function} A new function.
 * @see R.useWith
 * @example
 *
 *      var average = R.converge(R.divide, [R.sum, R.length])
 *      average([1, 2, 3, 4, 5, 6, 7]) //=> 4
 *
 *      var strangeConcat = R.converge(R.concat, [R.toUpper, R.toLower])
 *      strangeConcat("Yodel") //=> "YODELyodel"
 *
 * @symb R.converge(f, [g, h])(a, b) = f(g(a, b), h(a, b))
 */


var converge = /*#__PURE__*/_curry2_1(function converge(after, fns) {
  return curryN_1(reduce_1(max_1, 0, pluck_1('length', fns)), function () {
    var args = arguments;
    var context = this;
    return after.apply(context, _map_1(function (fn) {
      return fn.apply(context, args);
    }, fns));
  });
});
var converge_1 = converge;

/**
 * juxt applies a list of functions to a list of values.
 *
 * @func
 * @memberOf R
 * @since v0.19.0
 * @category Function
 * @sig [(a, b, ..., m) -> n] -> ((a, b, ..., m) -> [n])
 * @param {Array} fns An array of functions
 * @return {Function} A function that returns a list of values after applying each of the original `fns` to its parameters.
 * @see R.applySpec
 * @example
 *
 *      var getRange = R.juxt([Math.min, Math.max]);
 *      getRange(3, 4, 9, -3); //=> [-3, 9]
 * @symb R.juxt([f, g, h])(a, b) = [f(a, b), g(a, b), h(a, b)]
 */


var juxt = /*#__PURE__*/_curry1_1(function juxt(fns) {
  return converge_1(function () {
    return Array.prototype.slice.call(arguments, 0);
  }, fns);
});
var juxt_1 = juxt;

/**
 * Determine if the passed argument is an integer.
 *
 * @private
 * @param {*} n
 * @category Type
 * @return {Boolean}
 */
var _isInteger = Number.isInteger || function _isInteger(n) {
  return n << 0 === n;
};

/**
 * Makes a shallow clone of an object, setting or overriding the specified
 * property with the given value. Note that this copies and flattens prototype
 * properties onto the new object as well. All non-primitive properties are
 * copied by reference.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Object
 * @sig String -> a -> {k: v} -> {k: v}
 * @param {String} prop The property name to set
 * @param {*} val The new value
 * @param {Object} obj The object to clone
 * @return {Object} A new object equivalent to the original except for the changed property.
 * @see R.dissoc
 * @example
 *
 *      R.assoc('c', 3, {a: 1, b: 2}); //=> {a: 1, b: 2, c: 3}
 */


var assoc = /*#__PURE__*/_curry3_1(function assoc(prop, val, obj) {
  var result = {};
  for (var p in obj) {
    result[p] = obj[p];
  }
  result[prop] = val;
  return result;
});
var assoc_1 = assoc;

/**
 * Checks if the input value is `null` or `undefined`.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Type
 * @sig * -> Boolean
 * @param {*} x The value to test.
 * @return {Boolean} `true` if `x` is `undefined` or `null`, otherwise `false`.
 * @example
 *
 *      R.isNil(null); //=> true
 *      R.isNil(undefined); //=> true
 *      R.isNil(0); //=> false
 *      R.isNil([]); //=> false
 */


var isNil = /*#__PURE__*/_curry1_1(function isNil(x) {
  return x == null;
});
var isNil_1 = isNil;

/**
 * Makes a shallow clone of an object, setting or overriding the nodes required
 * to create the given path, and placing the specific value at the tail end of
 * that path. Note that this copies and flattens prototype properties onto the
 * new object as well. All non-primitive properties are copied by reference.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Object
 * @typedefn Idx = String | Int
 * @sig [Idx] -> a -> {a} -> {a}
 * @param {Array} path the path to set
 * @param {*} val The new value
 * @param {Object} obj The object to clone
 * @return {Object} A new object equivalent to the original except along the specified path.
 * @see R.dissocPath
 * @example
 *
 *      R.assocPath(['a', 'b', 'c'], 42, {a: {b: {c: 0}}}); //=> {a: {b: {c: 42}}}
 *
 *      // Any missing or non-object keys in path will be overridden
 *      R.assocPath(['a', 'b', 'c'], 42, {a: 5}); //=> {a: {b: {c: 42}}}
 */


var assocPath = /*#__PURE__*/_curry3_1(function assocPath(path, val, obj) {
  if (path.length === 0) {
    return val;
  }
  var idx = path[0];
  if (path.length > 1) {
    var nextObj = !isNil_1(obj) && _has_1(idx, obj) ? obj[idx] : _isInteger(path[1]) ? [] : {};
    val = assocPath(Array.prototype.slice.call(path, 1), val, nextObj);
  }
  if (_isInteger(idx) && _isArray(obj)) {
    var arr = [].concat(obj);
    arr[idx] = val;
    return arr;
  } else {
    return assoc_1(idx, val, obj);
  }
});
var assocPath_1 = assocPath;

/**
* Takes a value and applies a function to it.
*
* This function is also known as the `thrush` combinator.
*
* @func
* @memberOf R
 * @since v0.25.0
* @category Function
* @sig a -> (a -> b) -> b
* @param {*} x The value
* @param {Function} f The function to apply
* @return {*} The result of applying `f` to `x`
* @example
*
*      var t42 = R.applyTo(42);
*      t42(R.identity); //=> 42
*      t42(R.add(1)); //=> 43
*/


var applyTo = /*#__PURE__*/_curry2_1(function applyTo(x, f) {
  return f(x);
});
var applyTo_1 = applyTo;

function _isObject(x) {
  return Object.prototype.toString.call(x) === '[object Object]';
}
var _isObject_1 = _isObject;

/**
 * Returns the empty value of its argument's type. Ramda defines the empty
 * value of Array (`[]`), Object (`{}`), String (`''`), and Arguments. Other
 * types are supported if they define `<Type>.empty`,
 * `<Type>.prototype.empty` or implement the
 * [FantasyLand Monoid spec](https://github.com/fantasyland/fantasy-land#monoid).
 *
 * Dispatches to the `empty` method of the first argument, if present.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category Function
 * @sig a -> a
 * @param {*} x
 * @return {*}
 * @example
 *
 *      R.empty(Just(42));      //=> Nothing()
 *      R.empty([1, 2, 3]);     //=> []
 *      R.empty('unicorns');    //=> ''
 *      R.empty({x: 1, y: 2});  //=> {}
 */


var empty = /*#__PURE__*/_curry1_1(function empty(x) {
  return x != null && typeof x['fantasy-land/empty'] === 'function' ? x['fantasy-land/empty']() : x != null && x.constructor != null && typeof x.constructor['fantasy-land/empty'] === 'function' ? x.constructor['fantasy-land/empty']() : x != null && typeof x.empty === 'function' ? x.empty() : x != null && x.constructor != null && typeof x.constructor.empty === 'function' ? x.constructor.empty() : _isArray(x) ? [] : _isString_1(x) ? '' : _isObject_1(x) ? {} : _isArguments_1(x) ? function () {
    return arguments;
  }() :
  // else
  void 0;
});
var empty_1 = empty;

function _arrayFromIterator(iter) {
  var list = [];
  var next;
  while (!(next = iter.next()).done) {
    list.push(next.value);
  }
  return list;
}
var _arrayFromIterator_1 = _arrayFromIterator;

function _containsWith(pred, x, list) {
  var idx = 0;
  var len = list.length;

  while (idx < len) {
    if (pred(x, list[idx])) {
      return true;
    }
    idx += 1;
  }
  return false;
}
var _containsWith_1 = _containsWith;

function _functionName(f) {
  // String(x => x) evaluates to "x => x", so the pattern may not match.
  var match = String(f).match(/^function (\w*)/);
  return match == null ? '' : match[1];
}
var _functionName_1 = _functionName;

/**
 * Returns true if its arguments are identical, false otherwise. Values are
 * identical if they reference the same memory. `NaN` is identical to `NaN`;
 * `0` and `-0` are not identical.
 *
 * @func
 * @memberOf R
 * @since v0.15.0
 * @category Relation
 * @sig a -> a -> Boolean
 * @param {*} a
 * @param {*} b
 * @return {Boolean}
 * @example
 *
 *      var o = {};
 *      R.identical(o, o); //=> true
 *      R.identical(1, 1); //=> true
 *      R.identical(1, '1'); //=> false
 *      R.identical([], []); //=> false
 *      R.identical(0, -0); //=> false
 *      R.identical(NaN, NaN); //=> true
 */


var identical = /*#__PURE__*/_curry2_1(function identical(a, b) {
  // SameValue algorithm
  if (a === b) {
    // Steps 1-5, 7-10
    // Steps 6.b-6.e: +0 != -0
    return a !== 0 || 1 / a === 1 / b;
  } else {
    // Step 6.a: NaN == NaN
    return a !== a && b !== b;
  }
});
var identical_1 = identical;

/**
 * Gives a single-word string description of the (native) type of a value,
 * returning such answers as 'Object', 'Number', 'Array', or 'Null'. Does not
 * attempt to distinguish user Object types any further, reporting them all as
 * 'Object'.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Type
 * @sig (* -> {*}) -> String
 * @param {*} val The value to test
 * @return {String}
 * @example
 *
 *      R.type({}); //=> "Object"
 *      R.type(1); //=> "Number"
 *      R.type(false); //=> "Boolean"
 *      R.type('s'); //=> "String"
 *      R.type(null); //=> "Null"
 *      R.type([]); //=> "Array"
 *      R.type(/[A-z]/); //=> "RegExp"
 *      R.type(() => {}); //=> "Function"
 *      R.type(undefined); //=> "Undefined"
 */


var type = /*#__PURE__*/_curry1_1(function type(val) {
  return val === null ? 'Null' : val === undefined ? 'Undefined' : Object.prototype.toString.call(val).slice(8, -1);
});
var type_1 = type;

/**
 * private _uniqContentEquals function.
 * That function is checking equality of 2 iterator contents with 2 assumptions
 * - iterators lengths are the same
 * - iterators values are unique
 *
 * false-positive result will be returned for comparision of, e.g.
 * - [1,2,3] and [1,2,3,4]
 * - [1,1,1] and [1,2,3]
 * */

function _uniqContentEquals(aIterator, bIterator, stackA, stackB) {
  var a = _arrayFromIterator_1(aIterator);
  var b = _arrayFromIterator_1(bIterator);

  function eq(_a, _b) {
    return _equals(_a, _b, stackA.slice(), stackB.slice());
  }

  // if *a* array contains any element that is not included in *b*
  return !_containsWith_1(function (b, aItem) {
    return !_containsWith_1(eq, aItem, b);
  }, b, a);
}

function _equals(a, b, stackA, stackB) {
  if (identical_1(a, b)) {
    return true;
  }

  var typeA = type_1(a);

  if (typeA !== type_1(b)) {
    return false;
  }

  if (a == null || b == null) {
    return false;
  }

  if (typeof a['fantasy-land/equals'] === 'function' || typeof b['fantasy-land/equals'] === 'function') {
    return typeof a['fantasy-land/equals'] === 'function' && a['fantasy-land/equals'](b) && typeof b['fantasy-land/equals'] === 'function' && b['fantasy-land/equals'](a);
  }

  if (typeof a.equals === 'function' || typeof b.equals === 'function') {
    return typeof a.equals === 'function' && a.equals(b) && typeof b.equals === 'function' && b.equals(a);
  }

  switch (typeA) {
    case 'Arguments':
    case 'Array':
    case 'Object':
      if (typeof a.constructor === 'function' && _functionName_1(a.constructor) === 'Promise') {
        return a === b;
      }
      break;
    case 'Boolean':
    case 'Number':
    case 'String':
      if (!(typeof a === typeof b && identical_1(a.valueOf(), b.valueOf()))) {
        return false;
      }
      break;
    case 'Date':
      if (!identical_1(a.valueOf(), b.valueOf())) {
        return false;
      }
      break;
    case 'Error':
      return a.name === b.name && a.message === b.message;
    case 'RegExp':
      if (!(a.source === b.source && a.global === b.global && a.ignoreCase === b.ignoreCase && a.multiline === b.multiline && a.sticky === b.sticky && a.unicode === b.unicode)) {
        return false;
      }
      break;
  }

  var idx = stackA.length - 1;
  while (idx >= 0) {
    if (stackA[idx] === a) {
      return stackB[idx] === b;
    }
    idx -= 1;
  }

  switch (typeA) {
    case 'Map':
      if (a.size !== b.size) {
        return false;
      }

      return _uniqContentEquals(a.entries(), b.entries(), stackA.concat([a]), stackB.concat([b]));
    case 'Set':
      if (a.size !== b.size) {
        return false;
      }

      return _uniqContentEquals(a.values(), b.values(), stackA.concat([a]), stackB.concat([b]));
    case 'Arguments':
    case 'Array':
    case 'Object':
    case 'Boolean':
    case 'Number':
    case 'String':
    case 'Date':
    case 'Error':
    case 'RegExp':
    case 'Int8Array':
    case 'Uint8Array':
    case 'Uint8ClampedArray':
    case 'Int16Array':
    case 'Uint16Array':
    case 'Int32Array':
    case 'Uint32Array':
    case 'Float32Array':
    case 'Float64Array':
    case 'ArrayBuffer':
      break;
    default:
      // Values of other types are only equal if identical.
      return false;
  }

  var keysA = keys_1(a);
  if (keysA.length !== keys_1(b).length) {
    return false;
  }

  var extendedStackA = stackA.concat([a]);
  var extendedStackB = stackB.concat([b]);

  idx = keysA.length - 1;
  while (idx >= 0) {
    var key = keysA[idx];
    if (!(_has_1(key, b) && _equals(b[key], a[key], extendedStackA, extendedStackB))) {
      return false;
    }
    idx -= 1;
  }
  return true;
}
var _equals_1 = _equals;

/**
 * Returns `true` if its arguments are equivalent, `false` otherwise. Handles
 * cyclical data structures.
 *
 * Dispatches symmetrically to the `equals` methods of both arguments, if
 * present.
 *
 * @func
 * @memberOf R
 * @since v0.15.0
 * @category Relation
 * @sig a -> b -> Boolean
 * @param {*} a
 * @param {*} b
 * @return {Boolean}
 * @example
 *
 *      R.equals(1, 1); //=> true
 *      R.equals(1, '1'); //=> false
 *      R.equals([1, 2, 3], [1, 2, 3]); //=> true
 *
 *      var a = {}; a.v = a;
 *      var b = {}; b.v = b;
 *      R.equals(a, b); //=> true
 */


var equals = /*#__PURE__*/_curry2_1(function equals(a, b) {
  return _equals_1(a, b, [], []);
});
var equals_1 = equals;

/**
 * Returns `true` if the given value is its type's empty value; `false`
 * otherwise.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Logic
 * @sig a -> Boolean
 * @param {*} x
 * @return {Boolean}
 * @see R.empty
 * @example
 *
 *      R.isEmpty([1, 2, 3]);   //=> false
 *      R.isEmpty([]);          //=> true
 *      R.isEmpty('');          //=> true
 *      R.isEmpty(null);        //=> false
 *      R.isEmpty({});          //=> true
 *      R.isEmpty({length: 0}); //=> false
 */


var isEmpty = /*#__PURE__*/_curry1_1(function isEmpty(x) {
  return x != null && equals_1(x, empty_1(x));
});
var isEmpty_1 = isEmpty;

/**
 * Returns `true` if the specified object property satisfies the given
 * predicate; `false` otherwise. You can test multiple properties with
 * [`R.where`](#where).
 *
 * @func
 * @memberOf R
 * @since v0.16.0
 * @category Logic
 * @sig (a -> Boolean) -> String -> {String: a} -> Boolean
 * @param {Function} pred
 * @param {String} name
 * @param {*} obj
 * @return {Boolean}
 * @see R.where, R.propEq, R.propIs
 * @example
 *
 *      R.propSatisfies(x => x > 0, 'x', {x: 1, y: 2}); //=> true
 */


var propSatisfies = /*#__PURE__*/_curry3_1(function propSatisfies(pred, name, obj) {
  return pred(obj[name]);
});
var propSatisfies_1 = propSatisfies;

/**
 * A function that returns the `!` of its argument. It will return `true` when
 * passed false-y value, and `false` when passed a truth-y one.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Logic
 * @sig * -> Boolean
 * @param {*} a any value
 * @return {Boolean} the logical inverse of passed argument.
 * @see R.complement
 * @example
 *
 *      R.not(true); //=> false
 *      R.not(false); //=> true
 *      R.not(0); //=> true
 *      R.not(1); //=> false
 */


var not = /*#__PURE__*/_curry1_1(function not(a) {
  return !a;
});
var not_1 = not;

/**
 * Takes a list of predicates and returns a predicate that returns true for a
 * given list of arguments if at least one of the provided predicates is
 * satisfied by those arguments.
 *
 * The function returned is a curried function whose arity matches that of the
 * highest-arity predicate.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Logic
 * @sig [(*... -> Boolean)] -> (*... -> Boolean)
 * @param {Array} predicates An array of predicates to check
 * @return {Function} The combined predicate
 * @see R.allPass
 * @example
 *
 *      var isClub = R.propEq('suit', '♣');
 *      var isSpade = R.propEq('suit', '♠');
 *      var isBlackCard = R.anyPass([isClub, isSpade]);
 *
 *      isBlackCard({rank: '10', suit: '♣'}); //=> true
 *      isBlackCard({rank: 'Q', suit: '♠'}); //=> true
 *      isBlackCard({rank: 'Q', suit: '♦'}); //=> false
 */


var anyPass = /*#__PURE__*/_curry1_1(function anyPass(preds) {
  return curryN_1(reduce_1(max_1, 0, pluck_1('length', preds)), function () {
    var idx = 0;
    var len = preds.length;
    while (idx < len) {
      if (preds[idx].apply(this, arguments)) {
        return true;
      }
      idx += 1;
    }
    return false;
  });
});
var anyPass_1 = anyPass;

/**
 * Takes a list of predicates and returns a predicate that returns true for a
 * given list of arguments if every one of the provided predicates is satisfied
 * by those arguments.
 *
 * The function returned is a curried function whose arity matches that of the
 * highest-arity predicate.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Logic
 * @sig [(*... -> Boolean)] -> (*... -> Boolean)
 * @param {Array} predicates An array of predicates to check
 * @return {Function} The combined predicate
 * @see R.anyPass
 * @example
 *
 *      var isQueen = R.propEq('rank', 'Q');
 *      var isSpade = R.propEq('suit', '♠︎');
 *      var isQueenOfSpades = R.allPass([isQueen, isSpade]);
 *
 *      isQueenOfSpades({rank: 'Q', suit: '♣︎'}); //=> false
 *      isQueenOfSpades({rank: 'Q', suit: '♠︎'}); //=> true
 */


var allPass = /*#__PURE__*/_curry1_1(function allPass(preds) {
  return curryN_1(reduce_1(max_1, 0, pluck_1('length', preds)), function () {
    var idx = 0;
    var len = preds.length;
    while (idx < len) {
      if (!preds[idx].apply(this, arguments)) {
        return false;
      }
      idx += 1;
    }
    return true;
  });
});
var allPass_1 = allPass;

/**
 * Returns a curried equivalent of the provided function. The curried function
 * has two unusual capabilities. First, its arguments needn't be provided one
 * at a time. If `f` is a ternary function and `g` is `R.curry(f)`, the
 * following are equivalent:
 *
 *   - `g(1)(2)(3)`
 *   - `g(1)(2, 3)`
 *   - `g(1, 2)(3)`
 *   - `g(1, 2, 3)`
 *
 * Secondly, the special placeholder value [`R.__`](#__) may be used to specify
 * "gaps", allowing partial application of any combination of arguments,
 * regardless of their positions. If `g` is as above and `_` is [`R.__`](#__),
 * the following are equivalent:
 *
 *   - `g(1, 2, 3)`
 *   - `g(_, 2, 3)(1)`
 *   - `g(_, _, 3)(1)(2)`
 *   - `g(_, _, 3)(1, 2)`
 *   - `g(_, 2)(1)(3)`
 *   - `g(_, 2)(1, 3)`
 *   - `g(_, 2)(_, 3)(1)`
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig (* -> a) -> (* -> a)
 * @param {Function} fn The function to curry.
 * @return {Function} A new, curried function.
 * @see R.curryN
 * @example
 *
 *      var addFourNumbers = (a, b, c, d) => a + b + c + d;
 *
 *      var curriedAddFourNumbers = R.curry(addFourNumbers);
 *      var f = curriedAddFourNumbers(1, 2);
 *      var g = f(3);
 *      g(4); //=> 10
 */


var curry = /*#__PURE__*/_curry1_1(function curry(fn) {
  return curryN_1(fn.length, fn);
});
var curry_1 = curry;

/**
 * Returns the result of calling its first argument with the remaining
 * arguments. This is occasionally useful as a converging function for
 * [`R.converge`](#converge): the first branch can produce a function while the
 * remaining branches produce values to be passed to that function as its
 * arguments.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Function
 * @sig (*... -> a),*... -> a
 * @param {Function} fn The function to apply to the remaining arguments.
 * @param {...*} args Any number of positional arguments.
 * @return {*}
 * @see R.apply
 * @example
 *
 *      R.call(R.add, 1, 2); //=> 3
 *
 *      var indentN = R.pipe(R.repeat(' '),
 *                           R.join(''),
 *                           R.replace(/^(?!$)/gm));
 *
 *      var format = R.converge(R.call, [
 *                                  R.pipe(R.prop('indent'), indentN),
 *                                  R.prop('value')
 *                              ]);
 *
 *      format({indent: 2, value: 'foo\nbar\nbaz\n'}); //=> '  foo\n  bar\n  baz\n'
 * @symb R.call(f, a, b) = f(a, b)
 */


var call = /*#__PURE__*/curry_1(function call(fn) {
  return fn.apply(this, Array.prototype.slice.call(arguments, 1));
});
var call_1 = call;

/**
 * Returns `true` if the specified object property at given path satisfies the
 * given predicate; `false` otherwise.
 *
 * @func
 * @memberOf R
 * @since v0.19.0
 * @category Logic
 * @typedefn Idx = String | Int
 * @sig (a -> Boolean) -> [Idx] -> {a} -> Boolean
 * @param {Function} pred
 * @param {Array} propPath
 * @param {*} obj
 * @return {Boolean}
 * @see R.propSatisfies, R.path
 * @example
 *
 *      R.pathSatisfies(y => y > 0, ['x', 'y'], {x: {y: 2}}); //=> true
 */


var pathSatisfies = /*#__PURE__*/_curry3_1(function pathSatisfies(pred, propPath, obj) {
  return propPath.length > 0 && pred(path_1(propPath, obj));
});
var pathSatisfies_1 = pathSatisfies;

/**
 * Creates a function that will process either the `onTrue` or the `onFalse`
 * function depending upon the result of the `condition` predicate.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Logic
 * @sig (*... -> Boolean) -> (*... -> *) -> (*... -> *) -> (*... -> *)
 * @param {Function} condition A predicate function
 * @param {Function} onTrue A function to invoke when the `condition` evaluates to a truthy value.
 * @param {Function} onFalse A function to invoke when the `condition` evaluates to a falsy value.
 * @return {Function} A new unary function that will process either the `onTrue` or the `onFalse`
 *                    function depending upon the result of the `condition` predicate.
 * @see R.unless, R.when
 * @example
 *
 *      var incCount = R.ifElse(
 *        R.has('count'),
 *        R.over(R.lensProp('count'), R.inc),
 *        R.assoc('count', 1)
 *      );
 *      incCount({});           //=> { count: 1 }
 *      incCount({ count: 1 }); //=> { count: 2 }
 */


var ifElse = /*#__PURE__*/_curry3_1(function ifElse(condition, onTrue, onFalse) {
  return curryN_1(Math.max(condition.length, onTrue.length, onFalse.length), function _ifElse() {
    return condition.apply(this, arguments) ? onTrue.apply(this, arguments) : onFalse.apply(this, arguments);
  });
});
var ifElse_1 = ifElse;

/**
 * Returns `true` if the specified object property is equal, in
 * [`R.equals`](#equals) terms, to the given value; `false` otherwise.
 * You can test multiple properties with [`R.where`](#where).
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Relation
 * @sig String -> a -> Object -> Boolean
 * @param {String} name
 * @param {*} val
 * @param {*} obj
 * @return {Boolean}
 * @see R.whereEq, R.propSatisfies, R.equals
 * @example
 *
 *      var abby = {name: 'Abby', age: 7, hair: 'blond'};
 *      var fred = {name: 'Fred', age: 12, hair: 'brown'};
 *      var rusty = {name: 'Rusty', age: 10, hair: 'brown'};
 *      var alois = {name: 'Alois', age: 15, disposition: 'surly'};
 *      var kids = [abby, fred, rusty, alois];
 *      var hasBrownHair = R.propEq('hair', 'brown');
 *      R.filter(hasBrownHair, kids); //=> [fred, rusty]
 */


var propEq = /*#__PURE__*/_curry3_1(function propEq(name, val, obj) {
  return equals_1(val, obj[name]);
});
var propEq_1 = propEq;

/**
 * Returns a partial copy of an object containing only the keys specified. If
 * the key does not exist, the property is ignored.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig [k] -> {k: v} -> {k: v}
 * @param {Array} names an array of String property names to copy onto a new object
 * @param {Object} obj The object to copy from
 * @return {Object} A new object with only properties from `names` on it.
 * @see R.omit, R.props
 * @example
 *
 *      R.pick(['a', 'd'], {a: 1, b: 2, c: 3, d: 4}); //=> {a: 1, d: 4}
 *      R.pick(['a', 'e', 'f'], {a: 1, b: 2, c: 3, d: 4}); //=> {a: 1}
 */


var pick = /*#__PURE__*/_curry2_1(function pick(names, obj) {
  var result = {};
  var idx = 0;
  while (idx < names.length) {
    if (names[idx] in obj) {
      result[names[idx]] = obj[names[idx]];
    }
    idx += 1;
  }
  return result;
});
var pick_1 = pick;

/**
 * Creates a new object by recursively evolving a shallow copy of `object`,
 * according to the `transformation` functions. All non-primitive properties
 * are copied by reference.
 *
 * A `transformation` function will not be invoked if its corresponding key
 * does not exist in the evolved object.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Object
 * @sig {k: (v -> v)} -> {k: v} -> {k: v}
 * @param {Object} transformations The object specifying transformation functions to apply
 *        to the object.
 * @param {Object} object The object to be transformed.
 * @return {Object} The transformed object.
 * @example
 *
 *      var tomato  = {firstName: '  Tomato ', data: {elapsed: 100, remaining: 1400}, id:123};
 *      var transformations = {
 *        firstName: R.trim,
 *        lastName: R.trim, // Will not get invoked.
 *        data: {elapsed: R.add(1), remaining: R.add(-1)}
 *      };
 *      R.evolve(transformations, tomato); //=> {firstName: 'Tomato', data: {elapsed: 101, remaining: 1399}, id:123}
 */


var evolve = /*#__PURE__*/_curry2_1(function evolve(transformations, object) {
  var result = {};
  var transformation, key, type;
  for (key in object) {
    transformation = transformations[key];
    type = typeof transformation;
    result[key] = type === 'function' ? transformation(object[key]) : transformation && type === 'object' ? evolve(transformation, object[key]) : object[key];
  }
  return result;
});
var evolve_1 = evolve;

/**
 * Creates a new object with the own properties of the two provided objects. If
 * a key exists in both objects, the provided function is applied to the key
 * and the values associated with the key in each object, with the result being
 * used as the value associated with the key in the returned object.
 *
 * @func
 * @memberOf R
 * @since v0.19.0
 * @category Object
 * @sig ((String, a, a) -> a) -> {a} -> {a} -> {a}
 * @param {Function} fn
 * @param {Object} l
 * @param {Object} r
 * @return {Object}
 * @see R.mergeDeepWithKey, R.merge, R.mergeWith
 * @example
 *
 *      let concatValues = (k, l, r) => k == 'values' ? R.concat(l, r) : r
 *      R.mergeWithKey(concatValues,
 *                     { a: true, thing: 'foo', values: [10, 20] },
 *                     { b: true, thing: 'bar', values: [15, 35] });
 *      //=> { a: true, b: true, thing: 'bar', values: [10, 20, 15, 35] }
 * @symb R.mergeWithKey(f, { x: 1, y: 2 }, { y: 5, z: 3 }) = { x: 1, y: f('y', 2, 5), z: 3 }
 */


var mergeWithKey = /*#__PURE__*/_curry3_1(function mergeWithKey(fn, l, r) {
  var result = {};
  var k;

  for (k in l) {
    if (_has_1(k, l)) {
      result[k] = _has_1(k, r) ? fn(k, l[k], r[k]) : l[k];
    }
  }

  for (k in r) {
    if (_has_1(k, r) && !_has_1(k, result)) {
      result[k] = r[k];
    }
  }

  return result;
});
var mergeWithKey_1 = mergeWithKey;

/**
 * Creates a new object with the own properties of the two provided objects.
 * If a key exists in both objects:
 * - and both associated values are also objects then the values will be
 *   recursively merged.
 * - otherwise the provided function is applied to the key and associated values
 *   using the resulting value as the new value associated with the key.
 * If a key only exists in one object, the value will be associated with the key
 * of the resulting object.
 *
 * @func
 * @memberOf R
 * @since v0.24.0
 * @category Object
 * @sig ((String, a, a) -> a) -> {a} -> {a} -> {a}
 * @param {Function} fn
 * @param {Object} lObj
 * @param {Object} rObj
 * @return {Object}
 * @see R.mergeWithKey, R.mergeDeep, R.mergeDeepWith
 * @example
 *
 *      let concatValues = (k, l, r) => k == 'values' ? R.concat(l, r) : r
 *      R.mergeDeepWithKey(concatValues,
 *                         { a: true, c: { thing: 'foo', values: [10, 20] }},
 *                         { b: true, c: { thing: 'bar', values: [15, 35] }});
 *      //=> { a: true, b: true, c: { thing: 'bar', values: [10, 20, 15, 35] }}
 */


var mergeDeepWithKey = /*#__PURE__*/_curry3_1(function mergeDeepWithKey(fn, lObj, rObj) {
  return mergeWithKey_1(function (k, lVal, rVal) {
    if (_isObject_1(lVal) && _isObject_1(rVal)) {
      return mergeDeepWithKey(fn, lVal, rVal);
    } else {
      return fn(k, lVal, rVal);
    }
  }, lObj, rObj);
});
var mergeDeepWithKey_1 = mergeDeepWithKey;

/**
 * Creates a new object with the own properties of the first object merged with
 * the own properties of the second object. If a key exists in both objects:
 * - and both values are objects, the two values will be recursively merged
 * - otherwise the value from the second object will be used.
 *
 * @func
 * @memberOf R
 * @since v0.24.0
 * @category Object
 * @sig {a} -> {a} -> {a}
 * @param {Object} lObj
 * @param {Object} rObj
 * @return {Object}
 * @see R.merge, R.mergeDeepLeft, R.mergeDeepWith, R.mergeDeepWithKey
 * @example
 *
 *      R.mergeDeepRight({ name: 'fred', age: 10, contact: { email: 'moo@example.com' }},
 *                       { age: 40, contact: { email: 'baa@example.com' }});
 *      //=> { name: 'fred', age: 40, contact: { email: 'baa@example.com' }}
 */


var mergeDeepRight = /*#__PURE__*/_curry2_1(function mergeDeepRight(lObj, rObj) {
  return mergeDeepWithKey_1(function (k, lVal, rVal) {
    return rVal;
  }, lObj, rObj);
});
var mergeDeepRight_1 = mergeDeepRight;

/**
 * Creates an object containing a single key:value pair.
 *
 * @func
 * @memberOf R
 * @since v0.18.0
 * @category Object
 * @sig String -> a -> {String:a}
 * @param {String} key
 * @param {*} val
 * @return {Object}
 * @see R.pair
 * @example
 *
 *      var matchPhrases = R.compose(
 *        R.objOf('must'),
 *        R.map(R.objOf('match_phrase'))
 *      );
 *      matchPhrases(['foo', 'bar', 'baz']); //=> {must: [{match_phrase: 'foo'}, {match_phrase: 'bar'}, {match_phrase: 'baz'}]}
 */


var objOf = /*#__PURE__*/_curry2_1(function objOf(key, val) {
  var obj = {};
  obj[key] = val;
  return obj;
});
var objOf_1 = objOf;

/**
 * Private `concat` function to merge two array-like objects.
 *
 * @private
 * @param {Array|Arguments} [set1=[]] An array-like object.
 * @param {Array|Arguments} [set2=[]] An array-like object.
 * @return {Array} A new, merged array.
 * @example
 *
 *      _concat([4, 5, 6], [1, 2, 3]); //=> [4, 5, 6, 1, 2, 3]
 */
function _concat(set1, set2) {
  set1 = set1 || [];
  set2 = set2 || [];
  var idx;
  var len1 = set1.length;
  var len2 = set2.length;
  var result = [];

  idx = 0;
  while (idx < len1) {
    result[result.length] = set1[idx];
    idx += 1;
  }
  idx = 0;
  while (idx < len2) {
    result[result.length] = set2[idx];
    idx += 1;
  }
  return result;
}
var _concat_1 = _concat;

/**
 * ap applies a list of functions to a list of values.
 *
 * Dispatches to the `ap` method of the second argument, if present. Also
 * treats curried functions as applicatives.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category Function
 * @sig [a -> b] -> [a] -> [b]
 * @sig Apply f => f (a -> b) -> f a -> f b
 * @sig (a -> b -> c) -> (a -> b) -> (a -> c)
 * @param {*} applyF
 * @param {*} applyX
 * @return {*}
 * @example
 *
 *      R.ap([R.multiply(2), R.add(3)], [1,2,3]); //=> [2, 4, 6, 4, 5, 6]
 *      R.ap([R.concat('tasty '), R.toUpper], ['pizza', 'salad']); //=> ["tasty pizza", "tasty salad", "PIZZA", "SALAD"]
 *
 *      // R.ap can also be used as S combinator
 *      // when only two functions are passed
 *      R.ap(R.concat, R.toUpper)('Ramda') //=> 'RamdaRAMDA'
 * @symb R.ap([f, g], [a, b]) = [f(a), f(b), g(a), g(b)]
 */


var ap = /*#__PURE__*/_curry2_1(function ap(applyF, applyX) {
  return typeof applyX['fantasy-land/ap'] === 'function' ? applyX['fantasy-land/ap'](applyF) : typeof applyF.ap === 'function' ? applyF.ap(applyX) : typeof applyF === 'function' ? function (x) {
    return applyF(x)(applyX(x));
  } :
  // else
  _reduce_1(function (acc, f) {
    return _concat_1(acc, map_1(f, applyX));
  }, [], applyF);
});
var ap_1 = ap;

function _of(x) {
  return [x];
}
var _of_1 = _of;

/**
 * Returns a singleton array containing the value provided.
 *
 * Note this `of` is different from the ES6 `of`; See
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/of
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category Function
 * @sig a -> [a]
 * @param {*} x any value
 * @return {Array} An array wrapping `x`.
 * @example
 *
 *      R.of(null); //=> [null]
 *      R.of([42]); //=> [[42]]
 */


var of = /*#__PURE__*/_curry1_1(_of_1);
var of_1 = of;

// Based on https://developer.mozilla.org/en/docs/Web/JavaScript/Reference/Global_Objects/Object/assign


function _objectAssign(target) {
  if (target == null) {
    throw new TypeError('Cannot convert undefined or null to object');
  }

  var output = Object(target);
  var idx = 1;
  var length = arguments.length;
  while (idx < length) {
    var source = arguments[idx];
    if (source != null) {
      for (var nextKey in source) {
        if (_has_1(nextKey, source)) {
          output[nextKey] = source[nextKey];
        }
      }
    }
    idx += 1;
  }
  return output;
}
var _objectAssign_1 = _objectAssign;

var _assign = typeof Object.assign === 'function' ? Object.assign : _objectAssign_1;

/**
 * Merges a list of objects together into one object.
 *
 * @func
 * @memberOf R
 * @since v0.10.0
 * @category List
 * @sig [{k: v}] -> {k: v}
 * @param {Array} list An array of objects
 * @return {Object} A merged object.
 * @see R.reduce
 * @example
 *
 *      R.mergeAll([{foo:1},{bar:2},{baz:3}]); //=> {foo:1,bar:2,baz:3}
 *      R.mergeAll([{foo:1},{foo:2},{bar:2}]); //=> {foo:2,bar:2}
 * @symb R.mergeAll([{ x: 1 }, { y: 2 }, { z: 3 }]) = { x: 1, y: 2, z: 3 }
 */


var mergeAll = /*#__PURE__*/_curry1_1(function mergeAll(list) {
  return _assign.apply(null, [{}].concat(list));
});
var mergeAll_1 = mergeAll;

function _isFunction(x) {
  return Object.prototype.toString.call(x) === '[object Function]';
}
var _isFunction_1 = _isFunction;

function _indexOf(list, a, idx) {
  var inf, item;
  // Array.prototype.indexOf doesn't exist below IE9
  if (typeof list.indexOf === 'function') {
    switch (typeof a) {
      case 'number':
        if (a === 0) {
          // manually crawl the list to distinguish between +0 and -0
          inf = 1 / a;
          while (idx < list.length) {
            item = list[idx];
            if (item === 0 && 1 / item === inf) {
              return idx;
            }
            idx += 1;
          }
          return -1;
        } else if (a !== a) {
          // NaN
          while (idx < list.length) {
            item = list[idx];
            if (typeof item === 'number' && item !== item) {
              return idx;
            }
            idx += 1;
          }
          return -1;
        }
        // non-zero numbers can utilise Set
        return list.indexOf(a, idx);

      // all these types can utilise Set
      case 'string':
      case 'boolean':
      case 'function':
      case 'undefined':
        return list.indexOf(a, idx);

      case 'object':
        if (a === null) {
          // null can utilise Set
          return list.indexOf(a, idx);
        }
    }
  }
  // anything else not covered above, defer to R.equals
  while (idx < list.length) {
    if (equals_1(list[idx], a)) {
      return idx;
    }
    idx += 1;
  }
  return -1;
}
var _indexOf_1 = _indexOf;

function _contains(a, list) {
  return _indexOf_1(list, a, 0) >= 0;
}
var _contains_1 = _contains;

function _quote(s) {
  var escaped = s.replace(/\\/g, '\\\\').replace(/[\b]/g, '\\b') // \b matches word boundary; [\b] matches backspace
  .replace(/\f/g, '\\f').replace(/\n/g, '\\n').replace(/\r/g, '\\r').replace(/\t/g, '\\t').replace(/\v/g, '\\v').replace(/\0/g, '\\0');

  return '"' + escaped.replace(/"/g, '\\"') + '"';
}
var _quote_1 = _quote;

/**
 * Polyfill from <https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/toISOString>.
 */
var pad = function pad(n) {
  return (n < 10 ? '0' : '') + n;
};

var _toISOString = typeof Date.prototype.toISOString === 'function' ? function _toISOString(d) {
  return d.toISOString();
} : function _toISOString(d) {
  return d.getUTCFullYear() + '-' + pad(d.getUTCMonth() + 1) + '-' + pad(d.getUTCDate()) + 'T' + pad(d.getUTCHours()) + ':' + pad(d.getUTCMinutes()) + ':' + pad(d.getUTCSeconds()) + '.' + (d.getUTCMilliseconds() / 1000).toFixed(3).slice(2, 5) + 'Z';
};

var _toISOString_1 = _toISOString;

function _complement(f) {
  return function () {
    return !f.apply(this, arguments);
  };
}
var _complement_1 = _complement;

function _filter(fn, list) {
  var idx = 0;
  var len = list.length;
  var result = [];

  while (idx < len) {
    if (fn(list[idx])) {
      result[result.length] = list[idx];
    }
    idx += 1;
  }
  return result;
}
var _filter_1 = _filter;

var XFilter = /*#__PURE__*/function () {

  function XFilter(f, xf) {
    this.xf = xf;
    this.f = f;
  }
  XFilter.prototype['@@transducer/init'] = _xfBase.init;
  XFilter.prototype['@@transducer/result'] = _xfBase.result;
  XFilter.prototype['@@transducer/step'] = function (result, input) {
    return this.f(input) ? this.xf['@@transducer/step'](result, input) : result;
  };

  return XFilter;
}();

var _xfilter = /*#__PURE__*/_curry2_1(function _xfilter(f, xf) {
  return new XFilter(f, xf);
});
var _xfilter_1 = _xfilter;

/**
 * Takes a predicate and a `Filterable`, and returns a new filterable of the
 * same type containing the members of the given filterable which satisfy the
 * given predicate. Filterable objects include plain objects or any object
 * that has a filter method such as `Array`.
 *
 * Dispatches to the `filter` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Filterable f => (a -> Boolean) -> f a -> f a
 * @param {Function} pred
 * @param {Array} filterable
 * @return {Array} Filterable
 * @see R.reject, R.transduce, R.addIndex
 * @example
 *
 *      var isEven = n => n % 2 === 0;
 *
 *      R.filter(isEven, [1, 2, 3, 4]); //=> [2, 4]
 *
 *      R.filter(isEven, {a: 1, b: 2, c: 3, d: 4}); //=> {b: 2, d: 4}
 */


var filter = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['filter'], _xfilter_1, function (pred, filterable) {
  return _isObject_1(filterable) ? _reduce_1(function (acc, key) {
    if (pred(filterable[key])) {
      acc[key] = filterable[key];
    }
    return acc;
  }, {}, keys_1(filterable)) :
  // else
  _filter_1(pred, filterable);
}));
var filter_1 = filter;

/**
 * The complement of [`filter`](#filter).
 *
 * Acts as a transducer if a transformer is given in list position. Filterable
 * objects include plain objects or any object that has a filter method such
 * as `Array`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Filterable f => (a -> Boolean) -> f a -> f a
 * @param {Function} pred
 * @param {Array} filterable
 * @return {Array}
 * @see R.filter, R.transduce, R.addIndex
 * @example
 *
 *      var isOdd = (n) => n % 2 === 1;
 *
 *      R.reject(isOdd, [1, 2, 3, 4]); //=> [2, 4]
 *
 *      R.reject(isOdd, {a: 1, b: 2, c: 3, d: 4}); //=> {b: 2, d: 4}
 */


var reject = /*#__PURE__*/_curry2_1(function reject(pred, filterable) {
  return filter_1(_complement_1(pred), filterable);
});
var reject_1 = reject;

function _toString(x, seen) {
  var recur = function recur(y) {
    var xs = seen.concat([x]);
    return _contains_1(y, xs) ? '<Circular>' : _toString(y, xs);
  };

  //  mapPairs :: (Object, [String]) -> [String]
  var mapPairs = function (obj, keys) {
    return _map_1(function (k) {
      return _quote_1(k) + ': ' + recur(obj[k]);
    }, keys.slice().sort());
  };

  switch (Object.prototype.toString.call(x)) {
    case '[object Arguments]':
      return '(function() { return arguments; }(' + _map_1(recur, x).join(', ') + '))';
    case '[object Array]':
      return '[' + _map_1(recur, x).concat(mapPairs(x, reject_1(function (k) {
        return (/^\d+$/.test(k)
        );
      }, keys_1(x)))).join(', ') + ']';
    case '[object Boolean]':
      return typeof x === 'object' ? 'new Boolean(' + recur(x.valueOf()) + ')' : x.toString();
    case '[object Date]':
      return 'new Date(' + (isNaN(x.valueOf()) ? recur(NaN) : _quote_1(_toISOString_1(x))) + ')';
    case '[object Null]':
      return 'null';
    case '[object Number]':
      return typeof x === 'object' ? 'new Number(' + recur(x.valueOf()) + ')' : 1 / x === -Infinity ? '-0' : x.toString(10);
    case '[object String]':
      return typeof x === 'object' ? 'new String(' + recur(x.valueOf()) + ')' : _quote_1(x);
    case '[object Undefined]':
      return 'undefined';
    default:
      if (typeof x.toString === 'function') {
        var repr = x.toString();
        if (repr !== '[object Object]') {
          return repr;
        }
      }
      return '{' + mapPairs(x, keys_1(x)).join(', ') + '}';
  }
}
var _toString_1 = _toString;

/**
 * Returns the string representation of the given value. `eval`'ing the output
 * should result in a value equivalent to the input value. Many of the built-in
 * `toString` methods do not satisfy this requirement.
 *
 * If the given value is an `[object Object]` with a `toString` method other
 * than `Object.prototype.toString`, this method is invoked with no arguments
 * to produce the return value. This means user-defined constructor functions
 * can provide a suitable `toString` method. For example:
 *
 *     function Point(x, y) {
 *       this.x = x;
 *       this.y = y;
 *     }
 *
 *     Point.prototype.toString = function() {
 *       return 'new Point(' + this.x + ', ' + this.y + ')';
 *     };
 *
 *     R.toString(new Point(1, 2)); //=> 'new Point(1, 2)'
 *
 * @func
 * @memberOf R
 * @since v0.14.0
 * @category String
 * @sig * -> String
 * @param {*} val
 * @return {String}
 * @example
 *
 *      R.toString(42); //=> '42'
 *      R.toString('abc'); //=> '"abc"'
 *      R.toString([1, 2, 3]); //=> '[1, 2, 3]'
 *      R.toString({foo: 1, bar: 2, baz: 3}); //=> '{"bar": 2, "baz": 3, "foo": 1}'
 *      R.toString(new Date('2001-02-03T04:05:06Z')); //=> 'new Date("2001-02-03T04:05:06.000Z")'
 */


var toString$2 = /*#__PURE__*/_curry1_1(function toString(val) {
  return _toString_1(val, []);
});
var toString_1 = toString$2;

/**
 * Turns a named method with a specified arity into a function that can be
 * called directly supplied with arguments and a target object.
 *
 * The returned function is curried and accepts `arity + 1` parameters where
 * the final parameter is the target object.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig Number -> String -> (a -> b -> ... -> n -> Object -> *)
 * @param {Number} arity Number of arguments the returned function should take
 *        before the target object.
 * @param {String} method Name of the method to call.
 * @return {Function} A new curried function.
 * @see R.construct
 * @example
 *
 *      var sliceFrom = R.invoker(1, 'slice');
 *      sliceFrom(6, 'abcdefghijklm'); //=> 'ghijklm'
 *      var sliceFrom6 = R.invoker(2, 'slice')(6);
 *      sliceFrom6(8, 'abcdefghijklm'); //=> 'gh'
 * @symb R.invoker(0, 'method')(o) = o['method']()
 * @symb R.invoker(1, 'method')(a, o) = o['method'](a)
 * @symb R.invoker(2, 'method')(a, b, o) = o['method'](a, b)
 */


var invoker = /*#__PURE__*/_curry2_1(function invoker(arity, method) {
  return curryN_1(arity + 1, function () {
    var target = arguments[arity];
    if (target != null && _isFunction_1(target[method])) {
      return target[method].apply(target, Array.prototype.slice.call(arguments, 0, arity));
    }
    throw new TypeError(toString_1(target) + ' does not have a method named "' + method + '"');
  });
});
var invoker_1 = invoker;

/**
 * Splits a string into an array of strings based on the given
 * separator.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category String
 * @sig (String | RegExp) -> String -> [String]
 * @param {String|RegExp} sep The pattern.
 * @param {String} str The string to separate into an array.
 * @return {Array} The array of strings from `str` separated by `str`.
 * @see R.join
 * @example
 *
 *      var pathComponents = R.split('/');
 *      R.tail(pathComponents('/usr/local/bin/node')); //=> ['usr', 'local', 'bin', 'node']
 *
 *      R.split('.', 'a.b.c.xyz.d'); //=> ['a', 'b', 'c', 'xyz', 'd']
 */


var split = /*#__PURE__*/invoker_1(1, 'split');
var split_1 = split;

/**
 * Returns the nth element of the given list or string. If n is negative the
 * element at index length + n is returned.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Number -> [a] -> a | Undefined
 * @sig Number -> String -> String
 * @param {Number} offset
 * @param {*} list
 * @return {*}
 * @example
 *
 *      var list = ['foo', 'bar', 'baz', 'quux'];
 *      R.nth(1, list); //=> 'bar'
 *      R.nth(-1, list); //=> 'quux'
 *      R.nth(-99, list); //=> undefined
 *
 *      R.nth(2, 'abc'); //=> 'c'
 *      R.nth(3, 'abc'); //=> ''
 * @symb R.nth(-1, [a, b, c]) = c
 * @symb R.nth(0, [a, b, c]) = a
 * @symb R.nth(1, [a, b, c]) = b
 */


var nth = /*#__PURE__*/_curry2_1(function nth(offset, list) {
  var idx = offset < 0 ? list.length + offset : offset;
  return _isString_1(list) ? list.charAt(idx) : list[idx];
});
var nth_1 = nth;

/**
 * Returns the first element of the given list or string. In some libraries
 * this function is named `first`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> a | Undefined
 * @sig String -> String
 * @param {Array|String} list
 * @return {*}
 * @see R.tail, R.init, R.last
 * @example
 *
 *      R.head(['fi', 'fo', 'fum']); //=> 'fi'
 *      R.head([]); //=> undefined
 *
 *      R.head('abc'); //=> 'a'
 *      R.head(''); //=> ''
 */


var head = /*#__PURE__*/nth_1(0);
var head_1 = head;

/**
 * "lifts" a function to be the specified arity, so that it may "map over" that
 * many lists, Functions or other objects that satisfy the [FantasyLand Apply spec](https://github.com/fantasyland/fantasy-land#apply).
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category Function
 * @sig Number -> (*... -> *) -> ([*]... -> [*])
 * @param {Function} fn The function to lift into higher context
 * @return {Function} The lifted function.
 * @see R.lift, R.ap
 * @example
 *
 *      var madd3 = R.liftN(3, (...args) => R.sum(args));
 *      madd3([1,2,3], [1,2,3], [1]); //=> [3, 4, 5, 4, 5, 6, 5, 6, 7]
 */


var liftN = /*#__PURE__*/_curry2_1(function liftN(arity, fn) {
  var lifted = curryN_1(arity, fn);
  return curryN_1(arity, function () {
    return _reduce_1(ap_1, map_1(lifted, arguments[0]), Array.prototype.slice.call(arguments, 1));
  });
});
var liftN_1 = liftN;

/**
 * "lifts" a function of arity > 1 so that it may "map over" a list, Function or other
 * object that satisfies the [FantasyLand Apply spec](https://github.com/fantasyland/fantasy-land#apply).
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category Function
 * @sig (*... -> *) -> ([*]... -> [*])
 * @param {Function} fn The function to lift into higher context
 * @return {Function} The lifted function.
 * @see R.liftN
 * @example
 *
 *      var madd3 = R.lift((a, b, c) => a + b + c);
 *
 *      madd3([1,2,3], [1,2,3], [1]); //=> [3, 4, 5, 4, 5, 6, 5, 6, 7]
 *
 *      var madd5 = R.lift((a, b, c, d, e) => a + b + c + d + e);
 *
 *      madd5([1,2], [3], [4, 5], [6], [7, 8]); //=> [21, 22, 22, 23, 22, 23, 23, 24]
 */


var lift = /*#__PURE__*/_curry1_1(function lift(fn) {
  return liftN_1(fn.length, fn);
});
var lift_1 = lift;

/**
 * Returns `true` if one or both of its arguments are `true`. Returns `false`
 * if both arguments are `false`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Logic
 * @sig a -> b -> a | b
 * @param {Any} a
 * @param {Any} b
 * @return {Any} the first argument if truthy, otherwise the second argument.
 * @see R.either
 * @example
 *
 *      R.or(true, true); //=> true
 *      R.or(true, false); //=> true
 *      R.or(false, true); //=> true
 *      R.or(false, false); //=> false
 */


var or = /*#__PURE__*/_curry2_1(function or(a, b) {
  return a || b;
});
var or_1 = or;

/**
 * A function wrapping calls to the two functions in an `||` operation,
 * returning the result of the first function if it is truth-y and the result
 * of the second function otherwise. Note that this is short-circuited,
 * meaning that the second function will not be invoked if the first returns a
 * truth-y value.
 *
 * In addition to functions, `R.either` also accepts any fantasy-land compatible
 * applicative functor.
 *
 * @func
 * @memberOf R
 * @since v0.12.0
 * @category Logic
 * @sig (*... -> Boolean) -> (*... -> Boolean) -> (*... -> Boolean)
 * @param {Function} f a predicate
 * @param {Function} g another predicate
 * @return {Function} a function that applies its arguments to `f` and `g` and `||`s their outputs together.
 * @see R.or
 * @example
 *
 *      var gt10 = x => x > 10;
 *      var even = x => x % 2 === 0;
 *      var f = R.either(gt10, even);
 *      f(101); //=> true
 *      f(8); //=> true
 */


var either = /*#__PURE__*/_curry2_1(function either(f, g) {
  return _isFunction_1(f) ? function _either() {
    return f.apply(this, arguments) || g.apply(this, arguments);
  } : lift_1(or_1)(f, g);
});
var either_1 = either;

/**
 * See if an object (`val`) is an instance of the supplied constructor. This
 * function will check up the inheritance chain, if any.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category Type
 * @sig (* -> {*}) -> a -> Boolean
 * @param {Object} ctor A constructor
 * @param {*} val The value to test
 * @return {Boolean}
 * @example
 *
 *      R.is(Object, {}); //=> true
 *      R.is(Number, 1); //=> true
 *      R.is(Object, 1); //=> false
 *      R.is(String, 's'); //=> true
 *      R.is(String, new String('')); //=> true
 *      R.is(Object, new String('')); //=> true
 *      R.is(Object, 's'); //=> false
 *      R.is(Number, {}); //=> false
 */


var is = /*#__PURE__*/_curry2_1(function is(Ctor, val) {
  return val != null && val.constructor === Ctor || val instanceof Ctor;
});
var is_1 = is;

/**
 * A special placeholder value used to specify "gaps" within curried functions,
 * allowing partial application of any combination of arguments, regardless of
 * their positions.
 *
 * If `g` is a curried ternary function and `_` is `R.__`, the following are
 * equivalent:
 *
 *   - `g(1, 2, 3)`
 *   - `g(_, 2, 3)(1)`
 *   - `g(_, _, 3)(1)(2)`
 *   - `g(_, _, 3)(1, 2)`
 *   - `g(_, 2, _)(1, 3)`
 *   - `g(_, 2)(1)(3)`
 *   - `g(_, 2)(1, 3)`
 *   - `g(_, 2)(_, 3)(1)`
 *
 * @constant
 * @memberOf R
 * @since v0.6.0
 * @category Function
 * @example
 *
 *      var greet = R.replace('{name}', R.__, 'Hello, {name}!');
 *      greet('Alice'); //=> 'Hello, Alice!'
 */
var __ = { '@@functional/placeholder': true };

/**
 * Tests the final argument by passing it to the given predicate function. If
 * the predicate is satisfied, the function will return the result of calling
 * the `whenTrueFn` function with the same argument. If the predicate is not
 * satisfied, the argument is returned as is.
 *
 * @func
 * @memberOf R
 * @since v0.18.0
 * @category Logic
 * @sig (a -> Boolean) -> (a -> a) -> a -> a
 * @param {Function} pred       A predicate function
 * @param {Function} whenTrueFn A function to invoke when the `condition`
 *                              evaluates to a truthy value.
 * @param {*}        x          An object to test with the `pred` function and
 *                              pass to `whenTrueFn` if necessary.
 * @return {*} Either `x` or the result of applying `x` to `whenTrueFn`.
 * @see R.ifElse, R.unless
 * @example
 *
 *      // truncate :: String -> String
 *      var truncate = R.when(
 *        R.propSatisfies(R.gt(R.__, 10), 'length'),
 *        R.pipe(R.take(10), R.append('…'), R.join(''))
 *      );
 *      truncate('12345');         //=> '12345'
 *      truncate('0123456789ABC'); //=> '0123456789…'
 */


var when = /*#__PURE__*/_curry3_1(function when(pred, whenTrueFn, x) {
  return pred(x) ? whenTrueFn(x) : x;
});
var when_1 = when;

/**
 * Returns a function that always returns the given value. Note that for
 * non-primitives the value returned is a reference to the original value.
 *
 * This function is known as `const`, `constant`, or `K` (for K combinator) in
 * other languages and libraries.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig a -> (* -> a)
 * @param {*} val The value to wrap in a function
 * @return {Function} A Function :: * -> val.
 * @example
 *
 *      var t = R.always('Tee');
 *      t(); //=> 'Tee'
 */


var always = /*#__PURE__*/_curry1_1(function always(val) {
  return function () {
    return val;
  };
});
var always_1 = always;

/**
 * Tests the final argument by passing it to the given predicate function. If
 * the predicate is not satisfied, the function will return the result of
 * calling the `whenFalseFn` function with the same argument. If the predicate
 * is satisfied, the argument is returned as is.
 *
 * @func
 * @memberOf R
 * @since v0.18.0
 * @category Logic
 * @sig (a -> Boolean) -> (a -> a) -> a -> a
 * @param {Function} pred        A predicate function
 * @param {Function} whenFalseFn A function to invoke when the `pred` evaluates
 *                               to a falsy value.
 * @param {*}        x           An object to test with the `pred` function and
 *                               pass to `whenFalseFn` if necessary.
 * @return {*} Either `x` or the result of applying `x` to `whenFalseFn`.
 * @see R.ifElse, R.when
 * @example
 *
 *      let safeInc = R.unless(R.isNil, R.inc);
 *      safeInc(null); //=> null
 *      safeInc(1); //=> 2
 */


var unless = /*#__PURE__*/_curry3_1(function unless(pred, whenFalseFn, x) {
  return pred(x) ? x : whenFalseFn(x);
});
var unless_1 = unless;

function _identity(x) {
  return x;
}
var _identity_1 = _identity;

/**
 * A function that does nothing but return the parameter supplied to it. Good
 * as a default or placeholder function.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig a -> a
 * @param {*} x The value to return.
 * @return {*} The input value, `x`.
 * @example
 *
 *      R.identity(1); //=> 1
 *
 *      var obj = {};
 *      R.identity(obj) === obj; //=> true
 * @symb R.identity(a) = a
 */


var identity = /*#__PURE__*/_curry1_1(_identity_1);
var identity_1 = identity;

/**
 * Takes a function `fn`, which takes a single array argument, and returns a
 * function which:
 *
 *   - takes any number of positional arguments;
 *   - passes these arguments to `fn` as an array; and
 *   - returns the result.
 *
 * In other words, `R.unapply` derives a variadic function from a function which
 * takes an array. `R.unapply` is the inverse of [`R.apply`](#apply).
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Function
 * @sig ([*...] -> a) -> (*... -> a)
 * @param {Function} fn
 * @return {Function}
 * @see R.apply
 * @example
 *
 *      R.unapply(JSON.stringify)(1, 2, 3); //=> '[1,2,3]'
 * @symb R.unapply(f)(a, b) = f([a, b])
 */


var unapply = /*#__PURE__*/_curry1_1(function unapply(fn) {
  return function () {
    return fn(Array.prototype.slice.call(arguments, 0));
  };
});
var unapply_1 = unapply;

function _pipe(f, g) {
  return function () {
    return g.call(this, f.apply(this, arguments));
  };
}
var _pipe_1 = _pipe;

/**
 * This checks whether a function has a [methodname] function. If it isn't an
 * array it will execute that function otherwise it will default to the ramda
 * implementation.
 *
 * @private
 * @param {Function} fn ramda implemtation
 * @param {String} methodname property to check for a custom implementation
 * @return {Object} Whatever the return value of the method is.
 */


function _checkForMethod(methodname, fn) {
  return function () {
    var length = arguments.length;
    if (length === 0) {
      return fn();
    }
    var obj = arguments[length - 1];
    return _isArray(obj) || typeof obj[methodname] !== 'function' ? fn.apply(this, arguments) : obj[methodname].apply(obj, Array.prototype.slice.call(arguments, 0, length - 1));
  };
}
var _checkForMethod_1 = _checkForMethod;

/**
 * Returns the elements of the given list or string (or object with a `slice`
 * method) from `fromIndex` (inclusive) to `toIndex` (exclusive).
 *
 * Dispatches to the `slice` method of the third argument, if present.
 *
 * @func
 * @memberOf R
 * @since v0.1.4
 * @category List
 * @sig Number -> Number -> [a] -> [a]
 * @sig Number -> Number -> String -> String
 * @param {Number} fromIndex The start index (inclusive).
 * @param {Number} toIndex The end index (exclusive).
 * @param {*} list
 * @return {*}
 * @example
 *
 *      R.slice(1, 3, ['a', 'b', 'c', 'd']);        //=> ['b', 'c']
 *      R.slice(1, Infinity, ['a', 'b', 'c', 'd']); //=> ['b', 'c', 'd']
 *      R.slice(0, -1, ['a', 'b', 'c', 'd']);       //=> ['a', 'b', 'c']
 *      R.slice(-3, -1, ['a', 'b', 'c', 'd']);      //=> ['b', 'c']
 *      R.slice(0, 3, 'ramda');                     //=> 'ram'
 */


var slice = /*#__PURE__*/_curry3_1( /*#__PURE__*/_checkForMethod_1('slice', function slice(fromIndex, toIndex, list) {
  return Array.prototype.slice.call(list, fromIndex, toIndex);
}));
var slice_1 = slice;

/**
 * Returns all but the first element of the given list or string (or object
 * with a `tail` method).
 *
 * Dispatches to the `slice` method of the first argument, if present.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [a]
 * @sig String -> String
 * @param {*} list
 * @return {*}
 * @see R.head, R.init, R.last
 * @example
 *
 *      R.tail([1, 2, 3]);  //=> [2, 3]
 *      R.tail([1, 2]);     //=> [2]
 *      R.tail([1]);        //=> []
 *      R.tail([]);         //=> []
 *
 *      R.tail('abc');  //=> 'bc'
 *      R.tail('ab');   //=> 'b'
 *      R.tail('a');    //=> ''
 *      R.tail('');     //=> ''
 */


var tail = /*#__PURE__*/_curry1_1( /*#__PURE__*/_checkForMethod_1('tail', /*#__PURE__*/slice_1(1, Infinity)));
var tail_1 = tail;

/**
 * Performs left-to-right function composition. The leftmost function may have
 * any arity; the remaining functions must be unary.
 *
 * In some libraries this function is named `sequence`.
 *
 * **Note:** The result of pipe is not automatically curried.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig (((a, b, ..., n) -> o), (o -> p), ..., (x -> y), (y -> z)) -> ((a, b, ..., n) -> z)
 * @param {...Function} functions
 * @return {Function}
 * @see R.compose
 * @example
 *
 *      var f = R.pipe(Math.pow, R.negate, R.inc);
 *
 *      f(3, 4); // -(3^4) + 1
 * @symb R.pipe(f, g, h)(a, b) = h(g(f(a, b)))
 */


function pipe() {
  if (arguments.length === 0) {
    throw new Error('pipe requires at least one argument');
  }
  return _arity_1(arguments[0].length, reduce_1(_pipe_1, arguments[0], tail_1(arguments)));
}
var pipe_1 = pipe;

/**
 * Returns a new list or string with the elements or characters in reverse
 * order.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [a]
 * @sig String -> String
 * @param {Array|String} list
 * @return {Array|String}
 * @example
 *
 *      R.reverse([1, 2, 3]);  //=> [3, 2, 1]
 *      R.reverse([1, 2]);     //=> [2, 1]
 *      R.reverse([1]);        //=> [1]
 *      R.reverse([]);         //=> []
 *
 *      R.reverse('abc');      //=> 'cba'
 *      R.reverse('ab');       //=> 'ba'
 *      R.reverse('a');        //=> 'a'
 *      R.reverse('');         //=> ''
 */


var reverse = /*#__PURE__*/_curry1_1(function reverse(list) {
  return _isString_1(list) ? list.split('').reverse().join('') : Array.prototype.slice.call(list, 0).reverse();
});
var reverse_1 = reverse;

/**
 * Performs right-to-left function composition. The rightmost function may have
 * any arity; the remaining functions must be unary.
 *
 * **Note:** The result of compose is not automatically curried.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig ((y -> z), (x -> y), ..., (o -> p), ((a, b, ..., n) -> o)) -> ((a, b, ..., n) -> z)
 * @param {...Function} ...functions The functions to compose
 * @return {Function}
 * @see R.pipe
 * @example
 *
 *      var classyGreeting = (firstName, lastName) => "The name's " + lastName + ", " + firstName + " " + lastName
 *      var yellGreeting = R.compose(R.toUpper, classyGreeting);
 *      yellGreeting('James', 'Bond'); //=> "THE NAME'S BOND, JAMES BOND"
 *
 *      R.compose(Math.abs, R.add(1), R.multiply(2))(-4) //=> 7
 *
 * @symb R.compose(f, g, h)(a, b) = f(g(h(a, b)))
 */


function compose() {
  if (arguments.length === 0) {
    throw new Error('compose requires at least one argument');
  }
  return pipe_1.apply(this, reverse_1(arguments));
}
var compose_1 = compose;

function unwrapExports (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var _isPlaceholder$2 = function _isPlaceholder(a) {
  return a != null &&
         typeof a === 'object' &&
         a['@@functional/placeholder'] === true;
};

/**
 * Optimized internal one-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */
var _curry1$2 = function _curry1(fn) {
  return function f1(a) {
    if (arguments.length === 0 || _isPlaceholder$2(a)) {
      return f1;
    } else {
      return fn.apply(this, arguments);
    }
  };
};

var _identity$2 = function _identity(x) { return x; };

/**
 * A function that does nothing but return the parameter supplied to it. Good
 * as a default or placeholder function.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig a -> a
 * @param {*} x The value to return.
 * @return {*} The input value, `x`.
 * @example
 *
 *      R.identity(1); //=> 1
 *
 *      var obj = {};
 *      R.identity(obj) === obj; //=> true
 * @symb R.identity(a) = a
 */
var identity$1 = _curry1$2(_identity$2);

/**
 * Optimized internal two-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */
var _curry2$2 = function _curry2(fn) {
  return function f2(a, b) {
    switch (arguments.length) {
      case 0:
        return f2;
      case 1:
        return _isPlaceholder$2(a) ? f2
             : _curry1$2(function(_b) { return fn(a, _b); });
      default:
        return _isPlaceholder$2(a) && _isPlaceholder$2(b) ? f2
             : _isPlaceholder$2(a) ? _curry1$2(function(_a) { return fn(_a, b); })
             : _isPlaceholder$2(b) ? _curry1$2(function(_b) { return fn(a, _b); })
             : fn(a, b);
    }
  };
};

/**
 * Tests whether or not an object is an array.
 *
 * @private
 * @param {*} val The object to test.
 * @return {Boolean} `true` if `val` is an array, `false` otherwise.
 * @example
 *
 *      _isArray([]); //=> true
 *      _isArray(null); //=> false
 *      _isArray({}); //=> false
 */
var _isArray$3 = Array.isArray || function _isArray(val) {
  return (val != null &&
          val.length >= 0 &&
          Object.prototype.toString.call(val) === '[object Array]');
};

var _isTransformer$2 = function _isTransformer(obj) {
  return typeof obj['@@transducer/step'] === 'function';
};

/**
 * Returns a function that dispatches with different strategies based on the
 * object in list position (last argument). If it is an array, executes [fn].
 * Otherwise, if it has a function with one of the given method names, it will
 * execute that function (functor case). Otherwise, if it is a transformer,
 * uses transducer [xf] to return a new transformer (transducer case).
 * Otherwise, it will default to executing [fn].
 *
 * @private
 * @param {Array} methodNames properties to check for a custom implementation
 * @param {Function} xf transducer to initialize if object is transformer
 * @param {Function} fn default ramda implementation
 * @return {Function} A function that dispatches on object in list position
 */
var _dispatchable$2 = function _dispatchable(methodNames, xf, fn) {
  return function() {
    if (arguments.length === 0) {
      return fn();
    }
    var args = Array.prototype.slice.call(arguments, 0);
    var obj = args.pop();
    if (!_isArray$3(obj)) {
      var idx = 0;
      while (idx < methodNames.length) {
        if (typeof obj[methodNames[idx]] === 'function') {
          return obj[methodNames[idx]].apply(obj, args);
        }
        idx += 1;
      }
      if (_isTransformer$2(obj)) {
        var transducer = xf.apply(null, args);
        return transducer(obj);
      }
    }
    return fn.apply(this, arguments);
  };
};

var _map$2 = function _map(fn, functor) {
  var idx = 0;
  var len = functor.length;
  var result = Array(len);
  while (idx < len) {
    result[idx] = fn(functor[idx]);
    idx += 1;
  }
  return result;
};

var _isString$2 = function _isString(x) {
  return Object.prototype.toString.call(x) === '[object String]';
};

/**
 * Tests whether or not an object is similar to an array.
 *
 * @private
 * @category Type
 * @category List
 * @sig * -> Boolean
 * @param {*} x The object to test.
 * @return {Boolean} `true` if `x` has a numeric length property and extreme indices defined; `false` otherwise.
 * @example
 *
 *      _isArrayLike([]); //=> true
 *      _isArrayLike(true); //=> false
 *      _isArrayLike({}); //=> false
 *      _isArrayLike({length: 10}); //=> false
 *      _isArrayLike({0: 'zero', 9: 'nine', length: 10}); //=> true
 */
var _isArrayLike$2 = _curry1$2(function isArrayLike(x) {
  if (_isArray$3(x)) { return true; }
  if (!x) { return false; }
  if (typeof x !== 'object') { return false; }
  if (_isString$2(x)) { return false; }
  if (x.nodeType === 1) { return !!x.length; }
  if (x.length === 0) { return true; }
  if (x.length > 0) {
    return x.hasOwnProperty(0) && x.hasOwnProperty(x.length - 1);
  }
  return false;
});

var _xwrap$2 = (function() {
  function XWrap(fn) {
    this.f = fn;
  }
  XWrap.prototype['@@transducer/init'] = function() {
    throw new Error('init not implemented on XWrap');
  };
  XWrap.prototype['@@transducer/result'] = function(acc) { return acc; };
  XWrap.prototype['@@transducer/step'] = function(acc, x) {
    return this.f(acc, x);
  };

  return function _xwrap(fn) { return new XWrap(fn); };
}());

var _arity$2 = function _arity(n, fn) {
  /* eslint-disable no-unused-vars */
  switch (n) {
    case 0: return function() { return fn.apply(this, arguments); };
    case 1: return function(a0) { return fn.apply(this, arguments); };
    case 2: return function(a0, a1) { return fn.apply(this, arguments); };
    case 3: return function(a0, a1, a2) { return fn.apply(this, arguments); };
    case 4: return function(a0, a1, a2, a3) { return fn.apply(this, arguments); };
    case 5: return function(a0, a1, a2, a3, a4) { return fn.apply(this, arguments); };
    case 6: return function(a0, a1, a2, a3, a4, a5) { return fn.apply(this, arguments); };
    case 7: return function(a0, a1, a2, a3, a4, a5, a6) { return fn.apply(this, arguments); };
    case 8: return function(a0, a1, a2, a3, a4, a5, a6, a7) { return fn.apply(this, arguments); };
    case 9: return function(a0, a1, a2, a3, a4, a5, a6, a7, a8) { return fn.apply(this, arguments); };
    case 10: return function(a0, a1, a2, a3, a4, a5, a6, a7, a8, a9) { return fn.apply(this, arguments); };
    default: throw new Error('First argument to _arity must be a non-negative integer no greater than ten');
  }
};

/**
 * Creates a function that is bound to a context.
 * Note: `R.bind` does not provide the additional argument-binding capabilities of
 * [Function.prototype.bind](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Function/bind).
 *
 * @func
 * @memberOf R
 * @since v0.6.0
 * @category Function
 * @category Object
 * @sig (* -> *) -> {*} -> (* -> *)
 * @param {Function} fn The function to bind to context
 * @param {Object} thisObj The context to bind `fn` to
 * @return {Function} A function that will execute in the context of `thisObj`.
 * @see R.partial
 * @example
 *
 *      var log = R.bind(console.log, console);
 *      R.pipe(R.assoc('a', 2), R.tap(log), R.assoc('a', 3))({a: 1}); //=> {a: 3}
 *      // logs {a: 2}
 * @symb R.bind(f, o)(a, b) = f.call(o, a, b)
 */
var bind$2 = _curry2$2(function bind(fn, thisObj) {
  return _arity$2(fn.length, function() {
    return fn.apply(thisObj, arguments);
  });
});

var _reduce$2 = (function() {
  function _arrayReduce(xf, acc, list) {
    var idx = 0;
    var len = list.length;
    while (idx < len) {
      acc = xf['@@transducer/step'](acc, list[idx]);
      if (acc && acc['@@transducer/reduced']) {
        acc = acc['@@transducer/value'];
        break;
      }
      idx += 1;
    }
    return xf['@@transducer/result'](acc);
  }

  function _iterableReduce(xf, acc, iter) {
    var step = iter.next();
    while (!step.done) {
      acc = xf['@@transducer/step'](acc, step.value);
      if (acc && acc['@@transducer/reduced']) {
        acc = acc['@@transducer/value'];
        break;
      }
      step = iter.next();
    }
    return xf['@@transducer/result'](acc);
  }

  function _methodReduce(xf, acc, obj, methodName) {
    return xf['@@transducer/result'](obj[methodName](bind$2(xf['@@transducer/step'], xf), acc));
  }

  var symIterator = (typeof Symbol !== 'undefined') ? Symbol.iterator : '@@iterator';
  return function _reduce(fn, acc, list) {
    if (typeof fn === 'function') {
      fn = _xwrap$2(fn);
    }
    if (_isArrayLike$2(list)) {
      return _arrayReduce(fn, acc, list);
    }
    if (typeof list['fantasy-land/reduce'] === 'function') {
      return _methodReduce(fn, acc, list, 'fantasy-land/reduce');
    }
    if (list[symIterator] != null) {
      return _iterableReduce(fn, acc, list[symIterator]());
    }
    if (typeof list.next === 'function') {
      return _iterableReduce(fn, acc, list);
    }
    if (typeof list.reduce === 'function') {
      return _methodReduce(fn, acc, list, 'reduce');
    }

    throw new TypeError('reduce: list must be array or iterable');
  };
}());

var _xfBase$3 = {
  init: function() {
    return this.xf['@@transducer/init']();
  },
  result: function(result) {
    return this.xf['@@transducer/result'](result);
  }
};

var _xmap$2 = (function() {
  function XMap(f, xf) {
    this.xf = xf;
    this.f = f;
  }
  XMap.prototype['@@transducer/init'] = _xfBase$3.init;
  XMap.prototype['@@transducer/result'] = _xfBase$3.result;
  XMap.prototype['@@transducer/step'] = function(result, input) {
    return this.xf['@@transducer/step'](result, this.f(input));
  };

  return _curry2$2(function _xmap(f, xf) { return new XMap(f, xf); });
}());

/**
 * Internal curryN function.
 *
 * @private
 * @category Function
 * @param {Number} length The arity of the curried function.
 * @param {Array} received An array of arguments received thus far.
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */
var _curryN$2 = function _curryN(length, received, fn) {
  return function() {
    var combined = [];
    var argsIdx = 0;
    var left = length;
    var combinedIdx = 0;
    while (combinedIdx < received.length || argsIdx < arguments.length) {
      var result;
      if (combinedIdx < received.length &&
          (!_isPlaceholder$2(received[combinedIdx]) ||
           argsIdx >= arguments.length)) {
        result = received[combinedIdx];
      } else {
        result = arguments[argsIdx];
        argsIdx += 1;
      }
      combined[combinedIdx] = result;
      if (!_isPlaceholder$2(result)) {
        left -= 1;
      }
      combinedIdx += 1;
    }
    return left <= 0 ? fn.apply(this, combined)
                     : _arity$2(left, _curryN(length, combined, fn));
  };
};

/**
 * Returns a curried equivalent of the provided function, with the specified
 * arity. The curried function has two unusual capabilities. First, its
 * arguments needn't be provided one at a time. If `g` is `R.curryN(3, f)`, the
 * following are equivalent:
 *
 *   - `g(1)(2)(3)`
 *   - `g(1)(2, 3)`
 *   - `g(1, 2)(3)`
 *   - `g(1, 2, 3)`
 *
 * Secondly, the special placeholder value [`R.__`](#__) may be used to specify
 * "gaps", allowing partial application of any combination of arguments,
 * regardless of their positions. If `g` is as above and `_` is [`R.__`](#__),
 * the following are equivalent:
 *
 *   - `g(1, 2, 3)`
 *   - `g(_, 2, 3)(1)`
 *   - `g(_, _, 3)(1)(2)`
 *   - `g(_, _, 3)(1, 2)`
 *   - `g(_, 2)(1)(3)`
 *   - `g(_, 2)(1, 3)`
 *   - `g(_, 2)(_, 3)(1)`
 *
 * @func
 * @memberOf R
 * @since v0.5.0
 * @category Function
 * @sig Number -> (* -> a) -> (* -> a)
 * @param {Number} length The arity for the returned function.
 * @param {Function} fn The function to curry.
 * @return {Function} A new, curried function.
 * @see R.curry
 * @example
 *
 *      var sumArgs = (...args) => R.sum(args);
 *
 *      var curriedAddFourNumbers = R.curryN(4, sumArgs);
 *      var f = curriedAddFourNumbers(1, 2);
 *      var g = f(3);
 *      g(4); //=> 10
 */
var curryN$2 = _curry2$2(function curryN(length, fn) {
  if (length === 1) {
    return _curry1$2(fn);
  }
  return _arity$2(length, _curryN$2(length, [], fn));
});

var _has$2 = function _has(prop, obj) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
};

var _isArguments$2 = (function() {
  var toString = Object.prototype.toString;
  return toString.call(arguments) === '[object Arguments]' ?
    function _isArguments(x) { return toString.call(x) === '[object Arguments]'; } :
    function _isArguments(x) { return _has$2('callee', x); };
}());

/**
 * Returns a list containing the names of all the enumerable own properties of
 * the supplied object.
 * Note that the order of the output array is not guaranteed to be consistent
 * across different JS platforms.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig {k: v} -> [k]
 * @param {Object} obj The object to extract properties from
 * @return {Array} An array of the object's own properties.
 * @see R.keysIn, R.values
 * @example
 *
 *      R.keys({a: 1, b: 2, c: 3}); //=> ['a', 'b', 'c']
 */
var keys$2 = (function() {
  // cover IE < 9 keys issues
  var hasEnumBug = !({toString: null}).propertyIsEnumerable('toString');
  var nonEnumerableProps = ['constructor', 'valueOf', 'isPrototypeOf', 'toString',
                            'propertyIsEnumerable', 'hasOwnProperty', 'toLocaleString'];
  // Safari bug
  var hasArgsEnumBug = (function() {
    'use strict';
    return arguments.propertyIsEnumerable('length');
  }());

  var contains = function contains(list, item) {
    var idx = 0;
    while (idx < list.length) {
      if (list[idx] === item) {
        return true;
      }
      idx += 1;
    }
    return false;
  };

  return typeof Object.keys === 'function' && !hasArgsEnumBug ?
    _curry1$2(function keys(obj) {
      return Object(obj) !== obj ? [] : Object.keys(obj);
    }) :
    _curry1$2(function keys(obj) {
      if (Object(obj) !== obj) {
        return [];
      }
      var prop, nIdx;
      var ks = [];
      var checkArgsLength = hasArgsEnumBug && _isArguments$2(obj);
      for (prop in obj) {
        if (_has$2(prop, obj) && (!checkArgsLength || prop !== 'length')) {
          ks[ks.length] = prop;
        }
      }
      if (hasEnumBug) {
        nIdx = nonEnumerableProps.length - 1;
        while (nIdx >= 0) {
          prop = nonEnumerableProps[nIdx];
          if (_has$2(prop, obj) && !contains(ks, prop)) {
            ks[ks.length] = prop;
          }
          nIdx -= 1;
        }
      }
      return ks;
    });
}());

/**
 * Takes a function and
 * a [functor](https://github.com/fantasyland/fantasy-land#functor),
 * applies the function to each of the functor's values, and returns
 * a functor of the same shape.
 *
 * Ramda provides suitable `map` implementations for `Array` and `Object`,
 * so this function may be applied to `[1, 2, 3]` or `{x: 1, y: 2, z: 3}`.
 *
 * Dispatches to the `map` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * Also treats functions as functors and will compose them together.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Functor f => (a -> b) -> f a -> f b
 * @param {Function} fn The function to be called on every element of the input `list`.
 * @param {Array} list The list to be iterated over.
 * @return {Array} The new list.
 * @see R.transduce, R.addIndex
 * @example
 *
 *      var double = x => x * 2;
 *
 *      R.map(double, [1, 2, 3]); //=> [2, 4, 6]
 *
 *      R.map(double, {x: 1, y: 2, z: 3}); //=> {x: 2, y: 4, z: 6}
 * @symb R.map(f, [a, b]) = [f(a), f(b)]
 * @symb R.map(f, { x: a, y: b }) = { x: f(a), y: f(b) }
 * @symb R.map(f, functor_o) = functor_o.map(f)
 */
var map$2 = _curry2$2(_dispatchable$2(['fantasy-land/map', 'map'], _xmap$2, function map(fn, functor) {
  switch (Object.prototype.toString.call(functor)) {
    case '[object Function]':
      return curryN$2(functor.length, function() {
        return fn.call(this, functor.apply(this, arguments));
      });
    case '[object Object]':
      return _reduce$2(function(acc, key) {
        acc[key] = fn(functor[key]);
        return acc;
      }, {}, keys$2(functor));
    default:
      return _map$2(fn, functor);
  }
}));

/**
 * Optimized internal three-arity curry function.
 *
 * @private
 * @category Function
 * @param {Function} fn The function to curry.
 * @return {Function} The curried function.
 */
var _curry3$2 = function _curry3(fn) {
  return function f3(a, b, c) {
    switch (arguments.length) {
      case 0:
        return f3;
      case 1:
        return _isPlaceholder$2(a) ? f3
             : _curry2$2(function(_b, _c) { return fn(a, _b, _c); });
      case 2:
        return _isPlaceholder$2(a) && _isPlaceholder$2(b) ? f3
             : _isPlaceholder$2(a) ? _curry2$2(function(_a, _c) { return fn(_a, b, _c); })
             : _isPlaceholder$2(b) ? _curry2$2(function(_b, _c) { return fn(a, _b, _c); })
             : _curry1$2(function(_c) { return fn(a, b, _c); });
      default:
        return _isPlaceholder$2(a) && _isPlaceholder$2(b) && _isPlaceholder$2(c) ? f3
             : _isPlaceholder$2(a) && _isPlaceholder$2(b) ? _curry2$2(function(_a, _b) { return fn(_a, _b, c); })
             : _isPlaceholder$2(a) && _isPlaceholder$2(c) ? _curry2$2(function(_a, _c) { return fn(_a, b, _c); })
             : _isPlaceholder$2(b) && _isPlaceholder$2(c) ? _curry2$2(function(_b, _c) { return fn(a, _b, _c); })
             : _isPlaceholder$2(a) ? _curry1$2(function(_a) { return fn(_a, b, c); })
             : _isPlaceholder$2(b) ? _curry1$2(function(_b) { return fn(a, _b, c); })
             : _isPlaceholder$2(c) ? _curry1$2(function(_c) { return fn(a, b, _c); })
             : fn(a, b, c);
    }
  };
};

/**
 * Returns a single item by iterating through the list, successively calling
 * the iterator function and passing it an accumulator value and the current
 * value from the array, and then passing the result to the next call.
 *
 * The iterator function receives two values: *(acc, value)*. It may use
 * [`R.reduced`](#reduced) to shortcut the iteration.
 *
 * The arguments' order of [`reduceRight`](#reduceRight)'s iterator function
 * is *(value, acc)*.
 *
 * Note: `R.reduce` does not skip deleted or unassigned indices (sparse
 * arrays), unlike the native `Array.prototype.reduce` method. For more details
 * on this behavior, see:
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/reduce#Description
 *
 * Dispatches to the `reduce` method of the third argument, if present. When
 * doing so, it is up to the user to handle the [`R.reduced`](#reduced)
 * shortcuting, as this is not implemented by `reduce`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig ((a, b) -> a) -> a -> [b] -> a
 * @param {Function} fn The iterator function. Receives two values, the accumulator and the
 *        current element from the array.
 * @param {*} acc The accumulator value.
 * @param {Array} list The list to iterate over.
 * @return {*} The final, accumulated value.
 * @see R.reduced, R.addIndex, R.reduceRight
 * @example
 *
 *      R.reduce(R.subtract, 0, [1, 2, 3, 4]) // => ((((0 - 1) - 2) - 3) - 4) = -10
 *                -               -10
 *               / \              / \
 *              -   4           -6   4
 *             / \              / \
 *            -   3   ==>     -3   3
 *           / \              / \
 *          -   2           -1   2
 *         / \              / \
 *        0   1            0   1
 *
 * @symb R.reduce(f, a, [b, c, d]) = f(f(f(a, b), c), d)
 */
var reduce$2 = _curry3$2(_reduce$2);

/**
 * Returns a curried equivalent of the provided function. The curried function
 * has two unusual capabilities. First, its arguments needn't be provided one
 * at a time. If `f` is a ternary function and `g` is `R.curry(f)`, the
 * following are equivalent:
 *
 *   - `g(1)(2)(3)`
 *   - `g(1)(2, 3)`
 *   - `g(1, 2)(3)`
 *   - `g(1, 2, 3)`
 *
 * Secondly, the special placeholder value [`R.__`](#__) may be used to specify
 * "gaps", allowing partial application of any combination of arguments,
 * regardless of their positions. If `g` is as above and `_` is [`R.__`](#__),
 * the following are equivalent:
 *
 *   - `g(1, 2, 3)`
 *   - `g(_, 2, 3)(1)`
 *   - `g(_, _, 3)(1)(2)`
 *   - `g(_, _, 3)(1, 2)`
 *   - `g(_, 2)(1)(3)`
 *   - `g(_, 2)(1, 3)`
 *   - `g(_, 2)(_, 3)(1)`
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig (* -> a) -> (* -> a)
 * @param {Function} fn The function to curry.
 * @return {Function} A new, curried function.
 * @see R.curryN
 * @example
 *
 *      var addFourNumbers = (a, b, c, d) => a + b + c + d;
 *
 *      var curriedAddFourNumbers = R.curry(addFourNumbers);
 *      var f = curriedAddFourNumbers(1, 2);
 *      var g = f(3);
 *      g(4); //=> 10
 */
var curry$2 = _curry1$2(function curry(fn) {
  return curryN$2(fn.length, fn);
});

var _filter$2 = function _filter(fn, list) {
  var idx = 0;
  var len = list.length;
  var result = [];

  while (idx < len) {
    if (fn(list[idx])) {
      result[result.length] = list[idx];
    }
    idx += 1;
  }
  return result;
};

var _isObject$2 = function _isObject(x) {
  return Object.prototype.toString.call(x) === '[object Object]';
};

var _xfilter$2 = (function() {
  function XFilter(f, xf) {
    this.xf = xf;
    this.f = f;
  }
  XFilter.prototype['@@transducer/init'] = _xfBase$3.init;
  XFilter.prototype['@@transducer/result'] = _xfBase$3.result;
  XFilter.prototype['@@transducer/step'] = function(result, input) {
    return this.f(input) ? this.xf['@@transducer/step'](result, input) : result;
  };

  return _curry2$2(function _xfilter(f, xf) { return new XFilter(f, xf); });
}());

/**
 * Takes a predicate and a `Filterable`, and returns a new filterable of the
 * same type containing the members of the given filterable which satisfy the
 * given predicate. Filterable objects include plain objects or any object
 * that has a filter method such as `Array`.
 *
 * Dispatches to the `filter` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig Filterable f => (a -> Boolean) -> f a -> f a
 * @param {Function} pred
 * @param {Array} filterable
 * @return {Array} Filterable
 * @see R.reject, R.transduce, R.addIndex
 * @example
 *
 *      var isEven = n => n % 2 === 0;
 *
 *      R.filter(isEven, [1, 2, 3, 4]); //=> [2, 4]
 *
 *      R.filter(isEven, {a: 1, b: 2, c: 3, d: 4}); //=> {b: 2, d: 4}
 */
var filter$2 = _curry2$2(_dispatchable$2(['filter'], _xfilter$2, function(pred, filterable) {
  return (
    _isObject$2(filterable) ?
      _reduce$2(function(acc, key) {
        if (pred(filterable[key])) {
          acc[key] = filterable[key];
        }
        return acc;
      }, {}, keys$2(filterable)) :
    // else
      _filter$2(pred, filterable)
  );
}));

var _arrayFromIterator$2 = function _arrayFromIterator(iter) {
  var list = [];
  var next;
  while (!(next = iter.next()).done) {
    list.push(next.value);
  }
  return list;
};

var _functionName$2 = function _functionName(f) {
  // String(x => x) evaluates to "x => x", so the pattern may not match.
  var match = String(f).match(/^function (\w*)/);
  return match == null ? '' : match[1];
};

/**
 * Returns true if its arguments are identical, false otherwise. Values are
 * identical if they reference the same memory. `NaN` is identical to `NaN`;
 * `0` and `-0` are not identical.
 *
 * @func
 * @memberOf R
 * @since v0.15.0
 * @category Relation
 * @sig a -> a -> Boolean
 * @param {*} a
 * @param {*} b
 * @return {Boolean}
 * @example
 *
 *      var o = {};
 *      R.identical(o, o); //=> true
 *      R.identical(1, 1); //=> true
 *      R.identical(1, '1'); //=> false
 *      R.identical([], []); //=> false
 *      R.identical(0, -0); //=> false
 *      R.identical(NaN, NaN); //=> true
 */
var identical$2 = _curry2$2(function identical(a, b) {
  // SameValue algorithm
  if (a === b) { // Steps 1-5, 7-10
    // Steps 6.b-6.e: +0 != -0
    return a !== 0 || 1 / a === 1 / b;
  } else {
    // Step 6.a: NaN == NaN
    return a !== a && b !== b;
  }
});

/**
 * Gives a single-word string description of the (native) type of a value,
 * returning such answers as 'Object', 'Number', 'Array', or 'Null'. Does not
 * attempt to distinguish user Object types any further, reporting them all as
 * 'Object'.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Type
 * @sig (* -> {*}) -> String
 * @param {*} val The value to test
 * @return {String}
 * @example
 *
 *      R.type({}); //=> "Object"
 *      R.type(1); //=> "Number"
 *      R.type(false); //=> "Boolean"
 *      R.type('s'); //=> "String"
 *      R.type(null); //=> "Null"
 *      R.type([]); //=> "Array"
 *      R.type(/[A-z]/); //=> "RegExp"
 *      R.type(() => {}); //=> "Function"
 */
var type$2 = _curry1$2(function type(val) {
  return val === null      ? 'Null'      :
         val === undefined ? 'Undefined' :
         Object.prototype.toString.call(val).slice(8, -1);
});

var _equals$2 = function _equals(a, b, stackA, stackB) {
  if (identical$2(a, b)) {
    return true;
  }

  if (type$2(a) !== type$2(b)) {
    return false;
  }

  if (a == null || b == null) {
    return false;
  }

  if (typeof a['fantasy-land/equals'] === 'function' || typeof b['fantasy-land/equals'] === 'function') {
    return typeof a['fantasy-land/equals'] === 'function' && a['fantasy-land/equals'](b) &&
           typeof b['fantasy-land/equals'] === 'function' && b['fantasy-land/equals'](a);
  }

  if (typeof a.equals === 'function' || typeof b.equals === 'function') {
    return typeof a.equals === 'function' && a.equals(b) &&
           typeof b.equals === 'function' && b.equals(a);
  }

  switch (type$2(a)) {
    case 'Arguments':
    case 'Array':
    case 'Object':
      if (typeof a.constructor === 'function' &&
          _functionName$2(a.constructor) === 'Promise') {
        return a === b;
      }
      break;
    case 'Boolean':
    case 'Number':
    case 'String':
      if (!(typeof a === typeof b && identical$2(a.valueOf(), b.valueOf()))) {
        return false;
      }
      break;
    case 'Date':
      if (!identical$2(a.valueOf(), b.valueOf())) {
        return false;
      }
      break;
    case 'Error':
      return a.name === b.name && a.message === b.message;
    case 'RegExp':
      if (!(a.source === b.source &&
            a.global === b.global &&
            a.ignoreCase === b.ignoreCase &&
            a.multiline === b.multiline &&
            a.sticky === b.sticky &&
            a.unicode === b.unicode)) {
        return false;
      }
      break;
    case 'Map':
    case 'Set':
      if (!_equals(_arrayFromIterator$2(a.entries()), _arrayFromIterator$2(b.entries()), stackA, stackB)) {
        return false;
      }
      break;
    case 'Int8Array':
    case 'Uint8Array':
    case 'Uint8ClampedArray':
    case 'Int16Array':
    case 'Uint16Array':
    case 'Int32Array':
    case 'Uint32Array':
    case 'Float32Array':
    case 'Float64Array':
      break;
    case 'ArrayBuffer':
      break;
    default:
      // Values of other types are only equal if identical.
      return false;
  }

  var keysA = keys$2(a);
  if (keysA.length !== keys$2(b).length) {
    return false;
  }

  var idx = stackA.length - 1;
  while (idx >= 0) {
    if (stackA[idx] === a) {
      return stackB[idx] === b;
    }
    idx -= 1;
  }

  stackA.push(a);
  stackB.push(b);
  idx = keysA.length - 1;
  while (idx >= 0) {
    var key = keysA[idx];
    if (!(_has$2(key, b) && _equals(b[key], a[key], stackA, stackB))) {
      return false;
    }
    idx -= 1;
  }
  stackA.pop();
  stackB.pop();
  return true;
};

/**
 * Returns `true` if its arguments are equivalent, `false` otherwise. Handles
 * cyclical data structures.
 *
 * Dispatches symmetrically to the `equals` methods of both arguments, if
 * present.
 *
 * @func
 * @memberOf R
 * @since v0.15.0
 * @category Relation
 * @sig a -> b -> Boolean
 * @param {*} a
 * @param {*} b
 * @return {Boolean}
 * @example
 *
 *      R.equals(1, 1); //=> true
 *      R.equals(1, '1'); //=> false
 *      R.equals([1, 2, 3], [1, 2, 3]); //=> true
 *
 *      var a = {}; a.v = a;
 *      var b = {}; b.v = b;
 *      R.equals(a, b); //=> true
 */
var equals$2 = _curry2$2(function equals(a, b) {
  return _equals$2(a, b, [], []);
});

var _reduced = function _reduced(x) {
  return x && x['@@transducer/reduced'] ? x :
    {
      '@@transducer/value': x,
      '@@transducer/reduced': true
    };
};

var _xall = (function() {
  function XAll(f, xf) {
    this.xf = xf;
    this.f = f;
    this.all = true;
  }
  XAll.prototype['@@transducer/init'] = _xfBase$3.init;
  XAll.prototype['@@transducer/result'] = function(result) {
    if (this.all) {
      result = this.xf['@@transducer/step'](result, true);
    }
    return this.xf['@@transducer/result'](result);
  };
  XAll.prototype['@@transducer/step'] = function(result, input) {
    if (!this.f(input)) {
      this.all = false;
      result = _reduced(this.xf['@@transducer/step'](result, false));
    }
    return result;
  };

  return _curry2$2(function _xall(f, xf) { return new XAll(f, xf); });
}());

/**
 * Returns `true` if all elements of the list match the predicate, `false` if
 * there are any that don't.
 *
 * Dispatches to the `all` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig (a -> Boolean) -> [a] -> Boolean
 * @param {Function} fn The predicate function.
 * @param {Array} list The array to consider.
 * @return {Boolean} `true` if the predicate is satisfied by every element, `false`
 *         otherwise.
 * @see R.any, R.none, R.transduce
 * @example
 *
 *      var equals3 = R.equals(3);
 *      R.all(equals3)([3, 3, 3, 3]); //=> true
 *      R.all(equals3)([3, 3, 1, 3]); //=> false
 */
var all = _curry2$2(_dispatchable$2(['all'], _xall, function all(fn, list) {
  var idx = 0;
  while (idx < list.length) {
    if (!fn(list[idx])) {
      return false;
    }
    idx += 1;
  }
  return true;
}));

var lib = createCommonjsModule(function (module, exports) {
'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.validate = undefined;



var _identity2 = _interopRequireDefault(identity$1);



var _map2 = _interopRequireDefault(map$2);



var _reduce2 = _interopRequireDefault(reduce$2);



var _curry2 = _interopRequireDefault(curry$2);



var _filter2 = _interopRequireDefault(filter$2);



var _equals2 = _interopRequireDefault(equals$2);



var _all2 = _interopRequireDefault(all);

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 *
 * @param {Function} successFn callback function in case of valid input
 * @param {Function} failFn callback function in case of invalid input
 * @param {Array} input
 * @returns {*}
 */
var transform = function transform(successFn, failFn, input) {
  var valid = (0, _all2.default)((0, _equals2.default)(true), input);
  return valid ? successFn() : failFn((0, _filter2.default)(function (a) {
    return a !== true;
  }, input));
};

/**
 *
 * @param {Function} predicate validation function to apply inputs on
 * @param {String|Function} errorMsg error message to return in case of fail
 * @param {*} value the actual value
 * @param {Object} inputs the input object - in case the predicate function needs access to dependent values
 * @returns {Boolean}
 */

var runPredicate = function runPredicate(_ref, value, inputs, field) {
  var _ref2 = _slicedToArray(_ref, 2),
      predicate = _ref2[0],
      errorMsg = _ref2[1];

  return predicate(value, inputs // eslint-disable-line no-nested-ternary
  ) ? true : typeof errorMsg === 'function' ? errorMsg(value, field) : errorMsg;
};

/**
 *
 * @param {Function} successFn callback function in case of valid input
 * @param {Function} failFn callback function in case of invalid input
 * @param {Object} spec the rule object
 * @param {Object} input the validation input data
 * @returns {{}}
 */
var validate = exports.validate = (0, _curry2.default)(function (successFn, failFn, spec, input) {
  return (0, _reduce2.default)(function (result, key) {
    var value = input[key];
    var predicates = spec[key];
    if (Array.isArray(predicates)) {
      return _extends({}, result, _defineProperty({}, key, transform(function () {
        return successFn(value);
      }, failFn, (0, _map2.default)(function (f) {
        return runPredicate(f, value, input, key);
      }, predicates))));
    } else if ((typeof predicates === 'undefined' ? 'undefined' : _typeof(predicates)) === 'object') {
      return _extends({}, result, _defineProperty({}, key, validate(successFn, failFn, predicates, value)));
    } else if (typeof predicates === 'function') {
      var rules = predicates(value);
      return _extends({}, result, _defineProperty({}, key, validate(successFn, failFn, rules, value)));
    } else {
      return _extends({}, result, _defineProperty({}, key, successFn([])));
    }
  }, {}, Object.keys(input));
}

/**
 *
 * @param {Object} spec the rule object
 * @param {Object} input the validation input data
 * @returns {{}}
 */
);var spected = function spected(spec, input) {
  return validate(function () {
    return true;
  }, _identity2.default, spec, input);
};

exports.default = (0, _curry2.default)(spected);
});

var spected = unwrapExports(lib);

function _cloneRegExp(pattern) {
                                  return new RegExp(pattern.source, (pattern.global ? 'g' : '') + (pattern.ignoreCase ? 'i' : '') + (pattern.multiline ? 'm' : '') + (pattern.sticky ? 'y' : '') + (pattern.unicode ? 'u' : ''));
}
var _cloneRegExp_1 = _cloneRegExp;

function _isRegExp(x) {
  return Object.prototype.toString.call(x) === '[object RegExp]';
}
var _isRegExp_1 = _isRegExp;

/**
 * Determines whether a given string matches a given regular expression.
 *
 * @func
 * @memberOf R
 * @since v0.12.0
 * @category String
 * @sig RegExp -> String -> Boolean
 * @param {RegExp} pattern
 * @param {String} str
 * @return {Boolean}
 * @see R.match
 * @example
 *
 *      R.test(/^x/, 'xyz'); //=> true
 *      R.test(/^y/, 'xyz'); //=> false
 */


var test = /*#__PURE__*/_curry2_1(function test(pattern, str) {
  if (!_isRegExp_1(pattern)) {
    throw new TypeError('‘test’ requires a value of type RegExp as its first argument; received ' + toString_1(pattern));
  }
  return _cloneRegExp_1(pattern).test(str);
});
var test_1 = test;

/**
 * Returns `true` if both arguments are `true`; `false` otherwise.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Logic
 * @sig a -> b -> a | b
 * @param {Any} a
 * @param {Any} b
 * @return {Any} the first argument if it is falsy, otherwise the second argument.
 * @see R.both
 * @example
 *
 *      R.and(true, true); //=> true
 *      R.and(true, false); //=> false
 *      R.and(false, true); //=> false
 *      R.and(false, false); //=> false
 */


var and = /*#__PURE__*/_curry2_1(function and(a, b) {
  return a && b;
});
var and_1 = and;

/**
 * A function which calls the two provided functions and returns the `&&`
 * of the results.
 * It returns the result of the first function if it is false-y and the result
 * of the second function otherwise. Note that this is short-circuited,
 * meaning that the second function will not be invoked if the first returns a
 * false-y value.
 *
 * In addition to functions, `R.both` also accepts any fantasy-land compatible
 * applicative functor.
 *
 * @func
 * @memberOf R
 * @since v0.12.0
 * @category Logic
 * @sig (*... -> Boolean) -> (*... -> Boolean) -> (*... -> Boolean)
 * @param {Function} f A predicate
 * @param {Function} g Another predicate
 * @return {Function} a function that applies its arguments to `f` and `g` and `&&`s their outputs together.
 * @see R.and
 * @example
 *
 *      var gt10 = R.gt(R.__, 10)
 *      var lt20 = R.lt(R.__, 20)
 *      var f = R.both(gt10, lt20);
 *      f(15); //=> true
 *      f(30); //=> false
 */


var both = /*#__PURE__*/_curry2_1(function both(f, g) {
  return _isFunction_1(f) ? function _both() {
    return f.apply(this, arguments) && g.apply(this, arguments);
  } : lift_1(and_1)(f, g);
});
var both_1 = both;

/**
 * Returns a partial copy of an object omitting the keys specified.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig [String] -> {String: *} -> {String: *}
 * @param {Array} names an array of String property names to omit from the new object
 * @param {Object} obj The object to copy from
 * @return {Object} A new object with properties from `names` not on it.
 * @see R.pick
 * @example
 *
 *      R.omit(['a', 'd'], {a: 1, b: 2, c: 3, d: 4}); //=> {b: 2, c: 3}
 */


var omit = /*#__PURE__*/_curry2_1(function omit(names, obj) {
  var result = {};
  var index = {};
  var idx = 0;
  var len = names.length;

  while (idx < len) {
    index[names[idx]] = 1;
    idx += 1;
  }

  for (var prop in obj) {
    if (!index.hasOwnProperty(prop)) {
      result[prop] = obj[prop];
    }
  }
  return result;
});
var omit_1 = omit;

/**
 * `_makeFlat` is a helper function that returns a one-level or fully recursive
 * function based on the flag passed in.
 *
 * @private
 */


function _makeFlat(recursive) {
  return function flatt(list) {
    var value, jlen, j;
    var result = [];
    var idx = 0;
    var ilen = list.length;

    while (idx < ilen) {
      if (_isArrayLike_1(list[idx])) {
        value = recursive ? flatt(list[idx]) : list[idx];
        j = 0;
        jlen = value.length;
        while (j < jlen) {
          result[result.length] = value[j];
          j += 1;
        }
      } else {
        result[result.length] = list[idx];
      }
      idx += 1;
    }
    return result;
  };
}
var _makeFlat_1 = _makeFlat;

/**
 * Returns a new list by pulling every item out of it (and all its sub-arrays)
 * and putting them in a new array, depth-first.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [b]
 * @param {Array} list The array to consider.
 * @return {Array} The flattened list.
 * @see R.unnest
 * @example
 *
 *      R.flatten([1, 2, [3, 4], 5, [6, [7, 8, [9, [10, 11], 12]]]]);
 *      //=> [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
 */


var flatten = /*#__PURE__*/_curry1_1( /*#__PURE__*/_makeFlat_1(true));
var flatten_1 = flatten;

/**
 * Returns a string made by inserting the `separator` between each element and
 * concatenating all the elements into a single string.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig String -> [a] -> String
 * @param {Number|String} separator The string used to separate the elements.
 * @param {Array} xs The elements to join into a string.
 * @return {String} str The string made by concatenating `xs` with `separator`.
 * @see R.split
 * @example
 *
 *      var spacer = R.join(' ');
 *      spacer(['a', 2, 3.4]);   //=> 'a 2 3.4'
 *      R.join('|', [1, 2, 3]);    //=> '1|2|3'
 */


var join = /*#__PURE__*/invoker_1(1, 'join');
var join_1 = join;

/**
 * Adds two values.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Math
 * @sig Number -> Number -> Number
 * @param {Number} a
 * @param {Number} b
 * @return {Number}
 * @see R.subtract
 * @example
 *
 *      R.add(2, 3);       //=>  5
 *      R.add(7)(10);      //=> 17
 */


var add = /*#__PURE__*/_curry2_1(function add(a, b) {
  return Number(a) + Number(b);
});
var add_1 = add;

/**
 * Adds together all the elements of a list.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Math
 * @sig [Number] -> Number
 * @param {Array} list An array of numbers
 * @return {Number} The sum of all the numbers in the list.
 * @see R.reduce
 * @example
 *
 *      R.sum([2,4,6,8,100,1]); //=> 121
 */


var sum = /*#__PURE__*/reduce_1(add_1, 0);
var sum_1 = sum;

function _reduced$3(x) {
  return x && x['@@transducer/reduced'] ? x : {
    '@@transducer/value': x,
    '@@transducer/reduced': true
  };
}
var _reduced_1 = _reduced$3;

var XAll = /*#__PURE__*/function () {

  function XAll(f, xf) {
    this.xf = xf;
    this.f = f;
    this.all = true;
  }
  XAll.prototype['@@transducer/init'] = _xfBase.init;
  XAll.prototype['@@transducer/result'] = function (result) {
    if (this.all) {
      result = this.xf['@@transducer/step'](result, true);
    }
    return this.xf['@@transducer/result'](result);
  };
  XAll.prototype['@@transducer/step'] = function (result, input) {
    if (!this.f(input)) {
      this.all = false;
      result = _reduced_1(this.xf['@@transducer/step'](result, false));
    }
    return result;
  };

  return XAll;
}();

var _xall$3 = /*#__PURE__*/_curry2_1(function _xall(f, xf) {
  return new XAll(f, xf);
});
var _xall_1 = _xall$3;

/**
 * Returns `true` if all elements of the list match the predicate, `false` if
 * there are any that don't.
 *
 * Dispatches to the `all` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig (a -> Boolean) -> [a] -> Boolean
 * @param {Function} fn The predicate function.
 * @param {Array} list The array to consider.
 * @return {Boolean} `true` if the predicate is satisfied by every element, `false`
 *         otherwise.
 * @see R.any, R.none, R.transduce
 * @example
 *
 *      var equals3 = R.equals(3);
 *      R.all(equals3)([3, 3, 3, 3]); //=> true
 *      R.all(equals3)([3, 3, 1, 3]); //=> false
 */


var all$2 = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['all'], _xall_1, function all(fn, list) {
  var idx = 0;
  while (idx < list.length) {
    if (!fn(list[idx])) {
      return false;
    }
    idx += 1;
  }
  return true;
}));
var all_1 = all$2;

/**
 * Finds the set (i.e. no duplicates) of all elements in the first list not
 * contained in the second list. Objects and Arrays are compared in terms of
 * value equality, not reference equality.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Relation
 * @sig [*] -> [*] -> [*]
 * @param {Array} list1 The first list.
 * @param {Array} list2 The second list.
 * @return {Array} The elements in `list1` that are not in `list2`.
 * @see R.differenceWith, R.symmetricDifference, R.symmetricDifferenceWith, R.without
 * @example
 *
 *      R.difference([1,2,3,4], [7,6,5,4,3]); //=> [1,2]
 *      R.difference([7,6,5,4,3], [1,2,3,4]); //=> [7,6,5]
 *      R.difference([{a: 1}, {b: 2}], [{a: 1}, {c: 3}]) //=> [{b: 2}]
 */


var difference = /*#__PURE__*/_curry2_1(function difference(first, second) {
  var out = [];
  var idx = 0;
  var firstLen = first.length;
  while (idx < firstLen) {
    if (!_contains_1(first[idx], second) && !_contains_1(first[idx], out)) {
      out[out.length] = first[idx];
    }
    idx += 1;
  }
  return out;
});
var difference_1 = difference;

/**
 * Create a new object with the own properties of the first object merged with
 * the own properties of the second object. If a key exists in both objects,
 * the value from the second object will be used.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Object
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} l
 * @param {Object} r
 * @return {Object}
 * @see R.mergeDeepRight, R.mergeWith, R.mergeWithKey
 * @example
 *
 *      R.merge({ 'name': 'fred', 'age': 10 }, { 'age': 40 });
 *      //=> { 'name': 'fred', 'age': 40 }
 *
 *      var resetToDefault = R.merge(R.__, {x: 0});
 *      resetToDefault({x: 5, y: 2}); //=> {x: 0, y: 2}
 * @symb R.merge({ x: 1, y: 2 }, { y: 5, z: 3 }) = { x: 1, y: 5, z: 3 }
 */


var merge = /*#__PURE__*/_curry2_1(function merge(l, r) {
  return _assign({}, l, r);
});
var merge_1 = merge;

/**
 * A function that always returns `true`. Any passed in parameters are ignored.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Function
 * @sig * -> Boolean
 * @param {*}
 * @return {Boolean}
 * @see R.always, R.F
 * @example
 *
 *      R.T(); //=> true
 */


var T = /*#__PURE__*/always_1(true);
var T_1 = T;

/**
 * Returns a function, `fn`, which encapsulates `if/else, if/else, ...` logic.
 * `R.cond` takes a list of [predicate, transformer] pairs. All of the arguments
 * to `fn` are applied to each of the predicates in turn until one returns a
 * "truthy" value, at which point `fn` returns the result of applying its
 * arguments to the corresponding transformer. If none of the predicates
 * matches, `fn` returns undefined.
 *
 * @func
 * @memberOf R
 * @since v0.6.0
 * @category Logic
 * @sig [[(*... -> Boolean),(*... -> *)]] -> (*... -> *)
 * @param {Array} pairs A list of [predicate, transformer]
 * @return {Function}
 * @example
 *
 *      var fn = R.cond([
 *        [R.equals(0),   R.always('water freezes at 0°C')],
 *        [R.equals(100), R.always('water boils at 100°C')],
 *        [R.T,           temp => 'nothing special happens at ' + temp + '°C']
 *      ]);
 *      fn(0); //=> 'water freezes at 0°C'
 *      fn(50); //=> 'nothing special happens at 50°C'
 *      fn(100); //=> 'water boils at 100°C'
 */


var cond = /*#__PURE__*/_curry1_1(function cond(pairs) {
  var arity = reduce_1(max_1, 0, map_1(function (pair) {
    return pair[0].length;
  }, pairs));
  return _arity_1(arity, function () {
    var idx = 0;
    while (idx < pairs.length) {
      if (pairs[idx][0].apply(this, arguments)) {
        return pairs[idx][1].apply(this, arguments);
      }
      idx += 1;
    }
  });
});
var cond_1 = cond;

/**
 * `tryCatch` takes two functions, a `tryer` and a `catcher`. The returned
 * function evaluates the `tryer`; if it does not throw, it simply returns the
 * result. If the `tryer` *does* throw, the returned function evaluates the
 * `catcher` function and returns its result. Note that for effective
 * composition with this function, both the `tryer` and `catcher` functions
 * must return the same type of results.
 *
 * @func
 * @memberOf R
 * @since v0.20.0
 * @category Function
 * @sig (...x -> a) -> ((e, ...x) -> a) -> (...x -> a)
 * @param {Function} tryer The function that may throw.
 * @param {Function} catcher The function that will be evaluated if `tryer` throws.
 * @return {Function} A new function that will catch exceptions and send then to the catcher.
 * @example
 *
 *      R.tryCatch(R.prop('x'), R.F)({x: true}); //=> true
 *      R.tryCatch(R.prop('x'), R.F)(null);      //=> false
 */


var tryCatch = /*#__PURE__*/_curry2_1(function _tryCatch(tryer, catcher) {
  return _arity_1(tryer.length, function () {
    try {
      return tryer.apply(this, arguments);
    } catch (e) {
      return catcher.apply(this, _concat_1([e], arguments));
    }
  });
});
var tryCatch_1 = tryCatch;

/**
 * Converts an object into an array of key, value arrays. Only the object's
 * own properties are used.
 * Note that the order of the output array is not guaranteed to be consistent
 * across different JS platforms.
 *
 * @func
 * @memberOf R
 * @since v0.4.0
 * @category Object
 * @sig {String: *} -> [[String,*]]
 * @param {Object} obj The object to extract from
 * @return {Array} An array of key, value arrays from the object's own properties.
 * @see R.fromPairs
 * @example
 *
 *      R.toPairs({a: 1, b: 2, c: 3}); //=> [['a', 1], ['b', 2], ['c', 3]]
 */


var toPairs = /*#__PURE__*/_curry1_1(function toPairs(obj) {
  var pairs = [];
  for (var prop in obj) {
    if (_has_1(prop, obj)) {
      pairs[pairs.length] = [prop, obj[prop]];
    }
  }
  return pairs;
});
var toPairs_1 = toPairs;

/**
 * Creates a new object from a list key-value pairs. If a key appears in
 * multiple pairs, the rightmost pair is included in the object.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category List
 * @sig [[k,v]] -> {k: v}
 * @param {Array} pairs An array of two-element arrays that will be the keys and values of the output object.
 * @return {Object} The object made by pairing up `keys` and `values`.
 * @see R.toPairs, R.pair
 * @example
 *
 *      R.fromPairs([['a', 1], ['b', 2], ['c', 3]]); //=> {a: 1, b: 2, c: 3}
 */


var fromPairs = /*#__PURE__*/_curry1_1(function fromPairs(pairs) {
  var result = {};
  var idx = 0;
  while (idx < pairs.length) {
    result[pairs[idx][0]] = pairs[idx][1];
    idx += 1;
  }
  return result;
});
var fromPairs_1 = fromPairs;

/**
 * Determines whether a nested path on an object has a specific value, in
 * [`R.equals`](#equals) terms. Most likely used to filter a list.
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category Relation
 * @typedefn Idx = String | Int
 * @sig [Idx] -> a -> {a} -> Boolean
 * @param {Array} path The path of the nested property to use
 * @param {*} val The value to compare the nested property with
 * @param {Object} obj The object to check the nested property in
 * @return {Boolean} `true` if the value equals the nested object property,
 *         `false` otherwise.
 * @example
 *
 *      var user1 = { address: { zipCode: 90210 } };
 *      var user2 = { address: { zipCode: 55555 } };
 *      var user3 = { name: 'Bob' };
 *      var users = [ user1, user2, user3 ];
 *      var isFamous = R.pathEq(['address', 'zipCode'], 90210);
 *      R.filter(isFamous, users); //=> [ user1 ]
 */


var pathEq = /*#__PURE__*/_curry3_1(function pathEq(_path, val, obj) {
  return equals_1(path_1(_path, obj), val);
});
var pathEq_1 = pathEq;

var isPlainObj = pathEq_1(['constructor', 'name'], 'Object');

// eslint-disable-next-line no-console
var consoler = curry_1(function (key, val) {
    return console.error('\n  Transform on field "' + key + '" failed\n', val);
});

/**
 * Wraps every function on a given spec in a try/catch that captures exceptions
 * to the console. It also logs the field name whose transform produced the error.
 *
 * @func
 * @sig {k: (a -> b)} -> {k: (a -> b)}
 * @param {Object} spec An object whose values may be transform functions that
 * need to be wrapped in try/catch
 * @returns {Object} The same spec object but whose functions are
 * safely wrapped in in try/catch handlers
 */
var safeSpecTransforms = compose_1(fromPairs_1, map_1(function (_ref) {
    var fieldName = _ref[0];
    var transform = _ref[1];
    return [fieldName, when_1(is_1(Function), tryCatch_1(__, consoler(fieldName)))(transform)];
}), toPairs_1);

/**
 * A port of [Ramda's evolve()](http://ramdajs.com/docs/#evolve), but the
 * transforms are _always_ applied regardless if the key/value pair exists in
 * the input object or not.
 * 
 *
 * @func
 * @sig {k: (a -> b)} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values may be transform functions that
 * need to be wrapped in try/catch
 * @param {Object} input The input object to pass through prop-level transformations
 * @returns {Object} The modified input object, with any prop-level transforms
 * applied to it
 */
var alwaysEvolve = curry_1(function (spec, input) {
    var result = {};
    var object = unless_1(isPlainObj, always_1({}))(input);
    keys_1(spec).forEach(function (key) {
        result[key] = cond_1([[is_1(Function), compose_1(applyTo_1(object[key]), tryCatch_1(__, consoler(key)))], [isPlainObj, alwaysEvolve(__, object[key])], [T_1, always_1(object[key])]])(spec[key]);
    });
    return result;
});

var baseShaper = curry_1(function (evolver, spec, value) {
    return pipe_1(unless_1(isPlainObj, always_1({})), merge_1, applyTo_1(reject_1(is_1(Function), spec)), evolver(compose_1(safeSpecTransforms, filter_1(is_1(Function)))(spec)), converge_1(merge_1, [identity_1, converge_1(applyTo_1, [when_1(isEmpty_1, always_1(value)), compose_1(applySpec_1, safeSpecTransforms, applyTo_1(spec), pick_1, difference_1(keys_1(filter_1(is_1(Function), spec))), keys_1)])]))(value);
});

/**
 * Combines two values of the same type (if it makes sense to combine them).
 * Numbers are summarized, strings and arrays are concatenated together, and
 * true objects are merged (the second value merged on top of the first).
 * In any other case only the first value is returned.
 * 
 * @func
 * @sig a -> a -> a
 * @param {*} val1 The base value to be combined with
 * @param {*} val2 The value to combine
 * @returns {*} If the values are of the same type, this represents the combined
 * value of the two of them. Otherwise only the first value is returned
 */
var combine = curry_1(function (val1, val2) {
    return cond_1([[all_1(is_1(Number)), sum_1], [all_1(is_1(String)), join_1('')], [all_1(is_1(Array)), flatten_1], [all_1(isPlainObj), mergeAll_1], [T_1, always_1(val1)]])([val1, val2]);
});

/**
 * Similar to [Ramda's applySpec()](http://ramdajs.com/docs/#applySpec), however
 * the spec will be applied and then merged into the original input.
 * Ramda's applySpec() will omit the original input, returning an object containing _only_
 * the props named in your spec, but this function here will apply the spec
 * and then merge back it into the original input.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are functions which
 * expect the entire object to be fed in as input.
 * This is identical to the input you'd pass into applySpec()
 * @returns {Function} A function that is ready to receive an input object and
 * merge into it the result of applying the provied spec
 */
var mergeSpec = curry_1(function (spec, value) {
    return converge_1(combine, [identity_1, applySpec_1(safeSpecTransforms(spec))])(value);
});

/**
 * This function allows one to blend an object with a clone that is taken
 * through prop transformations. Unlike object mappers/transformers, not every
 * value in the spec need be a function. Non-function values are simply added to
 * the result. If a given value in the spec IS a function however, it will be
 * used to transform OR to create a prop. If the prop specified on the spec
 * exists on the value (object) being mapped, a prop-level transform will be
 * perform (just like [Ramda's evolve()](http://ramdajs.com/docs/#evolve)).
 * But if a prop in the spec doesn't exist in the object and if that prop is a
 * transformer function, then the transformer will be used to create that prop
 * on the resulting object BUT the entire object will be passed into the
 * transform function (just like [Ramda's applySpec()](http://ramdajs.com/docs/#applySpec), rather than just a single prop.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or pass-through values to be added to the return object
 * @returns {Function} A function that is ready to receive an input object and
 * merge into it the result of applying the provied spec
 */
var shapeLoosely = baseShaper(evolve_1);

/**
 * Applies a shaping spec to an input object, but will NOT pass through any props unless they are named in the spec.
 * In other words, you're providing it template for creating a brand new object from some raw input.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or pass-through values to be added to the return object
 * @returns {Function} A function that is ready to receive an input object and
 * then pick from it to apply the custom shape defined in the spec
 */
var shapeStrictly = curry_1(function (spec, input) {
    return compose_1(pick_1(keys_1(spec)), shapeLoosely(spec))(input);
});

/**
 * Applies a shaping spec to an input object, but will never pass in the input object to a transform
 * when it is unable to find a key on the input object that corresponds to the key of the transform spec.
 * In other words, the transform functions will be applied to the prop on the input object OR to a value of undefined.
 * Write your transform functions accordingly.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or pass-through values to be added to the return object
 * @returns {Function} A function that is ready to receive an input object and
 * then pick from it to apply the custom shape defined in the spec
 */
var shapeSuperStrictly = baseShaper(alwaysEvolve);

/**
 * Creates a transform function from either a function, a spec object with field transform functions,
 * or some other type of value. Can also support strict shaping mode
 * (only fields on the spec get created in the final output)
 *
 * @func
 * @sig * -> (a -> b)
 * @param {*} val A spec, a function, or some other value (which will be always returned)
 * @returns {Function} A function that is ready to apply a (potential)
 * transformation on an input value to be provided later
 */
var makeShaper = curry_1(function (val, input) {
    return compose_1(when_1(isPlainObj, omit_1(['shapeyMode'])), cond_1([[is_1(Function), identity_1], [both_1(isPlainObj, pathSatisfies_1(test_1(/strict/i), ['shapeyMode'])), shapeStrictly], [both_1(isPlainObj, pathSatisfies_1(test_1(/super(.){0,1}strict/i), ['shapeyMode'])), shapeSuperStrictly], [isPlainObj, shapeLoosely], [T_1, always_1]])(val))(input);
});

/**
 * Applies a list of functions (in sequence) to a single input, passing
 * the transformed output as the input value to the next function in the chain.
 * If a spec object is included in the transforms, `shapeLooseley()` is invoked
 * on it, to create a transform function for the pipeline.
 *
 * @func
 * @sig [a -> b, b -> c, c -> d, ...] -> * -> *
 * @param {Function[]} transforms A list of transform functions
 * @param {*} input The input value to pass through the enhancer pipeline
 * @returns {*} The output of the original input passed through the chain of
 * transform functions
 */
var shapeline = curry_1(function (transforms, input) {
    return reduce_1(function (inputObj, fn) {
        return fn(inputObj);
    }, input, map_1(makeShaper, transforms || []));
});

var XAny = /*#__PURE__*/function () {

  function XAny(f, xf) {
    this.xf = xf;
    this.f = f;
    this.any = false;
  }
  XAny.prototype['@@transducer/init'] = _xfBase.init;
  XAny.prototype['@@transducer/result'] = function (result) {
    if (!this.any) {
      result = this.xf['@@transducer/step'](result, false);
    }
    return this.xf['@@transducer/result'](result);
  };
  XAny.prototype['@@transducer/step'] = function (result, input) {
    if (this.f(input)) {
      this.any = true;
      result = _reduced_1(this.xf['@@transducer/step'](result, true));
    }
    return result;
  };

  return XAny;
}();

var _xany = /*#__PURE__*/_curry2_1(function _xany(f, xf) {
  return new XAny(f, xf);
});
var _xany_1 = _xany;

/**
 * Returns `true` if at least one of elements of the list match the predicate,
 * `false` otherwise.
 *
 * Dispatches to the `any` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig (a -> Boolean) -> [a] -> Boolean
 * @param {Function} fn The predicate function.
 * @param {Array} list The array to consider.
 * @return {Boolean} `true` if the predicate is satisfied by at least one element, `false`
 *         otherwise.
 * @see R.all, R.none, R.transduce
 * @example
 *
 *      var lessThan0 = R.flip(R.lt)(0);
 *      var lessThan2 = R.flip(R.lt)(2);
 *      R.any(lessThan0)([1, 2]); //=> false
 *      R.any(lessThan2)([1, 2]); //=> true
 */


var any = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['any'], _xany_1, function any(fn, list) {
  var idx = 0;
  while (idx < list.length) {
    if (fn(list[idx])) {
      return true;
    }
    idx += 1;
  }
  return false;
}));
var any_1 = any;

/**
 * Takes two arguments, `fst` and `snd`, and returns `[fst, snd]`.
 *
 * @func
 * @memberOf R
 * @since v0.18.0
 * @category List
 * @sig a -> b -> (a,b)
 * @param {*} fst
 * @param {*} snd
 * @return {Array}
 * @see R.objOf, R.of
 * @example
 *
 *      R.pair('foo', 'bar'); //=> ['foo', 'bar']
 */


var pair = /*#__PURE__*/_curry2_1(function pair(fst, snd) {
  return [fst, snd];
});
var pair_1 = pair;

/**
 * Returns the last element of the given list or string.
 *
 * @func
 * @memberOf R
 * @since v0.1.4
 * @category List
 * @sig [a] -> a | Undefined
 * @sig String -> String
 * @param {*} list
 * @return {*}
 * @see R.init, R.head, R.tail
 * @example
 *
 *      R.last(['fi', 'fo', 'fum']); //=> 'fum'
 *      R.last([]); //=> undefined
 *
 *      R.last('abc'); //=> 'c'
 *      R.last(''); //=> ''
 */


var last = /*#__PURE__*/nth_1(-1);
var last_1 = last;

/**
 * Returns `true` if the specified value is equal, in [`R.equals`](#equals)
 * terms, to at least one element of the given list; `false` otherwise.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig a -> [a] -> Boolean
 * @param {Object} a The item to compare against.
 * @param {Array} list The array to consider.
 * @return {Boolean} `true` if an equivalent item is in the list, `false` otherwise.
 * @see R.any
 * @example
 *
 *      R.contains(3, [1, 2, 3]); //=> true
 *      R.contains(4, [1, 2, 3]); //=> false
 *      R.contains({ name: 'Fred' }, [{ name: 'Fred' }]); //=> true
 *      R.contains([42], [[42]]); //=> true
 */


var contains$1 = /*#__PURE__*/_curry2_1(_contains_1);
var contains_1 = contains$1;

/**
 * Returns whether or not an object has an own property with the specified name
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category Object
 * @sig s -> {s: x} -> Boolean
 * @param {String} prop The name of the property to check for.
 * @param {Object} obj The object to query.
 * @return {Boolean} Whether the property exists.
 * @example
 *
 *      var hasName = R.has('name');
 *      hasName({name: 'alice'});   //=> true
 *      hasName({name: 'bob'});     //=> true
 *      hasName({});                //=> false
 *
 *      var point = {x: 0, y: 0};
 *      var pointHas = R.has(R.__, point);
 *      pointHas('x');  //=> true
 *      pointHas('y');  //=> true
 *      pointHas('z');  //=> false
 */


var has = /*#__PURE__*/_curry2_1(_has_1);
var has_1 = has;

/**
 * Accepts a function `fn` and a list of transformer functions and returns a
 * new curried function. When the new function is invoked, it calls the
 * function `fn` with parameters consisting of the result of calling each
 * supplied handler on successive arguments to the new function.
 *
 * If more arguments are passed to the returned function than transformer
 * functions, those arguments are passed directly to `fn` as additional
 * parameters. If you expect additional arguments that don't need to be
 * transformed, although you can ignore them, it's best to pass an identity
 * function so that the new function reports the correct arity.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig ((x1, x2, ...) -> z) -> [(a -> x1), (b -> x2), ...] -> (a -> b -> ... -> z)
 * @param {Function} fn The function to wrap.
 * @param {Array} transformers A list of transformer functions
 * @return {Function} The wrapped function.
 * @see R.converge
 * @example
 *
 *      R.useWith(Math.pow, [R.identity, R.identity])(3, 4); //=> 81
 *      R.useWith(Math.pow, [R.identity, R.identity])(3)(4); //=> 81
 *      R.useWith(Math.pow, [R.dec, R.inc])(3, 4); //=> 32
 *      R.useWith(Math.pow, [R.dec, R.inc])(3)(4); //=> 32
 * @symb R.useWith(f, [g, h])(a, b) = f(g(a), h(b))
 */


var useWith = /*#__PURE__*/_curry2_1(function useWith(fn, transformers) {
  return curryN_1(transformers.length, function () {
    var args = [];
    var idx = 0;
    while (idx < transformers.length) {
      args.push(transformers[idx].call(this, arguments[idx]));
      idx += 1;
    }
    return fn.apply(this, args.concat(Array.prototype.slice.call(arguments, transformers.length)));
  });
});
var useWith_1 = useWith;

var _Set = /*#__PURE__*/function () {

  function _Set() {
    /* globals Set */
    this._nativeSet = typeof Set === 'function' ? new Set() : null;
    this._items = {};
  }

  // until we figure out why jsdoc chokes on this
  // @param item The item to add to the Set
  // @returns {boolean} true if the item did not exist prior, otherwise false
  //
  _Set.prototype.add = function (item) {
    return !hasOrAdd(item, true, this);
  };

  //
  // @param item The item to check for existence in the Set
  // @returns {boolean} true if the item exists in the Set, otherwise false
  //
  _Set.prototype.has = function (item) {
    return hasOrAdd(item, false, this);
  };

  //
  // Combines the logic for checking whether an item is a member of the set and
  // for adding a new item to the set.
  //
  // @param item       The item to check or add to the Set instance.
  // @param shouldAdd  If true, the item will be added to the set if it doesn't
  //                   already exist.
  // @param set        The set instance to check or add to.
  // @return {boolean} true if the item already existed, otherwise false.
  //
  return _Set;
}();

function hasOrAdd(item, shouldAdd, set) {
  var type = typeof item;
  var prevSize, newSize;
  switch (type) {
    case 'string':
    case 'number':
      // distinguish between +0 and -0
      if (item === 0 && 1 / item === -Infinity) {
        if (set._items['-0']) {
          return true;
        } else {
          if (shouldAdd) {
            set._items['-0'] = true;
          }
          return false;
        }
      }
      // these types can all utilise the native Set
      if (set._nativeSet !== null) {
        if (shouldAdd) {
          prevSize = set._nativeSet.size;
          set._nativeSet.add(item);
          newSize = set._nativeSet.size;
          return newSize === prevSize;
        } else {
          return set._nativeSet.has(item);
        }
      } else {
        if (!(type in set._items)) {
          if (shouldAdd) {
            set._items[type] = {};
            set._items[type][item] = true;
          }
          return false;
        } else if (item in set._items[type]) {
          return true;
        } else {
          if (shouldAdd) {
            set._items[type][item] = true;
          }
          return false;
        }
      }

    case 'boolean':
      // set._items['boolean'] holds a two element array
      // representing [ falseExists, trueExists ]
      if (type in set._items) {
        var bIdx = item ? 1 : 0;
        if (set._items[type][bIdx]) {
          return true;
        } else {
          if (shouldAdd) {
            set._items[type][bIdx] = true;
          }
          return false;
        }
      } else {
        if (shouldAdd) {
          set._items[type] = item ? [false, true] : [true, false];
        }
        return false;
      }

    case 'function':
      // compare functions for reference equality
      if (set._nativeSet !== null) {
        if (shouldAdd) {
          prevSize = set._nativeSet.size;
          set._nativeSet.add(item);
          newSize = set._nativeSet.size;
          return newSize === prevSize;
        } else {
          return set._nativeSet.has(item);
        }
      } else {
        if (!(type in set._items)) {
          if (shouldAdd) {
            set._items[type] = [item];
          }
          return false;
        }
        if (!_contains_1(item, set._items[type])) {
          if (shouldAdd) {
            set._items[type].push(item);
          }
          return false;
        }
        return true;
      }

    case 'undefined':
      if (set._items[type]) {
        return true;
      } else {
        if (shouldAdd) {
          set._items[type] = true;
        }
        return false;
      }

    case 'object':
      if (item === null) {
        if (!set._items['null']) {
          if (shouldAdd) {
            set._items['null'] = true;
          }
          return false;
        }
        return true;
      }
    /* falls through */
    default:
      // reduce the search size of heterogeneous sets by creating buckets
      // for each type.
      type = Object.prototype.toString.call(item);
      if (!(type in set._items)) {
        if (shouldAdd) {
          set._items[type] = [item];
        }
        return false;
      }
      // scan through all previously applied items
      if (!_contains_1(item, set._items[type])) {
        if (shouldAdd) {
          set._items[type].push(item);
        }
        return false;
      }
      return true;
  }
}

// A simple Set type that honours R.equals semantics
var _Set_1 = _Set;

/**
 * Returns a new list containing only one copy of each element in the original
 * list, based upon the value returned by applying the supplied function to
 * each list element. Prefers the first item if the supplied function produces
 * the same value on two items. [`R.equals`](#equals) is used for comparison.
 *
 * @func
 * @memberOf R
 * @since v0.16.0
 * @category List
 * @sig (a -> b) -> [a] -> [a]
 * @param {Function} fn A function used to produce a value to use during comparisons.
 * @param {Array} list The array to consider.
 * @return {Array} The list of unique items.
 * @example
 *
 *      R.uniqBy(Math.abs, [-1, -5, 2, 10, 1, 2]); //=> [-1, -5, 2, 10]
 */


var uniqBy = /*#__PURE__*/_curry2_1(function uniqBy(fn, list) {
  var set = new _Set_1();
  var result = [];
  var idx = 0;
  var appliedItem, item;

  while (idx < list.length) {
    item = list[idx];
    appliedItem = fn(item);
    if (set.add(appliedItem)) {
      result.push(item);
    }
    idx += 1;
  }
  return result;
});
var uniqBy_1 = uniqBy;

/**
 * Returns a new list containing only one copy of each element in the original
 * list. [`R.equals`](#equals) is used to determine equality.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [a]
 * @param {Array} list The array to consider.
 * @return {Array} The list of unique items.
 * @example
 *
 *      R.uniq([1, 1, 2, 1]); //=> [1, 2]
 *      R.uniq([1, '1']);     //=> [1, '1']
 *      R.uniq([[42], [42]]); //=> [[42]]
 */


var uniq = /*#__PURE__*/uniqBy_1(identity_1);
var uniq_1 = uniq;

/**
 * Returns the second argument if it is not `null`, `undefined` or `NaN`;
 * otherwise the first argument is returned.
 *
 * @func
 * @memberOf R
 * @since v0.10.0
 * @category Logic
 * @sig a -> b -> a | b
 * @param {a} default The default value.
 * @param {b} val `val` will be returned instead of `default` unless `val` is `null`, `undefined` or `NaN`.
 * @return {*} The second value if it is not `null`, `undefined` or `NaN`, otherwise the default value
 * @example
 *
 *      var defaultTo42 = R.defaultTo(42);
 *
 *      defaultTo42(null);  //=> 42
 *      defaultTo42(undefined);  //=> 42
 *      defaultTo42('Ramda');  //=> 'Ramda'
 *      // parseInt('string') results in NaN
 *      defaultTo42(parseInt('string')); //=> 42
 */


var defaultTo = /*#__PURE__*/_curry2_1(function defaultTo(d, v) {
  return v == null || v !== v ? d : v;
});
var defaultTo_1 = defaultTo;

/**
 * Takes a function `f` and returns a function `g` such that if called with the same arguments
 * when `f` returns a "truthy" value, `g` returns `false` and when `f` returns a "falsy" value `g` returns `true`.
 *
 * `R.complement` may be applied to any functor
 *
 * @func
 * @memberOf R
 * @since v0.12.0
 * @category Logic
 * @sig (*... -> *) -> (*... -> Boolean)
 * @param {Function} f
 * @return {Function}
 * @see R.not
 * @example
 *
 *      var isNotNil = R.complement(R.isNil);
 *      isNil(null); //=> true
 *      isNotNil(null); //=> false
 *      isNil(7); //=> false
 *      isNotNil(7); //=> true
 */


var complement = /*#__PURE__*/lift_1(not_1);
var complement_1 = complement;

/**
 * Checks to see if a prop name is an alphanumeric string (plus some symbols
 * allowed, like underscores, dashes and dots)
 *
 * @func
 * @sig a -> Boolean
 * @param {String} prop A prop name to check for formatting
 * @returns {Boolean} whether or not the prop name passed validation
 */
var isValidPropName = test_1(/^(?:[A-Z])([A-Z0-9_\-.]+)([A-Z0-9])$/i);

/**
 * Checks to see if a given value is a JavaScript Promise
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value to check to see if is a Promise
 * @returns {Boolean} whether or not the val is a Promise
 */
var isPromise = compose_1(equals_1('Promise'), path_1(['constructor', 'name']), defaultTo_1(''));

/**
 * Checks to see if a given value is an Object
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value to check for objectishness
 * @returns {Boolean} whether or not the val is a plain old JS object
 */
var isPlainObj$1 = compose_1(equals_1('Object'), path_1(['constructor', 'name']), defaultTo_1(''));

/**
 * Checks to make sure a given value isn't null or undefined
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value which may or may not be null / undefined
 * @returns {Boolean} whether or not the value was non-nil
 */
var isNotNil = complement_1(isNil_1);

/**
 * Checks to make sure a given value isn't empty
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value which may or may not be empty
 * @returns {Boolean} whether or not the value was non-empty
 */
var isNotEmpty = complement_1(isEmpty_1);

/**
 * Checks to see whether or not a given String is non-blank (one or more chars)
 *
 * @func
 * @sig String -> Boolean
 * @param {String} val A String which may or may not be blank
 * @returns {Boolean} whether or not a given String is non-blank
 */
var isNotBlankString = compose_1(not_1, test_1(/^\s*$/));

/**
 * Checks to see whether or not a given value is a non-blank String (one or more chars)
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value which may or may not be a non-blank String
 * @returns {Boolean} whether or not a given value is a non-blank String
 */
var isStringieThingie = allPass_1([isNotBlankString, either_1(is_1(Number), is_1(String)), isNotNil]);

/**
 * Checks to see whether or not a given value is either: Boolean, Number, String, Data, or RegExp
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value which may or may not be "primitive-ish"
 * @returns {Boolean} whether or not a given value is "primitive-ish"
 */
var isPrimitiveish = anyPass_1([is_1(Boolean), is_1(Number), is_1(String), is_1(RegExp), is_1(Date)]);

/**
 * Check that action is an object with a "type" prop that is a non-blank string
 * 
 * @func
 * @sig * -> Boolean
 * @param {*} val A value which may or may not be a Redux action
 * @returns {Boolean}
 */
var isAction = both_1(isPlainObj$1, pathSatisfies_1(allPass_1([is_1(String), test_1(/\S/)]), ['type']));

/**
 * Checks to see if a provided object has a given prop (path)
 *
 * @func
 * @sig [String] -> {k: v} -> Boolean
 * @param {String[]} propPath An array of string values representing the nested path to a prop
 * @param {Object} obj An object on which a given prop may exist
 * @returns {Boolean} whether or not the provided prop path exists on the provided object
 */
var hasNestedProp = curry_1(function (propPath, obj) {
  return compose_1(isNotNil, path_1(propPath))(obj);
});

var _extends$1 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

/**
 * Takes a prop of any type and depending will wrap it in an array -
 * unless the prop is an array, in which case it will be returned [as-is](http://ramdajs.com/docs/#identity)
 *
 * @func
 * @sig * -> [*]
 * @param {Array|*} p A prop of any type (arrays will be returned as-is)
 * @returns {Array} either the original prop (if it was an array) or an Array wrapping the original prop
 */
var coerceToArray = unless_1(is_1(Array), Array);

/**
 * Takes a prop of any type and depending will [wrap a function around it](http://ramdajs.com/docs/#always) -
 * unless the prop is a function, in which case it will be returned [as-is](http://ramdajs.com/docs/#identity)
 *
 * @func
 * @sig * -> (() -> *)
 * @param {Function|*} p A prop of any type (functions will be returned as-is)
 * @returns {Function} either the original prop (if it was a Function) or a Function wrapping the original prop
 */
var coerceToFn = unless_1(is_1(Function), always_1);

/**
 * Converts a value of any type to [its string equivalent](http://ramdajs.com/docs/#toString),
 * but passes through a String value [as-is](http://ramdajs.com/docs/#identity)
 *
 * @func
 * @sig * -> String
 * @param {*} s A value of any type (strings will be returned as-is)
 * @returns {String} either the original prop (if it was a String) or a stringified rendering of the original prop
 */
var coerceToString = unless_1(is_1(String), toString);

/**
 * Simplest of common reducer functions that merges an array of key/value pairs into a single object
 *
 * @func
 * @sig ({k: v}, [k, v]) -> {k: v}
 * @param {Object} returnObj Accumulator object that each key/val pair will be
 * merged into
 * @param {Array} pairs A simple key/value pair array (just two indexes in the array)
 * @returns {Objedt} The merged object of key/value pairs
 */
var listOfPairsToOneObject = function listOfPairsToOneObject(returnObj, _ref) {
    var _extends2;

    var key = _ref[0];
    var val = _ref[1];
    return _extends$1({}, returnObj, (_extends2 = {}, _extends2[key] = val, _extends2));
};

/**
 * A function which merges two Objects, with the latter merging over the former
 * upon finding any duplicate fields. Also will return one or the other in full
 * (un-merged) if one of the params is not an Object.
 *
 * @func
 * @sig ({k: v}, {k: v}) -> *
 * @param {Object} parent An Object whose props will have lower precedence when
 * merged with another Object
 * @param {Object} child An Object whose props will have higher precedence when
 * merged with another Object
 * @returns {*} the merged result of two values
 */
var simpleMergeStrategy = function simpleMergeStrategy(parent, child) {
    if (type_1(parent) !== type_1(child) || isNil_1(child)) {
        if (is_1(Array, parent) && !isNil_1(child)) {
            return [].concat(parent, coerceToArray(child));
        }
        if (!isPlainObj$1(parent)) {
            return parent;
        }
    } else if (isPrimitiveish(child) || is_1(Function, child)) {
        return child;
    } else if (is_1(Array, parent)) {
        return [].concat(parent, coerceToArray(child));
    }

    return mergeDeepRight_1(parent, child);
};

/**
 * Creates a new object out of a list of keys and a list of values.
 * Key/value pairing is truncated to the length of the shorter of the two lists.
 * Note: `zipObj` is equivalent to `pipe(zip, fromPairs)`.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category List
 * @sig [String] -> [*] -> {String: *}
 * @param {Array} keys The array that will be properties on the output object.
 * @param {Array} values The list of values on the output object.
 * @return {Object} The object made by pairing up same-indexed elements of `keys` and `values`.
 * @example
 *
 *      R.zipObj(['a', 'b', 'c'], [1, 2, 3]); //=> {a: 1, b: 2, c: 3}
 */


var zipObj = /*#__PURE__*/_curry2_1(function zipObj(keys, values) {
  var idx = 0;
  var len = Math.min(keys.length, values.length);
  var out = {};
  while (idx < len) {
    out[keys[idx]] = values[idx];
    idx += 1;
  }
  return out;
});
var zipObj_1 = zipObj;

/**
 * Retrieves all a duck's namespaced types
 *
 * @func
 * @sig {k: v} -> [String]
 * @param {Object} duck An object containing an object of action types, whose values are namespaced
 * @returns {String[]} A list of all the namespaced types for a given duck
 */
var getTypes = compose_1(values_1, prop_1('types'));

/**
 * Merges together either two lists or appends a non-list type into another list
 *
 * @func
 * @sig ([a], b) -> [a, b]
 * @param {Array} left A list of any type of values
 * @param {Array|*} right Either a list of values or any kind of non-list value
 * to append to the 'left' list
 * @returns {Array|*} a merged list (if 'left' was an array, otherwise 'right')
 */
var concatOrReplace = function concatOrReplace(left, right) {
    return is_1(Array, left) ? [].concat(left, coerceToArray(right)) : right;
};

/**
 * A function which validates and then deep freezes an object of constants
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} constants An Object which should contain only primitive values or arrays of primitives values
 * @returns {Object} A frozen Object with only the values which have been validated
 */
var createConstants = ifElse_1(
/* no nested functions or objects, just primitives or conversion of arrays
 * of primitives into simple objects whose keys and vals are the same */
allPass_1([compose_1(not_1, isPrimitiveish), compose_1(not_1, is_1(Array)), is_1(Object)]), compose_1(Object.freeze, reduce_1(listOfPairsToOneObject, {}), reject_1(any_1(either_1(isNil_1, isEmpty_1))), map_1(function (_ref) {
    var name = _ref[0];
    var value = _ref[1];

    if (is_1(Array, value)) {
        /* Creates an object whose keys and values are identical */
        return [name, Object.freeze(zipObj_1(value.filter(isPrimitiveish), value.filter(isPrimitiveish)))];
    } else if (isPlainObj$1(value)) {
        return [name, compose_1(ifElse_1(isEmpty_1, always_1(null), Object.freeze), filter_1(isPrimitiveish))(value)];
    }
    return isPrimitiveish(value) ? [name, value] : null;
}), toPairs_1, reject_1(isEmpty_1)), always_1({}));

/**
 * A function which takes an Object containing namespace and store values and returns
 * a function that will formats an array of strings to have the namespace and
 * store as prefixes.
 *
 * @func
 * @sig {k: v} -> ([String] -> {k: v})
 * @param {Object} p An Object with both a 'namespace' and 'store' string value,
 * from which to format each action type
 * @returns {Function} A function which takes an array of Strings (action types)
 * and creates an Object formatted such that the each key and value matches the string,
 * but formatted as 'namespace/store/type'
 */
var createTypes = curry_1(function (_ref2, types) {
    var namespace = _ref2.namespace;
    var store = _ref2.store;
    return zipObj_1(types, types.map(function (type) {
        return (namespace || '') + '/' + (store || '') + '/' + type;
    }));
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

/**
 * Simple check to see whether a given duck's machines is empty.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} machines A duck's collection of state machines
 * @returns {Boolean} whether or not the machines are empty
 */
var noMachines = anyPass_1([isNil_1, isEmpty_1]);

/**
 * Adds an object containing the current state of all the state machines onto
 * a given section of the redux store
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck A duck instance, containing the state machines and the
 * prop name/path to the section of the store where their state is tracked
 * @param {*} state The current state of this portion of the store
 * @returns {Object} original state (or initialState, if state is nil) plus
 * current state of all the state machines
 */
var addTransitionsToState = curry_1(function (state, _ref) {
    var _ref$initialState = _ref.initialState;
    var initialState = _ref$initialState === undefined ? {} : _ref$initialState;
    var _ref$stateMachinesPro = _ref.stateMachinesPropName;
    var stateMachinesPropName = _ref$stateMachinesPro === undefined ? ['states'] : _ref$stateMachinesPro;
    return compose_1(unless_1(hasNestedProp(stateMachinesPropName), assocPath_1(stateMachinesPropName, {})), defaultTo_1(initialState))(state);
});

/**
 * Retrieves a unique list all the state transitions possible for a given machine.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the transition-to state is the value of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> [String]
 * @param {Object} machine A single state machine
 * @returns {String[]} A list of unique state transitions for the provided machine
 */
var getTransitionsForMachine = compose_1(uniq_1, filter_1(isStringieThingie), flatten_1, map_1(values_1), values_1);

/**
 * Retrieves a unique list all the inputs for a given state machine (which correspond to Action Types).
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> [String]
 * @param {Object} machine A single state machine
 * @returns {String[]} A list of unique inputs to the provided state machine
 */
var getStateInputsForMachine = compose_1(uniq_1, filter_1(isStringieThingie), flatten_1, map_1(keys_1), values_1);

/**
 * Retrieves a unique list all the inputs for all of a given duck's state machines.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> [String]
 * @param {Object} duck A duck containing one or more state machines
 * @returns {String[]} A list of unique inputs to all the state machines for the provided duck
 */
var getStateInputsForAllMachines = compose_1(uniq_1, flatten_1, map_1(getStateInputsForMachine), values_1, prop_1('machines'));

/**
 * Retrieves any invalid inputs from a given duck's state machines.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> [String]
 * @param {Object} duck A duck containing one or more state machines and action types
 * @returns {String[]} A list of any invalid inputs for the state machines of the provided duck
 */
var invalidStateMachineInputs = converge_1(difference_1, [getStateInputsForAllMachines, getTypes]);

/**
 * Checks whether all a given state machine's states are string values.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} machine A single state machine
 * @returns {Boolean} whether or not the states for the state machine are strings
 */
var areStateNamesStrings = compose_1(all_1(is_1(String)), keys_1);

/**
 * Checks whether all a given state machine's inputs and transition names are string values.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key and transition is the value of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} machine A single state machine
 * @returns {Boolean} whether or not the inputs and transitions for the state machine are strings
 */
var areInputsAndTransitionsStrings = compose_1(all_1(allPass_1([isPlainObj$1, either_1(isEmpty_1, compose_1(all_1(all_1(is_1(String))), toPairs_1))])), values_1);

/**
 * Checks whether all the state machine inputs for a given duck correspond to its action types.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} duck A duck containing one or more state machines and action types
 * @returns {Boolean} whether or not all the inputs for the state machines are among the duck's action types
 */
var areStateMachineInputsActionTypes = compose_1(all_1(isEmpty_1), invalidStateMachineInputs);

/**
 * Checks if all a state machine's transitions are states in that machine.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the transition-to state is the value of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} machine A single state machine
 * @returns {Boolean} whether or not all the transition names are also states for a given state machine
 */
var isEachTransitionAmongMachineStates = useWith_1(difference_1, [getTransitionsForMachine, keys_1]);

/**
 * Checks to see whether a given machine allows for a given transition, provided its current state
 *
 * @func
 * @sig String -> String -> {k: v} -> Boolean
 * @param {String} transitionName A string value representing a potential state transition
 * @param {String} currentState A string value representing the current state
 * @param {Object} machine A state machine which contains the possible states and their associated transitions
 * @returns {Boolean} whether or not the transition is possible from the current state
 */
var isTransitionPossible = curry_1(function (transitionName, currentState, machine) {
    return compose_1(has_1(coerceToString(transitionName)), defaultTo_1({}), path_1([coerceToString(currentState)]), defaultTo_1({}))(machine);
});

/**
 * Creates a "machine state" object, which can be combined with others to create a single state machine.
 * The keys for a machine state object (called inputs or input values) correspond to the names of redux action types.
 * The values for a machine state object (called transitions) are merely
 * the names of fellow machine state objects (also created by this function).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} machine A single state machine
 * @param {Object} types An object whose keys/values correspond
 * to redux actions and will be used to set up inputs for the state machine
 * @returns {Object} a validated, immutable state machine
 */
var createMachineStates = curry_1(function () {
    var machine = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    var types = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
    return compose_1(Object.freeze, reduce_1(listOfPairsToOneObject, {}), map_1(function (_ref2) {
        var state = _ref2[0];
        var transitions = _ref2[1];
        return [state, pick_1(compose_1(map_1(head_1), filter_1(compose_1(contains_1(__, values_1(types)), head_1)), filter_1(compose_1(has_1(__, machine), last_1)), filter_1(all_1(is_1(String))), toPairs_1)(transitions), transitions)];
    }), filter_1(both_1(compose_1(is_1(String), head_1), compose_1(isPlainObj$1, last_1))), toPairs_1)(machine);
});

/**
 * Creates a frozen (immutable) object whose keys are all the possible "current state"
 * values for the machine and whose values are objects containing each action type that
 * may produce a transition to another state.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} machines An object full of one or more state machines
 * @param {Object} duck A duck instance
 * @returns {Object} an object of validated, immutable state machines
 */
var createMachines = curry_1(function (machines, duck) {
    return compose_1(Object.freeze, reduce_1(listOfPairsToOneObject, {}), map_1(function (_ref3) {
        var name = _ref3[0];
        var machine = _ref3[1];
        return [name, createMachineStates(machine, duck.types)];
    }), toPairs_1)(machines);
});

/**
 * Creates an string path (array) that the state machine functions will use to
 * lookup and alter the current state of one or more of the state machines.
 * The path supplied to this function can be a single prop name (String) or
 * an Array of prop names that represent a nested prop's path.
 * And if a String is passed in, a dot-separated format is also supported.
 *
 * @func
 * @sig String -> [String]
 * @param {String|String[]} path The prop name or path to the section of the
 * store where the state machines current state is tracked.
 * @returns {String[]} an array of strings representing the nested path to the
 * prop where state machines current state is tracked in the redux store
 */
var getStateMachinesPropPath = ifElse_1(allPass_1([is_1(String), contains_1('.'), isValidPropName]), split_1('.'), compose_1(when_1(isEmpty_1, always_1(['states'])), filter_1(isNotBlankString), when_1(is_1(String), Array)));

/**
 * Builds a flattened object where the keys are the names of each of the duck's state machines,
 * and the values are each set to 'initial'.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} machines An object containing the definitions for one or more state machines
 * @returns {Object} a flattened object where the current state of each state machine will be tracked
 */
var getDefaultStateForMachines = function getDefaultStateForMachines() {
    var machines = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return reduce_1(function (obj, key) {
        var _extends2;

        return _extends({}, obj, (_extends2 = {}, _extends2[key] = 'initial', _extends2));
    }, {}, keys_1(machines || {}));
};

/**
 * Retrieves the section of the redux store where the current state is being
 * tracked for one or more state machines. The shape of this section of the store
 * is usually just a flattened object with a single value for each named state
 * machine, representing its current state. Each possible state for any of those
 * is defined in the Ducks.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck A duck instance, containing the state machines and the
 * prop name/path to the section of the store where their state is tracked
 * @param {*} state The current state of this portion of the store
 * @returns {Object} A flattened Object containing the names of each state
 * machine (String) and its current state (String) as key/value pairs
 */
var getCurrentState = function getCurrentState() {
    var _ref4 = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

    var _ref4$machines = _ref4.machines;
    var machines = _ref4$machines === undefined ? {} : _ref4$machines;
    var stateMachinesPropName = _ref4.stateMachinesPropName;
    return compose_1(mergeDeepRight_1(map_1(always_1('initial'), machines)), pick_1(keys_1(machines)), defaultTo_1({}), path_1(coerceToArray(stateMachinesPropName)));
};

/**
 * Based on the current state of the machines (managed in the store itself) and
 * the rules defined for each state machine (in the ducks), an action is processed
 * and may cause one or more ore the state machines to transition.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object|*} state The current state of this portion of the store
 * @param {Object} action A dispatched redux action
 * @param {Object} duck A duck instance, containing the state machines and the
 * prop name/path to the section of the store where their state is tracked
 * @returns {Object} A flattened Object containing the names of each state
 * machine (String) and its current state (String) as key/value pairs
 */
var getNextState = curry_1(function (state, action, _ref5) {
    var _ref5$machines = _ref5.machines;
    var machines = _ref5$machines === undefined ? {} : _ref5$machines;
    var stateMachinesPropName = _ref5.stateMachinesPropName;

    var currentState = getCurrentState({ machines: machines, stateMachinesPropName: stateMachinesPropName })(state);
    return compose_1(reduce_1(listOfPairsToOneObject, {}), map_1(function (_ref6) {
        var name = _ref6[0];
        var machine = _ref6[1];
        return compose_1(pair_1(name), unless_1(isNotNil, always_1(currentState[name])), prop_1(action.type), defaultTo_1({}), prop_1(currentState[name]))(machine);
    }), toPairs_1)(machines);
});

/**
 * Checks whether a dispatched action's type is listed among the
 * current state's input values for one or more of the state machines.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v} -> Boolean
 * @param {Object|*} state The current state of this portion of the store
 * @param {Object} action A dispatch redux action
 * @param {Object} duck A duck, which contains both state machines and the
 * corresponding prop name/path in the redux store where their current state is tracked
 * @returns {Boolean} whether or not the dispatched action is listed as an input
 * value for the state machine(s) current state
 */
var isActionTypeInCurrentState = curry_1(function (state, action, _ref7) {
    var machines = _ref7.machines;
    var stateMachinesPropName = _ref7.stateMachinesPropName;

    var currentState = getCurrentState({ machines: machines, stateMachinesPropName: stateMachinesPropName })(state);
    return compose_1(any_1(function (_ref8) {
        var name = _ref8[0];
        var machine = _ref8[1];
        return compose_1(isNotNil, prop_1(action.type), defaultTo_1({}), prop_1(__, machine), prop_1(name))(currentState);
    }), toPairs_1)(machines);
});

var _extends$2 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var initialIfNil = function initialIfNil(state, duck) {
    return isNil_1(state) ? prop_1('initialState', duck) : state;
};

/**
 * Creates a standard ducks reducer, where the duck is the context object,
 * passed as the third param (standard redux reducers are two params, first for
 * the current state of the store and the second is the dispatched redux
 * action).
 * This function is then applied to the state along with a dispatched action
 * that may induce a change in state.
 *
 * @func ((a, b, c) -> a) -> c -> a -> b -> a
 * @param {Function} reducer A function which processes state, action, and duck params
 * to validate the action and (potentially) apply a change to state.
 * @param {Object} duck A duck, containing types, consts, validators and other
 * helpers that are used to carry out the reducer's logic
 * @param {Object} state The current state of a certain portion of the store
 * @param {Object} action A dispatched redux action, with (at minimum) a 'type' prop
 * @returns {Object} Returns a clone of the state, potentially different than
 * the original state if the action object caused a change. Because this is a
 * curried function, you can omit the final two params (state and action) to set
 * this partially curried function as a redux reducer in the combineReducers() setup.
 */
var makeReducer = curry_1(function (reducer, duck, state, action) {
    return reducer(initialIfNil(state, duck), action, duck);
});

/**
 * Creates a standard ducks reducer, where the duck is the context object,
 * passed as the third param (standard redux reducers are two params, first for
 * the current state of the store and the second is the dispatched redux
 * action).
 * This function is then applied to the state along with a dispatched action
 * that may induce a change in state.
 *
 * Appended to this reducer is an automatic validation of the action to see if
 * it is registered as an accepted input to one or more of the state machines
 * (inside the duck), and if so, the corresponding state transition is then applied
 * to the store (at the prop/path which the user specified when they setup their duck).
 *
 * @func
 * @sig ((a, b, c) -> a) -> c -> a -> b -> a
 * @param {Function} reducer A function which processes state, action, and duck params
 * to validate the action and (potentially) apply a change to state.
 * @param {Object} duck A duck, containing types, consts, validators and other
 * helpers that are used to carry out the reducer's logic
 * @param {Object} state The current state of a certain portion of the store
 * @param {Object} action A dispatched redux action, with (at minimum) a 'type' prop
 * @returns {Object} Returns a clone of the state, potentially different than
 * the original state if the action object caused a change. Because this is a
 * curried function, you can omit the final two params (state and action) to set
 * this partially curried function as a redux reducer in the combineReducers() setup.
 */
var makeTransitionsPostReducer = curry_1(function (reducer, dux, state, action) {
    var stateWithTransitions = addTransitionsToState(initialIfNil(state, dux), dux);

    var updatedStates = assocPath_1(dux.stateMachinesPropName, getNextState(stateWithTransitions, action, dux), {});

    return _extends$2({}, reducer(_extends$2({}, stateWithTransitions, updatedStates), action, dux), updatedStates);
});

/**
 * Creates a layered reducer which invokes one and then the other, passing in
 * the result of the first reducer as the 'state' param to the second reducer.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v} -> {k: v} -> {k: v}
 * @param {Object} parentOptions An options object for a duck being extended, containing
 * types, consts, validators and other helpers that are used to carry out the reducer's logic
 * @param {Object} childOptions An options object for a duck that has yet to be created
 * but which are being extened from an existing duck to validate the action and (potentially)
 * apply a change to state.
 * @param {Object} state The current state of a certain portion of the store
 * @param {Object} action A dispatched redux action, with (at minimum) a 'type' prop
 * @param {Object} duck A duck, containing types, consts, validators and other
 * helpers that are used to carry out the reducer's logic
 * @returns {Object} Returns a clone of the state, potentially different than
 * the original state if the action object caused a change. Because this is a
 * curried function, you can omit the final two params (state and action) to set
 * this partially curried function as a redux reducer in the combineReducers() setup.
 */
var makeExtendedReducer = curry_1(function (parentOptions, childOptions, state, action, duck) {
    return childOptions.reducer(parentOptions.reducer(initialIfNil(state), action, duck), action, duck);
});

/**
 * Parses the reducer from a duck (inside the 'options' prop),
 * wraps it with the duck (as context) so that the reducer code will be able to execute
 * (as it likely parses the types or other items from the duck to carry out its logic on the action payload).
 * The resulting function is then ready to be passed state and a dispatched action.
 *
 * @func
 * @sig {k: v} -> ((a, b) -> a)
 * @param {Object} duck A duck, containing a reducer prop inside of its options prop
 * @returns {Function} A wrapped reducer, ready to be invoked with state and action params
 */
var createReducer = converge_1(makeReducer, [compose_1(defaultTo_1(identity_1), path_1(['options', 'reducer'])), identity_1]);

/**
 * Parses the reducer from a duck (inside the 'options' prop),
 * wraps it with the duck (as context) so that the reducer code will be able to execute
 * (as it likely parses the types or other items from the duck to carry out its logic on the action payload),
 * and finally a state machine transitions validation is appended to the
 * reducer, so that the portion of the store that tracks state transitions will
 * be updated.
 * The resulting function is then ready to be passed state and a dispatched action.
 *
 * @func
 * @sig {k: v} -> ((a, b) -> a)
 * @param {Object} duck A duck, containing a reducer prop inside of its options prop
 * @returns {Function} A wrapped reducer, ready to be invoked with state and action params
 */
var createTransitionsPostReducer = converge_1(makeTransitionsPostReducer, [compose_1(defaultTo_1(identity_1), path_1(['options', 'reducer'])), identity_1]);

/**
 * Returns all but the last element of the given list or string.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category List
 * @sig [a] -> [a]
 * @sig String -> String
 * @param {*} list
 * @return {*}
 * @see R.last, R.head, R.tail
 * @example
 *
 *      R.init([1, 2, 3]);  //=> [1, 2]
 *      R.init([1, 2]);     //=> [1]
 *      R.init([1]);        //=> []
 *      R.init([]);         //=> []
 *
 *      R.init('abc');  //=> 'ab'
 *      R.init('ab');   //=> 'a'
 *      R.init('a');    //=> ''
 *      R.init('');     //=> ''
 */


var init = /*#__PURE__*/slice_1(0, -1);
var init_1 = init;

/**
 * A customisable version of [`R.memoize`](#memoize). `memoizeWith` takes an
 * additional function that will be applied to a given argument set and used to
 * create the cache key under which the results of the function to be memoized
 * will be stored. Care must be taken when implementing key generation to avoid
 * clashes that may overwrite previous entries erroneously.
 *
 *
 * @func
 * @memberOf R
 * @since v0.24.0
 * @category Function
 * @sig (*... -> String) -> (*... -> a) -> (*... -> a)
 * @param {Function} fn The function to generate the cache key.
 * @param {Function} fn The function to memoize.
 * @return {Function} Memoized version of `fn`.
 * @see R.memoize
 * @example
 *
 *      let count = 0;
 *      const factorial = R.memoizeWith(R.identity, n => {
 *        count += 1;
 *        return R.product(R.range(1, n + 1));
 *      });
 *      factorial(5); //=> 120
 *      factorial(5); //=> 120
 *      factorial(5); //=> 120
 *      count; //=> 1
 */


var memoizeWith = /*#__PURE__*/_curry2_1(function memoizeWith(mFn, fn) {
  var cache = {};
  return _arity_1(fn.length, function () {
    var key = mFn.apply(this, arguments);
    if (!_has_1(key, cache)) {
      cache[key] = fn.apply(this, arguments);
    }
    return cache[key];
  });
});
var memoizeWith_1 = memoizeWith;

/**
 * Creates a new function that, when invoked, caches the result of calling `fn`
 * for a given argument set and returns the result. Subsequent calls to the
 * memoized `fn` with the same argument set will not result in an additional
 * call to `fn`; instead, the cached result for that set of arguments will be
 * returned.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category Function
 * @sig (*... -> a) -> (*... -> a)
 * @param {Function} fn The function to memoize.
 * @return {Function} Memoized version of `fn`.
 * @see R.memoizeWith
 * @deprecated since v0.25.0
 * @example
 *
 *      let count = 0;
 *      const factorial = R.memoize(n => {
 *        count += 1;
 *        return R.product(R.range(1, n + 1));
 *      });
 *      factorial(5); //=> 120
 *      factorial(5); //=> 120
 *      factorial(5); //=> 120
 *      count; //=> 1
 */


var memoize = /*#__PURE__*/memoizeWith_1(function () {
  return toString_1(arguments);
});
var memoize_1 = memoize;

/**
 * Returns a partial copy of an object containing only the keys that satisfy
 * the supplied predicate.
 *
 * @func
 * @memberOf R
 * @since v0.8.0
 * @category Object
 * @sig ((v, k) -> Boolean) -> {k: v} -> {k: v}
 * @param {Function} pred A predicate to determine whether or not a key
 *        should be included on the output object.
 * @param {Object} obj The object to copy from
 * @return {Object} A new object with only properties that satisfy `pred`
 *         on it.
 * @see R.pick, R.filter
 * @example
 *
 *      var isUpperCase = (val, key) => key.toUpperCase() === key;
 *      R.pickBy(isUpperCase, {a: 1, b: 2, A: 3, B: 4}); //=> {A: 3, B: 4}
 */


var pickBy = /*#__PURE__*/_curry2_1(function pickBy(test, obj) {
  var result = {};
  for (var prop in obj) {
    if (test(obj[prop], prop, obj)) {
      result[prop] = obj[prop];
    }
  }
  return result;
});
var pickBy_1 = pickBy;

var _extends$3 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

/**
 * A simple function that checks an Object for a prop called "needsExtraction"
 * and determines if it is set to `true`
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object}
 * @returns {Boolean}
 */
var needsExtraction = compose_1(equals_1(true), prop_1('needsExtraction'));

/**
 * Helper utility to assist in composing the selectors.
 * Previously defined selectors can be used to derive future selectors.
 * Any selector which is a pre-extracted state is expected to have a
 * `needsExtraction: true` and an accompanying `justAddDuckSelectors` prop
 * which performs the extraction. This function creates a sort of running total
 * of all selectors and passes them into each `justAddDuckSelectors()`.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} selectors An Object of selector functions, some of which may
 * need to be extracted (those which require access to fellow selectors)
 * @returns {Object} selectors, but now with access to fellow selectors
 */
var deriveSelectors = function deriveSelectors() {
    var selectors = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return compose_1(reduce_1(function (composedSelectors, _ref) {
        var _extends2;

        var key = _ref[0];
        var selector = _ref[1];
        return _extends$3({}, composedSelectors, (_extends2 = {}, _extends2[key] = selector.justAddDuckSelectors(composedSelectors), _extends2));
    }, selectors), toPairs_1, pickBy_1(needsExtraction))(selectors);
};

/**
 * A simple, Ramda implementation of Reselect's `createSelector()` function,
 * taken from [this example](https://twitter.com/sharifsbeat/status/891001130632830976)
 * The cost of all the Ramda functions which make up Attadux has already been
 * paid, so may as well save on another dependency. You can still use Reselect
 * if you prefer; they both work the same.
 *
 * @func
 * @sig (*... -> a) -> (*... -> a)
 * @ param {...Function} selectors One or more selector functions to be "merged" into one
 * @returns {Function} A memoized "selector" function
 */
var createSelector = function createSelector() {
    for (var _len = arguments.length, selectors = Array(_len), _key = 0; _key < _len; _key++) {
        selectors[_key] = arguments[_key];
    }

    return memoize_1(converge_1(last_1(selectors), init_1(selectors)));
};

/**
 * Selectors that require one or more of the existing selectors are built using this helper,
 * and the extraction function is passed in as a param. The simplest type of
 * extractor function that this helper was initially built to handle would be:
 *
 * selectors => state => selectors.getSomethingFrom(state)
 *
 * Additionally, the return Object is in a ready-to-be-extracted state (which is
 * why it is flagged `needsExtraction: true`) and so invoking its
 * `justAddDuckSelectors()` will perform the final step of running the extractor
 * (ideally when the rest of the selectors are all being derived)
 *
 * @func
 * @sig Function -> {k: v}
 * @param {Function} extractFunction A helper that will extract all the existing selectors
 * @returns {Object}
 */
var createDuckSelector = function createDuckSelector(extractFunction) {
    return {
        needsExtraction: true,
        justAddDuckSelectors: function justAddDuckSelectors() {
            var allSelectorsObject = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

            var extracted = extractFunction(allSelectorsObject);
            if (Array.isArray(extracted)) {
                return createSelector.apply(undefined, extracted);
            }
            return extracted;
        }
    };
};

/**
 * If the given, non-null object has a value at the given path, returns the
 * value at that path. Otherwise returns the provided default value.
 *
 * @func
 * @memberOf R
 * @since v0.18.0
 * @category Object
 * @typedefn Idx = String | Int
 * @sig a -> [Idx] -> {a} -> a
 * @param {*} d The default value.
 * @param {Array} p The path to use.
 * @param {Object} obj The object to retrieve the nested property from.
 * @return {*} The data at `path` of the supplied object or the default value.
 * @example
 *
 *      R.pathOr('N/A', ['a', 'b'], {a: {b: 2}}); //=> 2
 *      R.pathOr('N/A', ['a', 'b'], {c: {b: 2}}); //=> "N/A"
 */


var pathOr = /*#__PURE__*/_curry3_1(function pathOr(d, p, obj) {
  return defaultTo_1(d, path_1(p, obj));
});
var pathOr_1 = pathOr;

/**
 * Replace a substring or regex match in a string with a replacement.
 *
 * @func
 * @memberOf R
 * @since v0.7.0
 * @category String
 * @sig RegExp|String -> String -> String -> String
 * @param {RegExp|String} pattern A regular expression or a substring to match.
 * @param {String} replacement The string to replace the matches with.
 * @param {String} str The String to do the search and replacement in.
 * @return {String} The result.
 * @example
 *
 *      R.replace('foo', 'bar', 'foo foo foo'); //=> 'bar foo foo'
 *      R.replace(/foo/, 'bar', 'foo foo foo'); //=> 'bar foo foo'
 *
 *      // Use the "g" (global) flag to replace all occurrences:
 *      R.replace(/foo/g, 'bar', 'foo foo foo'); //=> 'bar bar bar'
 */


var replace = /*#__PURE__*/_curry3_1(function replace(regex, replacement, str) {
  return str.replace(regex, replacement);
});
var replace_1 = replace;

function e(e,t,n){e.addEventListener("message",function(r){var o=r.data,a=o.id;if("RPC"===o.type&&null!=a)if(o.method){var s=t[o.method];null==s?e.postMessage({type:"RPC",id:a,error:"NO_SUCH_METHOD"}):Promise.resolve().then(function(){return s.apply(null,o.params)}).then(function(t){e.postMessage({type:"RPC",id:a,result:t});}).catch(function(t){e.postMessage({type:"RPC",id:a,error:""+t});});}else{var i=n[a];if(null==i)throw Error("Unknown callback "+a);delete n[a],o.error?i[1](Error(o.error)):i[0](o.result);}});}var workerize = function(t,n){var r=this,o={},a="__xpo"+Math.random().toString().substring(2)+"__";"function"==typeof t&&(t="("+Function.prototype.toString.call(t)+")("+a+")"),t=function(e,t,n){return e=(e=e.replace(/^(\s*)export\s+default\s+/m,function(e,r){return n.default=!0,""+r+t+".default="})).replace(/^(\s*)export\s+((?:async\s*)?function(?:\s*\*)?|const|let|var)(\s+)([a-zA-Z$_][a-zA-Z0-9$_]*)/gm,function(e,r,o,a,s){return n[s]=!0,""+r+t+"."+s+"="+o+a+s}),"var "+t+"={};\n"+e+"\n"+t+";"}(t,a,o)+"\n("+Function.prototype.toString.call(e)+")(self,"+a+",{})";var s,i=URL.createObjectURL(new Blob([t])),l=new Worker(i,n),c=l.terminate,u={},p=0;for(s in l.kill=function(e){l.postMessage({type:"KILL",signal:e}),setTimeout(l.terminate);},l.terminate=function(){URL.revokeObjectURL(i),c.call(r);},l.call=function(e,t){return new Promise(function(n,r){var o="rpc"+ ++p;u[o]=[n,r],l.postMessage({type:"RPC",id:o,method:e,params:t});})},l.rpcMethods={},e(l,l.rpcMethods,u),l.expose=function(e){l[s]=function(){return l.call(e,[].slice.call(arguments))};},o)s in l||l.expose(s);return l};

/* eslint-disable no-undef */
var hasWorkers = typeof Worker !== 'undefined' && Worker !== null && typeof Blob !== 'undefined' && Blob !== null;
/* eslint-enable no-undef */

var makeWorkerString = cond_1([[is_1(Function), invoker_1(0, 'toString')], [is_1(String), identity_1], [T_1, always_1('')]]);

var makeExportable = pipe_1(replace_1(/^(\s*)\(/m, 'export const run = ('), replace_1(/^(\s*)const/m, 'export const'), replace_1(/^(\s*)function\s*\(/m, 'export function run('), replace_1(/^(\s*)function\s+/m, 'export function '), replace_1(/^(\s*)function\*\s+/m, 'export function* '), replace_1(/^(\s*)async\s+\(/m, 'export const run = async ('), replace_1(/^(\s*)async\s+function\s*\(/m, 'export async function run('));

var makeWorker = pipe_1(makeWorkerString, makeExportable, unless_1(test_1(/^\s*$/), function (str) {
    return workerize('' + str);
}));

var getWorkers = converge_1(call_1, [compose_1(coerceToFn, pathOr_1({}, ['options', 'workers'])), identity_1]);

var makeWorkers = pipe_1(getWorkers, map_1(hasWorkers ? makeWorker : pipe_1(makeWorkerString, makeExportable)), reject_1(test_1(/^\s*$/)));

var XFind = /*#__PURE__*/function () {

  function XFind(f, xf) {
    this.xf = xf;
    this.f = f;
    this.found = false;
  }
  XFind.prototype['@@transducer/init'] = _xfBase.init;
  XFind.prototype['@@transducer/result'] = function (result) {
    if (!this.found) {
      result = this.xf['@@transducer/step'](result, void 0);
    }
    return this.xf['@@transducer/result'](result);
  };
  XFind.prototype['@@transducer/step'] = function (result, input) {
    if (this.f(input)) {
      this.found = true;
      result = _reduced_1(this.xf['@@transducer/step'](result, input));
    }
    return result;
  };

  return XFind;
}();

var _xfind = /*#__PURE__*/_curry2_1(function _xfind(f, xf) {
  return new XFind(f, xf);
});
var _xfind_1 = _xfind;

/**
 * Returns the first element of the list which matches the predicate, or
 * `undefined` if no element matches.
 *
 * Dispatches to the `find` method of the second argument, if present.
 *
 * Acts as a transducer if a transformer is given in list position.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig (a -> Boolean) -> [a] -> a | undefined
 * @param {Function} fn The predicate function used to determine if the element is the
 *        desired one.
 * @param {Array} list The array to consider.
 * @return {Object} The element found, or `undefined`.
 * @see R.transduce
 * @example
 *
 *      var xs = [{a: 1}, {a: 2}, {a: 3}];
 *      R.find(R.propEq('a', 2))(xs); //=> {a: 2}
 *      R.find(R.propEq('a', 4))(xs); //=> undefined
 */


var find = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['find'], _xfind_1, function find(fn, list) {
  var idx = 0;
  var len = list.length;
  while (idx < len) {
    if (fn(list[idx])) {
      return list[idx];
    }
    idx += 1;
  }
}));
var find_1 = find;

var makeQuery = function makeQuery(query) {
    if (is_1(Array, query)) {
        var buildQuery = find_1(is_1(Function), query) || identity_1;
        var raw = find_1(is_1(String), query) || '';
        return buildQuery(raw);
    } else if (is_1(String, query)) {
        return query;
    }
    return null;
};

var getQueries = converge_1(call_1, [compose_1(coerceToFn, pathOr_1({}, ['options', 'queries'])), identity_1]);

var makeQueries = compose_1(reject_1(isNil_1), map_1(makeQuery), getQueries);

var getRawQueries = compose_1(filter_1(is_1(String)), map_1(ifElse_1(is_1(Array), find_1(is_1(String)), always_1(null))), prop_1('queries'));

/**
 * Copies the string values from the Duck's queries into `consts.queries`.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain queries (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with `consts.queries`.
 */
var copyRawQueriesToConsts = converge_1(assocPath_1(['consts', 'queries']), [getRawQueries, identity_1]);

/**
 * Returns `true` if no elements of the list match the predicate, `false`
 * otherwise.
 *
 * Dispatches to the `any` method of the second argument, if present.
 *
 * @func
 * @memberOf R
 * @since v0.12.0
 * @category List
 * @sig (a -> Boolean) -> [a] -> Boolean
 * @param {Function} fn The predicate function.
 * @param {Array} list The array to consider.
 * @return {Boolean} `true` if the predicate is not satisfied by every element, `false` otherwise.
 * @see R.all, R.any
 * @example
 *
 *      var isEven = n => n % 2 === 0;
 *      var isOdd = n => n % 2 === 1;
 *
 *      R.none(isEven, [1, 3, 5, 7, 9, 11]); //=> true
 *      R.none(isOdd, [1, 3, 5, 7, 8, 11]); //=> false
 */


var none = /*#__PURE__*/_curry2_1( /*#__PURE__*/_complement_1( /*#__PURE__*/_dispatchable_1(['any'], _xany_1, any_1)));
var none_1 = none;

function _isNumber(x) {
  return Object.prototype.toString.call(x) === '[object Number]';
}
var _isNumber_1 = _isNumber;

/**
 * Returns the number of elements in the array by returning `list.length`.
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category List
 * @sig [a] -> Number
 * @param {Array} list The array to inspect.
 * @return {Number} The length of the array.
 * @example
 *
 *      R.length([]); //=> 0
 *      R.length([1, 2, 3]); //=> 3
 */


var length = /*#__PURE__*/_curry1_1(function length(list) {
  return list != null && _isNumber_1(list.length) ? list.length : NaN;
});
var length_1 = length;

/**
 * Returns a list of all the properties, including prototype properties, of the
 * supplied object.
 * Note that the order of the output array is not guaranteed to be consistent
 * across different JS platforms.
 *
 * @func
 * @memberOf R
 * @since v0.2.0
 * @category Object
 * @sig {k: v} -> [v]
 * @param {Object} obj The object to extract values from
 * @return {Array} An array of the values of the object's own and prototype properties.
 * @see R.values, R.keysIn
 * @example
 *
 *      var F = function() { this.x = 'X'; };
 *      F.prototype.y = 'Y';
 *      var f = new F();
 *      R.valuesIn(f); //=> ['X', 'Y']
 */


var valuesIn = /*#__PURE__*/_curry1_1(function valuesIn(obj) {
  var prop;
  var vs = [];
  for (prop in obj) {
    vs[vs.length] = obj[prop];
  }
  return vs;
});
var valuesIn_1 = valuesIn;

/**
 * Converts an object into an array of key, value arrays. The object's own
 * properties and prototype properties are used. Note that the order of the
 * output array is not guaranteed to be consistent across different JS
 * platforms.
 *
 * @func
 * @memberOf R
 * @since v0.4.0
 * @category Object
 * @sig {String: *} -> [[String,*]]
 * @param {Object} obj The object to extract from
 * @return {Array} An array of key, value arrays from the object's own
 *         and prototype properties.
 * @example
 *
 *      var F = function() { this.x = 'X'; };
 *      F.prototype.y = 'Y';
 *      var f = new F();
 *      R.toPairsIn(f); //=> [['x','X'], ['y','Y']]
 */


var toPairsIn = /*#__PURE__*/_curry1_1(function toPairsIn(obj) {
  var pairs = [];
  for (var prop in obj) {
    pairs[pairs.length] = [prop, obj[prop]];
  }
  return pairs;
});
var toPairsIn_1 = toPairsIn;

var ws = '\x09\x0A\x0B\x0C\x0D\x20\xA0\u1680\u180E\u2000\u2001\u2002\u2003' + '\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u202F\u205F\u3000\u2028' + '\u2029\uFEFF';
var zeroWidth = '\u200b';
var hasProtoTrim = typeof String.prototype.trim === 'function';
/**
 * Removes (strips) whitespace from both ends of the string.
 *
 * @func
 * @memberOf R
 * @since v0.6.0
 * @category String
 * @sig String -> String
 * @param {String} str The string to trim.
 * @return {String} Trimmed version of `str`.
 * @example
 *
 *      R.trim('   xyz  '); //=> 'xyz'
 *      R.map(R.trim, R.split(',', 'x, y, z')); //=> ['x', 'y', 'z']
 */
var _trim = !hasProtoTrim || /*#__PURE__*/ws.trim() || ! /*#__PURE__*/zeroWidth.trim() ? function trim(str) {
  var beginRx = new RegExp('^[' + ws + '][' + ws + ']*');
  var endRx = new RegExp('[' + ws + '][' + ws + ']*$');
  return str.replace(beginRx, '').replace(endRx, '');
} : function trim(str) {
  return str.trim();
};
var trim = /*#__PURE__*/_curry1_1(_trim);
var trim_1 = trim;

/**
 * The upper case version of a string.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category String
 * @sig String -> String
 * @param {String} str The string to upper case.
 * @return {String} The upper case version of `str`.
 * @see R.toLower
 * @example
 *
 *      R.toUpper('abc'); //=> 'ABC'
 */


var toUpper = /*#__PURE__*/invoker_1(0, 'toUpperCase');
var toUpper_1 = toUpper;

var _createConstants = createConstants({
    VALIDATION_LEVELS: ['STRICT', 'CANCEL', 'PRUNE', 'LOG']
});

var VALIDATION_LEVELS = _createConstants.VALIDATION_LEVELS;

var _extends$4 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

/**
 * Checks to see if a given value is one of the valid options for
 * Validation Level, which are: STRICT, CANCEL, PRUNE, LOG
 *
 * @func
 * @sig String -> Boolean
 * @param {String} level One of the four valid options: STRICT, CANCEL, PRUNE, LOG
 * @returns {Boolean} Whether or not a level is valid for middleware validations
 */
var isValidationLevel = contains_1(__, keys_1(VALIDATION_LEVELS));

/**
 * Sets a validation level, one of: STRICT, CANCEL, PRUNE, LOG.
 * The default - if no valid level is passed in - is CANCEL.
 *
 * @func
 * @sig String -> String
 * @param {String|*} level A validation level for the middleware
 * @returns {String} One of [STRICT, CANCEL, PRUNE, LOG]
 */
var makeValidationLevel = compose_1(unless_1(isValidationLevel, always_1('CANCEL')), toUpper_1, trim_1, coerceToString);

/**
 * A curried function that examines the result of a [spected](https://github.com/25th-floor/spected) validator on a given
 * input object and replaces the validation failures with the original value
 * from the input object.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} original The input object which when pushed through a validator yielded the associated validations
 * @param {Object} validations An object of validation results (true for pass, String[] for fails)
 * @returns {Object} The validations result pruned of all invalid fields
 */
var pruneInvalidFields = curry_1(function (original, validations) {
    return compose_1(reduce_1(function (prunedObj, _ref) {
        var _extends2;

        var key = _ref[0];
        var val = _ref[1];

        var value = isPlainObj$1(val) ? pruneInvalidFields(original[key], val) : val;
        if (isPlainObj$1(value) && isEmpty_1(value)) {
            return prunedObj;
        }
        return _extends$4({}, prunedObj, (_extends2 = {}, _extends2[key] = value === true ? original[key] : value, _extends2));
    }, {}), reject_1(compose_1(is_1(Array), nth_1(1))), toPairsIn_1)(validations);
});

/**
 * A function that inspects a validations object (validated fields are Bools of `true` and
 * invalid fields are arrays of error messages) and removes everything but the errors
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} validations An object of validation results (true for pass, String[] for fails)
 * @returns {Object} The validations result pruned of all valid fields
 */
var pruneValidatedFields = compose_1(reduce_1(function (prunedObj, _ref2) {
    var _extends3;

    var key = _ref2[0];
    var val = _ref2[1];

    var value = isPlainObj$1(val) ? pruneValidatedFields(val) : val;
    if (isPlainObj$1(value) && isEmpty_1(value)) {
        return prunedObj;
    }
    return _extends$4({}, prunedObj, (_extends3 = {}, _extends3[key] = value, _extends3));
}, {}), reject_1(function (pairs) {
    return pairs[1] === true;
}), toPairsIn_1);

/**
 * Simple retrieval function that looks for an action payload validator matching
 * a given action.
 *
 * @func
 * @sig {k: v} -> {k: v} -> ({k: v} -> Boolean)
 * @returns {Function} A validator matching a dispatched action (or a Function
 * always returning True if none exists)
 */
var getValidatorForAction = function getValidatorForAction(validators) {
    return compose_1(defaultTo_1(T_1), prop_1(__, validators), prop_1('type'));
};

/**
 * Creates a function that will receive a dispatched action and apply a validator function that
 * matches its action type (if one exists) and prune away any fields that failed validation.
 * The set of validator functions from which to match the action type is the only dependency.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} validators An object of curried [spected](https://github.com/25th-floor/spected) validator functions
 * @returns {Object} The original action paylod, minus any invalid fields
 */
var createPayloadPruner = function createPayloadPruner() {
    var validators = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return function () {
        var action = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
        return compose_1(when_1(isEmpty_1, always_1(null)), pruneInvalidFields(action), function (validate) {
            return validate(action);
        }, getValidatorForAction(validators))(action);
    };
};

/**
 * Creates a function that will receive a dispatched action and apply a validator function that
 * matches its action type (if one exists).
 * The set of validator functions from which to match the action type is the only dependency.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} validators An object of curried [spected](https://github.com/25th-floor/spected) validator functions
 * @returns {Object|null} An Object of validation errors, if any, otherwise returns null
 */
var createPayloadValidationsLogger = function createPayloadValidationsLogger() {
    var validators = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return function () {
        var action = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
        return compose_1(when_1(isEmpty_1, always_1(null)), pruneValidatedFields, applyTo_1(action), getValidatorForAction(validators))(action);
    };
};

/**
 * Inspects an Object (can even be nested) of validation results, for props which failed validation.
 * Because the [spected](https://github.com/25th-floor/spected) tool follows a common validations format
 * where failed validations are Arrays of validation errors (strings) and validated props are simply marked `true`,
 * this function is only doing a quick, simple check for any props that are arrays.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} validations
 * @returns {Boolean}
 */
var anyValidationFailures = function anyValidationFailures() {
    var validations = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return compose_1(any_1(either_1(both_1(isPlainObj$1, anyValidationFailures), is_1(Array))), valuesIn_1)(validations);
};

/**
 * Creates a function that will receive a dispatched action and apply a validator function that
 * matches its action type (if one exists) and do a basic pass/fail check.
 * The set of validator functions from which to match the action type is the only dependency.
 *
 * @func
 * @sig {k: v} -> {k: v} -> Boolean
 * @param {Object} validators An object of curried [spected](https://github.com/25th-floor/spected) validator functions
 * @returns {Function} A validator function to be applied against a dispatched action
 */
var createPayloadValidator = function createPayloadValidator() {
    var validators = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
    return function () {
        var action = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
        return compose_1(not_1, anyValidationFailures, applyTo_1(action), getValidatorForAction(validators))(action);
    };
};

/**
 * This object of functions is meant to be applied to an object which has
 * some or all of its keys sharing the same name as these evolver functions.
 *
 * Using [evolve()](http://ramdajs.docs/#evolve) on an object with this set of evolver functions
 * will create a clone of the original object with the result of each evolver in place of the original value.
 */
var metadataEvolvers = {
    namespace: identity_1,
    store: identity_1,
    validationLevel: makeValidationLevel,
    stateMachinesPropName: getStateMachinesPropPath,
    consts: createConstants

    /**
     * These are values which will always be set for a new duck,
     * even though the user may not explicitly supply a value for any of them.
     */
};var duxDefaults = {
    consts: {},
    creators: {},
    machines: {},
    queries: {},
    workers: {},
    selectors: {},
    stateMachinesPropName: 'states',
    types: [],
    validationLevel: 'CANCEL',
    validators: {}

    /**
     * This set of rules is expected to be supplied to the (curried) [spected](https://github.com/25th-floor/spected)
     * function, and is then invoked against the set of values supplied when creating a new duck.
     *
     * The spected validator function will then apply this collection of validator functions against
     * any matching keys on the input values passed in, and set either a value of 'true' if validation succeeded
     * or an array of error messages for any value(s) that failed validation.
     */
};var duxRules = {
    validationLevel: [[isValidationLevel, 'must be: STRICT, CANCEL, PRUNE, or LOG. CANCEL is the default.']],
    store: [[isStringieThingie, 'must be a (non-blank) string']],
    namespace: [[isStringieThingie, 'must be a (non-blank) string']],
    stateMachinesPropName: [[either_1(isStringieThingie, both_1(is_1(Array), all_1(isStringieThingie))), 'must be a string (or array of strings)']],
    consts: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    creators: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    machines: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    selectors: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    types: [[both_1(is_1(Array), all_1(is_1(String))), 'must be an object (or a function returning an object)']],
    validators: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    enhancers: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    multipliers: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    queries: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    workers: [[either_1(isPlainObj$1, is_1(Function)), 'must be an object (or a function returning an object)']],
    reducer: [[is_1(Function), 'must be a function']],
    effects: [[either_1(both_1(is_1(Array), all_1(is_1(Array))), is_1(Function)), 'must be an array of key/value pairs (or a function returning an one)']],
    throttling: [[either_1(both_1(is_1(Array), all_1(is_1(Array))), is_1(Function)), 'must be an array of key/value pairs (or a function returning an one)']],
    debouncing: [[either_1(both_1(is_1(Array), all_1(is_1(Array))), is_1(Function)), 'must be an array of key/value pairs (or a function returning an one)']],
    initialState: [[anyPass_1([isPrimitiveish, isPlainObj$1, is_1(Function)]), 'must be an object, a function returning an object, or a primitive value']]

    /**
     * This set of rules is expected to be supplied to the (curried) [spected](https://github.com/25th-floor/spected)
     * function, and is then invoked against one or more ducks supplied to the validator middleware.
     *
     * The spected validator function will then apply this collection of validator functions against
     * any matching keys on the input values passed in, and set either a value of 'true' if validation succeeded
     * or an array of error messages for any value(s) that failed validation.
     *
     * The validator middleware makes use of a small portion of a given duck, so the schema rules
     * for validating a usable duck in the middleware function is less strict than those used to create a duck.
     */
};var duckMiddlewareRules = {
    store: [[isStringieThingie, 'must be a (non-blank) string']],
    namespace: [[isStringieThingie, 'must be a (non-blank) string']],
    types: [[isPlainObj$1, 'must be an object'], [compose_1(all_1(compose_1(equals_1(1), length_1, uniq_1)), map_1(function (_ref) {
        var key = _ref[0];
        var val = _ref[1];
        return [key, compose_1(last_1, split_1('/'))(val)];
    }), filter_1(all_1(is_1(String))), toPairs_1), 'each key and value are identical']],
    machines: [[compose_1(all_1(isPlainObj$1), values_1), 'must be an object'], [compose_1(all_1(isNotEmpty), values_1), 'must not be empty'], [compose_1(all_1(allPass_1([areStateNamesStrings, areInputsAndTransitionsStrings])), values_1), 'each machine contains nested objects (states) whose inputs and transitions are strings'], [compose_1(all_1(isEachTransitionAmongMachineStates), values_1), 'each transition value must also be a state']],
    stateMachinesPropName: [[both_1(is_1(Array), all_1(isStringieThingie)), 'must be an array of strings (representing the path to the "current state" prop)']]

    /**
     * Checks to see if a given value is a Duck for middleware,
     * which means it is a namespaced unit that contains types and validators for one or more of those types.
     *
     * @func
     * @sig * -> Boolean
     * @param {*} val A value which may or may not be a Duck instance
     * @returns {Boolean} whether or not the object is an instance of a Duck
     */
};var isDux = compose_1(isEmpty_1, difference_1(['store', 'namespace', 'validators', 'types']), keys_1);

/**
 * Checks whether or not a given Object contains ducks
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} row An Object which may contain ducks among its props
 * @returns {Boolean} whether or not there are any ducks inside of a given Object
 */
var noDucks = anyPass_1([complement_1(isPlainObj$1), compose_1(isEmpty_1, keys_1), compose_1(none_1(isDux), values_1)]);

/**
 * Tests a regular expression against a String. Note that this function will
 * return an empty array when there are no matches. This differs from
 * [`String.prototype.match`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/match)
 * which returns `null` when there are no matches.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category String
 * @sig RegExp -> String -> [String | Undefined]
 * @param {RegExp} rx A regular expression.
 * @param {String} str The string to match against
 * @return {Array} The list of matches or empty array.
 * @see R.test
 * @example
 *
 *      R.match(/([a-z]a)/g, 'bananas'); //=> ['ba', 'na', 'na']
 *      R.match(/a/, 'b'); //=> []
 *      R.match(/a/, null); //=> TypeError: null does not have a method named "match"
 */


var match = /*#__PURE__*/_curry2_1(function match(rx, str) {
  return str.match(rx) || [];
});
var match_1 = match;

/**
 * Creates a new list out of the two supplied by pairing up equally-positioned
 * items from both lists. The returned list is truncated to the length of the
 * shorter of the two input lists.
 * Note: `zip` is equivalent to `zipWith(function(a, b) { return [a, b] })`.
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [b] -> [[a,b]]
 * @param {Array} list1 The first array to consider.
 * @param {Array} list2 The second array to consider.
 * @return {Array} The list made by pairing up same-indexed elements of `list1` and `list2`.
 * @example
 *
 *      R.zip([1, 2, 3], ['a', 'b', 'c']); //=> [[1, 'a'], [2, 'b'], [3, 'c']]
 * @symb R.zip([a, b, c], [d, e, f]) = [[a, d], [b, e], [c, f]]
 */


var zip = /*#__PURE__*/_curry2_1(function zip(a, b) {
  var rv = [];
  var idx = 0;
  var len = Math.min(a.length, b.length);
  while (idx < len) {
    rv[idx] = [a[idx], b[idx]];
    idx += 1;
  }
  return rv;
});
var zip_1 = zip;

/**
 * A port of [format-util](https://www.npmjs.com/package/format-util) but with
 * support for %o in addition to %j, %s and %d.
 *
 * @param {String} pattern A string value representing the message with any
 * placeholders marked by %s, %d, %j or %o
 * @returns {String} A string formatted with the supplied values inserted
 * (stringified) into its placeholders
 */
var utilFormat = function utilFormat(pattern) {
    for (var _len = arguments.length, params = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        params[_key - 1] = arguments[_key];
    }

    return isStringieThingie(pattern) && params.length ? zip_1(match_1(/(%?)(%([jods]))/g, pattern), map_1(cond_1([[either_1(is_1(Number), is_1(String)), identity_1], [either_1(is_1(Array), isPlainObj$1), JSON.stringify]]))(params)).reduce(function (patt, _ref) {
        var placeholder = _ref[0];
        var value = _ref[1];
        return patt.replace(placeholder, value);
    }, pattern).replace(/%{2,2}/g, '%') : pattern;
};

/**
 * A curried wrapper around [format-util](https://www.npmjs.com/package/format-util)
 * that places the second param into the placeholder from the first param.
 *
 * @func
 * @sig String -> String -> String
 * @param {String} description A description of the error message
 * (friendly to standard message formatting charaacters: %s %o %f etc)
 * @param {String} message An value to plug into the description using formatting placeholders
 * @returns {String} A formatted message
 */
var format = curry_1(function (description, message) {
    return utilFormat(description, message);
});

/**
 * Validates all the configuration options for a new Duck, returning any
 * validation errors and a filtered version of the config options (just those
 * that passed validation).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} options All the configuration options to be passed into the duck constructor
 * @returns {Object} An object containing the 'validationsResult'
 * and 'validatedOptions' (which are just all the validated fields)
 */
var createDuckSchemaValidator = function createDuckSchemaValidator() {
    var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

    var optionsToValidate = pick_1(keys_1(duxRules), options);
    var validationsResult = spected(duxRules, optionsToValidate);
    var validatedOptions = pruneInvalidFields(optionsToValidate, validationsResult);
    return { validationsResult: validationsResult, validatedOptions: validatedOptions };
};

// export const createDuckSchemaValidator = compose(
//     converge(merge, [
//         compose(
//             objOf('validatedOptions'),
//             converge(pruneInvalidFields, [identity, spected(duxRules)])
//         ),
//         compose(objOf('validationsResult'), spected(duxRules))
//     ]),
//     pick(keys(duxRules))
// )

/**
 * For a row of ducks, retrieves any invalid inputs for each duck's state machines.
 * State machines are objects whose keys are states and whose values are objects
 * inside of which the input value is the key of each of its key/value pairs.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} row An object of ducks, each containing one or more state machines and action types
 * @returns {Object} An object of validation error messages about invalid states inputs for each duck's state machines
 */
var getInvalidStateMachineInputs = compose_1(map_1(format('These inputs are not valid Action Types: %o')), filter_1(isNotEmpty), map_1(invalidStateMachineInputs), values_1, filter_1(isDux));

/**
 * For a row of ducks, checks the schema rules for valiation errors.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} row An object of ducks
 * @returns {Object} An object of validation error messages corresponding to one or more ducks in the row
 */
var getInvalidDucks = compose_1(map_1(format('These violations of the schema rules for the middleware were found: %o')), filter_1(isNotEmpty), map_1(compose_1(pruneValidatedFields, spected(duckMiddlewareRules))), values_1, filter_1(isDux));

/**
 * For a row of ducks, checks if the name given to the duck inside the row maches that duck's store prop.
 *
 * @func
 * @sig {k: v} -> Boolean
 * @param {Object} row An object of ducks
 * @returns {Boolean} whether the name for each duck (in the row) matches each duck's store prop
 */
var duckNamesMatchStoreNames = compose_1(all_1(function (_ref2) {
    var key = _ref2[0];
    var val = _ref2[1];
    return equals_1(key, prop_1('store', val));
}), toPairs_1, filter_1(isDux));

/**
 * Checks for any validation errors on a row of ducks.
 * If there are errors, a string description of the problem(s) is returned.
 *
 * @func
 * @sig {k: v} -> String
 * @param {Object} row An object containing one or more ducks
 * @returns {String|null} A validation error message or null (if no validation errors)
 */
var getRowValidationErrors = function getRowValidationErrors(row) {
    if (noDucks(row)) {
        // eslint-disable-next-line max-len
        return 'No ducks have been provided! To create the Attadux middleware please provide an Object containing one or more ducks';
    }

    if (!duckNamesMatchStoreNames(row)) {
        // eslint-disable-next-line max-len
        return 'The name of each duck should match its \'store\' prop, otherwise it will not be possible to find the correct duck for each dispatched action';
    }

    var invalidStateMachines = getInvalidStateMachineInputs(row);
    if (isNotEmpty(invalidStateMachines)) {
        return format('Invalid State Machines: %o', invalidStateMachines);
    }

    var invalidDucks = getInvalidDucks(row);
    if (isNotEmpty(invalidDucks)) {
        return format('Invalid Ducks: %o', invalidDucks);
    }

    return null;
};

/**
 * A function that always returns `false`. Any passed in parameters are ignored.
 *
 * @func
 * @memberOf R
 * @since v0.9.0
 * @category Function
 * @sig * -> Boolean
 * @param {*}
 * @return {Boolean}
 * @see R.always, R.T
 * @example
 *
 *      R.F(); //=> false
 */


var F = /*#__PURE__*/always_1(false);
var F_1 = F;

var _extends$5 = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

/**
 * A slightly opinionated way to handle successful effects and creating a new
 * Action to be dispatched. This default success handler assumes you're alright
 * with appending _SUCCESS to the original action type, and merging the return object
 * (from your effect function) into the new Action. However, if your effect does NOT
 * return an object, that return value will be merged onto the new Action onto a prop
 * called "payload". The other assumption it makes is that if you have suffixed your
 * original Action with _REQUEST or _EFFECT, that will be removed. A (somewhat) standard
 * pattern in the Redux community is to suffix _SUCCESS for effects that succeed
 * and either _ERROR or _FAIL for those that failed. You can always override
 * with your own success handler by providing it when you create an effect in
 * the ducks.
 *
 * @func
 * @sig String|{k: v} -> {k: v} -> {k: v}
 * @param {String|Object} result The result returned from your effect function
 * @param {Object} action The original Redux action that triggered the effect
 * @returns {Object} A new Action object that contains the "payload" that was
 * caught when the effect was created (unless your effect function returned an
 * object, in which case it will be merged onto this)
 */
var defaultSuccessHandler = curry_1(function (result, action) {
    return compose_1(merge_1(__, _extends$5({}, unless_1(isPlainObj$1, objOf_1('payload'))(result))), when_1(pathSatisfies_1(is_1(String), ['type']), compose_1(evolve_1({ type: function type(t) {
            return t.replace(/_REQUEST$/i, '').replace(/_EFFECT$/i, '') + '_SUCCESS';
        } }), pick_1(['type']))))(action);
});

/**
 * A slightly opinionated way to handle catching an exception and creating a new
 * Action to be dispatched. This default error handler assumes you're alright
 * with appending _ERROR to the original action type, and pruning out all of the
 * other fields from the original action and setting only an "error" prop on the
 * new Action. The other assumption it makes is that if you have suffixed your
 * original Action with _REQUEST or _EFFECT, that will be removed. A (somewhat) standard
 * pattern in the Redux community is to suffix _SUCCESS for effects that succeed
 * and either _ERROR or _FAIL for those that failed. You can always override
 * with your own error handler by providing it when you create an effect in
 * the ducks.
 *
 * @func
 * @sig String|{k: v} -> {k: v} -> {k: v}
 * @param {String|Object} error An error that was caught when the effect was
 * triggered
 * @param {Object} action The original Redux action that triggered the effect
 * @returns {Object} A new Action object that contains the "error" that was
 * caught when the effect was created
 */
var defaultErrorHandler = curry_1(function (error, action) {
    return compose_1(merge_1({ error: unless_1(isPlainObj$1, String)(error) }), when_1(pathSatisfies_1(is_1(String), ['type']), compose_1(evolve_1({ type: function type(t) {
            return t.replace(/_REQUEST$/i, '').replace(/_EFFECT$/i, '') + '_ERROR';
        } }), pick_1(['type']))))(action);
});

/**
 * Makes a predicate function out of either a String, RegExp, or a Function,
 * which can then be applied to an Object that contains a "type" property.
 * This is meant to be used to evaluate if a Redux action matches the predicate.
 * If neither a String, RegExp, nor Function is supplied here, then the result
 * will be to give back a function that always returns false.
 *
 * @func
 * @sig String|RegExp|({k: v} -> Boolean) -> ({k: v} -> Boolean)
 * @param {String|RegExp|Function} pattern A String to be exactly matched to an
 * Object's "type" property, OR a Regular expression to be matched against it,
 * OR a function to be used to match against any custom criteria on that Object
 * @returns {Function} A predicate function to be applied later to an Object
 * (ideally, one containing a "type" property)
 */
var makePredicate = cond_1([[is_1(String), pathEq_1(['type'])], [is_1(RegExp), compose_1(pathSatisfies_1(__, ['type']), test_1)], [is_1(Function), function (fn) {
    return compose_1(Boolean, fn);
}], [T_1, always_1(F_1)]]);

/**
 * Creates a function that handles the effect result (either success or failure)
 *
 * @func
 * @sig (a -> b) -> (a -> b) -> (a -> b)
 * @param {Function} defaultHandler The fall through handler function to be used
* in case the handler passed into this function is null/undefined
* @param {String|Object|Function} handler The handler function or the
* String/Object to be turned into a handler function (if String/Object, it will
* be turned into a Shapey spec-mapping function)
 * @returns {Function} A success OR error handler function to be applied after
 * the effect is finished
 */
var makeResponseHandler = curry_1(function (defaultHandler, handler) {
    return cond_1([[is_1(String), compose_1(makeShaper, objOf_1('type'))], [isPlainObj$1, makeShaper], [is_1(Function), identity_1], [T_1, always_1(defaultHandler)]])(handler);
});

/**
 * Creates a robust effect handler from a predicate, an effect creating
 * function, as well as a success and error handler. This curried function can
 * then be applied safely to any Action that matches the predicate (if the
 * Action does NOT match, this is simply an identity function - which returns
 * the Action as-is).
 * 
 * @func
 * @param {String|RegExp|Function} pattern A String to be exactly matched to an
 * Object's "type" property, OR a Regular expression to be matched against it,
 * OR a function to be used to match against any custom criteria on that Object
 * @param {Function} effectHandler A custom function that creates some kind of 
 * normal (but "impure") effect which may succeed or it may fail
 * @param {Function} successHandler A function that receives the input of the
 * effect creating function and creates a new Action containing its result
 * @param {Function} errorHandler A function that receives the caught exception
 * from the effect creating function and creates a new Action from it
 * @returns {Object} Either the original action (if it didn't match the
 * predicate) or a new Action that represents the succes of the effect or
 * alternatively it's failure
 */
var createEffectHandler = curry_1(function (pattern, effectHandler, successHandler, errorHandler, action) {
    return when_1(makePredicate(pattern), tryCatch_1(pipe_1(effectHandler, ifElse_1(isPromise, function (promise) {
        return promise.then(function (res) {
            return successHandler(res, action);
        })['catch'](function (err) {
            return errorHandler(err, action);
        });
    }, function (res) {
        return successHandler(res, action);
    })), function (ex) {
        return errorHandler(ex, action);
    }))(action);
});

/**
 * Merges multiple ducks into one flattened Object (row).
 *
 * @func
 * @sig ({k: v}...) -> {k: v}
 * @param {...Object} ducks One or more ducks to merge into one row
 * @returns {Object} A single object composed of many ducks
 */
var createRow = compose_1(reduce_1(function (row, duck) {
    return assoc_1(duck.store, duck, row);
}, {}), filter_1(isDux), unapply_1(identity_1));

/**
 * A function which takes an Object of one or more ducks and finds the duck
 * which matches a (potentially) namespaced Redux action
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} row An Object containing one or more ducks
 * @param {Object} action An dispatch Redux action (if belonging to this lib,
 * should have a namespaced action type)
 * @returns {Object} A single duck that corresponds to the dispatched action
 * (defaults to an empty object if none found)
 */
var createDuckLookup = curry_1(function (row, action) {
    return compose_1(unless_1(isPlainObj$1, always_1({})), when_1(isStringieThingie, prop_1(__, filter_1(isDux, row))), when_1(is_1(String), compose_1(either_1(nth_1(1), head_1), split_1('/'))), path_1(['type']))(action);
});

/**
 * Validates and applies the configuration options for a new Duck, also
 * performing final formatting for many of the duck's simpler props
 * (consts, types, validationLevel, namespace, store, etc.).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} options All the configuration options to be passed into the duck constructor
 * @returns {Object} An object containing all the validated configuration options for the new Duck
 */
var createDuckMetadata = compose_1(mergeAll_1, ap_1([compose_1(objOf_1('stateMachinesPropName'), always_1(['states'])), compose_1(converge_1(mergeDeepRight_1, [compose_1(evolve_1(metadataEvolvers), pick_1(keys_1(metadataEvolvers)), copyRawQueriesToConsts), compose_1(objOf_1('types'), converge_1(createTypes, [identity_1, prop_1('types')]))]), prop_1('validatedOptions')), compose_1(when_1(isNotEmpty, objOf_1('invalidOptions')), pruneValidatedFields, prop_1('validationsResult')), compose_1(objOf_1('options'), prop_1('validatedOptions'))]), of_1, createDuckSchemaValidator);

/**
 * Creates the Duck's validators (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain validators (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with validators (if they were found inside of 'options').
 */
var createDuckValidators = compose_1(unless_1(propEq_1('validationLevel', 'PRUNE'), evolve_1({
    validators: map_1(function (validator) {
        return compose_1(pruneValidatedFields, validator);
    })
})), converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'validators']), always_1({}), compose_1(objOf_1('validators'), map_1(spected), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'validators'])), identity_1])))]));

/**
 * Creates the Duck's effect handlers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain effect handlers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with effect handlers (if they were found inside of 'options').
 */
var createDuckEffects = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'effects']), always_1([]), compose_1(objOf_1('effects'), map_1(function (_ref) {
    var pattern = _ref[0];
    var effectHandler = _ref[1];
    var successHandler = _ref[2];
    var errorHandler = _ref[3];
    return createEffectHandler(pattern, effectHandler || identity_1, makeResponseHandler(defaultSuccessHandler, successHandler), makeResponseHandler(defaultErrorHandler, errorHandler));
}), filter_1(allPass_1([is_1(Array), pathSatisfies_1(anyPass_1([is_1(String), is_1(RegExp), is_1(Function)]), [0]), pathSatisfies_1(is_1(Function), [1]), pathSatisfies_1(anyPass_1([isNil_1, isPlainObj$1, is_1(Function), is_1(String)]), [2]), pathSatisfies_1(anyPass_1([isNil_1, isPlainObj$1, is_1(Function), is_1(String)]), [3])])), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'effects'])), identity_1])))]);

/**
 * Creates the Duck's throttlers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain throttlers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with throttlers (if they were found inside of 'options').
 */
var createDuckThrottlers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'throttling']), always_1([]), compose_1(objOf_1('throttling'), map_1(function (_ref2) {
    var pattern = _ref2[0];
    var milliseconds = _ref2[1];
    return [pattern, Number(milliseconds)];
}), filter_1(allPass_1([is_1(Array), pathSatisfies_1(anyPass_1([is_1(String), is_1(RegExp), is_1(Function)]), [0]), pathSatisfies_1(compose_1(not_1, isNaN, Number), [1])])), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'throttling'])), identity_1])))]);

/**
 * Creates the Duck's debouncers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain debouncers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with debouncers (if they were found inside of 'options').
 */
var createDuckDebouncers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'debouncing']), always_1([]), compose_1(objOf_1('debouncing'), map_1(function (_ref3) {
    var pattern = _ref3[0];
    var milliseconds = _ref3[1];
    return [pattern, Number(milliseconds)];
}), filter_1(allPass_1([is_1(Array), pathSatisfies_1(anyPass_1([is_1(String), is_1(RegExp), is_1(Function)]), [0]), pathSatisfies_1(compose_1(not_1, isNaN, Number), [1])])), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'debouncing'])), identity_1])))]);

/**
 * Creates the Duck's queries (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain queries (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with queries (if they were found inside of 'options').
 */
var createDuckQueries = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'queries']), always_1({}), compose_1(objOf_1('queries'), makeQueries))]);

/**
 * Creates the Duck's workers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain workers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with workers (if they were found inside of 'options').
 */
var createDuckWorkers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'workers']), always_1({}), compose_1(objOf_1('workers'), makeWorkers))]);

/**
 * Creates the Duck's state machines (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain state machines (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with state machines (if they were found inside of 'options').
 */
var createDuckMachines = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'machines']), always_1({}), compose_1(objOf_1('machines'), converge_1(createMachines, [converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'machines'])), identity_1]), identity_1])))]);

/**
 * Creates the Duck's initial state for the section of the store that affects this duck.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain initialState (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with initialState (if it was found inside of 'options').
 */
var createDuckInitialState = converge_1(mergeDeepRight_1, [identity_1, compose_1(objOf_1('initialState'), converge_1(mergeDeepRight_1, [converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'initialState'])), identity_1]), ifElse_1(propSatisfies_1(isEmpty_1, 'machines'), always_1({}), compose_1(applyTo_1({}), converge_1(assocPath_1, [prop_1('stateMachinesPropName'), compose_1(getDefaultStateForMachines, prop_1('machines'))])))]))]);

/**
 * Creates the Duck's selectors (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain selectors (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with selectors (if they were found inside of 'options').
 */
var createDuckSelectors = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'selectors']), always_1({}), compose_1(objOf_1('selectors'), deriveSelectors, converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'selectors'])), identity_1])))]);

/**
 * Creates an Object of enhancement functions out of an Object of spec objects
 * (or arrays of spec objects).
 *
 * @func
 * @sig {k: [({k: v} -> {k: v}), ({k: v} -> {k: v}), ...]|({k: v} -> {k: v}) } -> {k: ({k: v} -> {k: v}) }
 * @param {Object[]|Object} enhancements A single enhancement spec or an Array of enhancement specs
 * @returns {Object} An object of enhancer functions, each ready to receive an
 * input object and apply their single or chain of enhancer functions to it.
 */
var makeEnhancers = ifElse_1(is_1(Array), shapeline, makeShaper);

/**
 * Creates the Duck's action enhancers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain action enhancers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with action enhancers (if they were found inside of 'options').
 */
var createDuckActionEnhancers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'enhancers']), always_1({}), compose_1(objOf_1('enhancers'), map_1(makeEnhancers), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'enhancers'])), identity_1])))]);

/**
 * Creates an Object of multiplier functions out of an Object of spec objects
 * (or arrays of spec objects).
 *
 * @func
 * @sig {k: [({k: v} -> {k: v}), ({k: v} -> {k: v}), ...]|({k: v} -> {k: v}) } -> {k: ({k: v} -> {k: v}) }
 * @param {Object[]|Object} multipliers A single multiplier spec or an Array of multipler specs
 * @returns {Object} An object of multipler functions, each ready to receive an
 * input object and apply their single or many multipler functions to it.
 */
var makeMultipliers = compose_1(juxt_1, map_1(makeShaper), unless_1(is_1(Array), of_1));

/**
 * Creates the Duck's action multipliers (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain action multipliers (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with action multipliers (if they were found inside of 'options').
 */
var createDuckActionMultipliers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'multipliers']), always_1({}), compose_1(objOf_1('multipliers'), map_1(makeMultipliers), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'multipliers'])), identity_1])))]);
/**
 * Creates the Duck's action creators (if they are present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain action creators (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with action creators (if they were found inside of 'options').
 */
var createDuckActionCreators = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(pathSatisfies_1(isNil_1, ['options', 'creators']), always_1({}), compose_1(objOf_1('creators'), converge_1(call_1, [compose_1(coerceToFn, path_1(['options', 'creators'])), identity_1])))]);

/**
 * Creates the Duck's reducer (if it was present inside of its 'options' prop).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain a reducer (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with a reducer (if it was found inside of 'options').
 */
var createDuckReducer = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(either_1(pathSatisfies_1(isNil_1, ['options', 'reducer']), propSatisfies_1(isEmpty_1, 'machines')), compose_1(objOf_1('reducer'), createReducer), compose_1(objOf_1('reducer'), createTransitionsPostReducer))]);

/**
 * Creates validation helpers for a given Duck, to be used in the middleware chain.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} duck A duck which (may) contain validators (inside of its 'options')
 * @returns {Object} A clone of the duck, but now with validator helpers (if validators were found inside of 'options').
 */
var createValidationMiddlewareHelpers = converge_1(mergeDeepRight_1, [identity_1, ifElse_1(propSatisfies_1(isNil_1, 'validators'), always_1({}), compose_1(applySpec_1({
    isPayloadValid: createPayloadValidator,
    getValidationErrors: createPayloadValidationsLogger,
    pruneInvalidFields: createPayloadPruner
}), prop_1('validators')))]);

/**
 * Returns a function of arity `n` from a (manually) curried function.
 *
 * @func
 * @memberOf R
 * @since v0.14.0
 * @category Function
 * @sig Number -> (a -> b) -> (a -> c)
 * @param {Number} length The arity for the returned function.
 * @param {Function} fn The function to uncurry.
 * @return {Function} A new function.
 * @see R.curry
 * @example
 *
 *      var addFour = a => b => c => d => a + b + c + d;
 *
 *      var uncurriedAddFour = R.uncurryN(4, addFour);
 *      uncurriedAddFour(1, 2, 3, 4); //=> 10
 */


var uncurryN = /*#__PURE__*/_curry2_1(function uncurryN(depth, fn) {
  return curryN_1(depth, function () {
    var currentDepth = 1;
    var value = fn;
    var idx = 0;
    var endIdx;
    while (currentDepth <= depth && typeof value === 'function') {
      endIdx = currentDepth === depth ? arguments.length : idx + value.length;
      value = value.apply(this, Array.prototype.slice.call(arguments, idx, endIdx));
      currentDepth += 1;
      idx = endIdx;
    }
    return value;
  });
});
var uncurryN_1 = uncurryN;

/**
 * The `mapAccum` function behaves like a combination of map and reduce; it
 * applies a function to each element of a list, passing an accumulating
 * parameter from left to right, and returning a final value of this
 * accumulator together with the new list.
 *
 * The iterator function receives two arguments, *acc* and *value*, and should
 * return a tuple *[acc, value]*.
 *
 * @func
 * @memberOf R
 * @since v0.10.0
 * @category List
 * @sig ((acc, x) -> (acc, y)) -> acc -> [x] -> (acc, [y])
 * @param {Function} fn The function to be called on every element of the input `list`.
 * @param {*} acc The accumulator value.
 * @param {Array} list The list to iterate over.
 * @return {*} The final, accumulated value.
 * @see R.addIndex, R.mapAccumRight
 * @example
 *
 *      var digits = ['1', '2', '3', '4'];
 *      var appender = (a, b) => [a + b, a + b];
 *
 *      R.mapAccum(appender, 0, digits); //=> ['01234', ['01', '012', '0123', '01234']]
 * @symb R.mapAccum(f, a, [b, c, d]) = [
 *   f(f(f(a, b)[0], c)[0], d)[0],
 *   [
 *     f(a, b)[1],
 *     f(f(a, b)[0], c)[1],
 *     f(f(f(a, b)[0], c)[0], d)[1]
 *   ]
 * ]
 */


var mapAccum = /*#__PURE__*/_curry3_1(function mapAccum(fn, acc, list) {
  var idx = 0;
  var len = list.length;
  var result = [];
  var tuple = [acc];
  while (idx < len) {
    tuple = fn(tuple[0], list[idx]);
    result[idx] = tuple[1];
    idx += 1;
  }
  return [tuple[0], result];
});
var mapAccum_1 = mapAccum;

/**
 * Returns the result of concatenating the given lists or strings.
 *
 * Note: `R.concat` expects both arguments to be of the same type,
 * unlike the native `Array.prototype.concat` method. It will throw
 * an error if you `concat` an Array with a non-Array value.
 *
 * Dispatches to the `concat` method of the first argument, if present.
 * Can also concatenate two members of a [fantasy-land
 * compatible semigroup](https://github.com/fantasyland/fantasy-land#semigroup).
 *
 * @func
 * @memberOf R
 * @since v0.1.0
 * @category List
 * @sig [a] -> [a] -> [a]
 * @sig String -> String -> String
 * @param {Array|String} firstList The first list
 * @param {Array|String} secondList The second list
 * @return {Array|String} A list consisting of the elements of `firstList` followed by the elements of
 * `secondList`.
 *
 * @example
 *
 *      R.concat('ABC', 'DEF'); // 'ABCDEF'
 *      R.concat([4, 5, 6], [1, 2, 3]); //=> [4, 5, 6, 1, 2, 3]
 *      R.concat([], []); //=> []
 */


var concat = /*#__PURE__*/_curry2_1(function concat(a, b) {
  if (_isArray(a)) {
    if (_isArray(b)) {
      return a.concat(b);
    }
    throw new TypeError(toString_1(b) + ' is not an array');
  }
  if (_isString_1(a)) {
    if (_isString_1(b)) {
      return a + b;
    }
    throw new TypeError(toString_1(b) + ' is not a string');
  }
  if (a != null && _isFunction_1(a['fantasy-land/concat'])) {
    return a['fantasy-land/concat'](b);
  }
  if (a != null && _isFunction_1(a.concat)) {
    return a.concat(b);
  }
  throw new TypeError(toString_1(a) + ' does not have a method named "concat" or "fantasy-land/concat"');
});
var concat_1 = concat;

/**
 * Creates a new object with the own properties of the two provided objects.
 * If a key exists in both objects:
 * - and both associated values are also objects then the values will be
 *   recursively merged.
 * - otherwise the provided function is applied to associated values using the
 *   resulting value as the new value associated with the key.
 * If a key only exists in one object, the value will be associated with the key
 * of the resulting object.
 *
 * @func
 * @memberOf R
 * @since v0.24.0
 * @category Object
 * @sig ((a, a) -> a) -> {a} -> {a} -> {a}
 * @param {Function} fn
 * @param {Object} lObj
 * @param {Object} rObj
 * @return {Object}
 * @see R.mergeWith, R.mergeDeep, R.mergeDeepWithKey
 * @example
 *
 *      R.mergeDeepWith(R.concat,
 *                      { a: true, c: { values: [10, 20] }},
 *                      { b: true, c: { values: [15, 35] }});
 *      //=> { a: true, b: true, c: { values: [10, 20, 15, 35] }}
 */


var mergeDeepWith = /*#__PURE__*/_curry3_1(function mergeDeepWith(fn, lObj, rObj) {
  return mergeDeepWithKey_1(function (k, lVal, rVal) {
    return fn(lVal, rVal);
  }, lObj, rObj);
});
var mergeDeepWith_1 = mergeDeepWith;

/**
 * A function that creates an "extender" function which will merges
 * a portion of two duck together at a specified prop
 *
 * @func
 * @sig {k: v} -> {k: v} -> String -> *
 * @param {Object} childDuckOptions A set of options from which a duck can be built
 * @param {Object} parentDuck A duck that has already been built and whose original constructor options will be re-invoked
 * @param {String} key A key name from which to match child options to parent (same prop for both)
 * @returns {Function} A function that takes the name of a prop on a duck, and
 * will merge the parent and child together for that prop
 */
var createOptionsExtender = curry_1(function (childDuckOptions, _ref, key) {
    var parentOptions = _ref.options;
    return cond_1([
    /* If the key at the child or parent is a function */
    [compose_1(any_1(is_1(Function)), map_1(prop_1(key)), pair_1(parentOptions)),
    /**
     * then they both need to be invoked
     * (coerced to fn, if not already)
     * and then their results are merged
     */
    always_1(converge_1(mergeDeepWith_1(simpleMergeStrategy), [coerceToFn(parentOptions[key]), converge_1(coerceToFn(childDuckOptions[key]), [identity_1, coerceToFn(parentOptions[key])])]))],
    /* If the child doesn't have anything at that key, just return from the parent */
    [compose_1(isNil_1, prop_1(key)), always_1(parentOptions[key])],
    /* Otherwise, simply merge the parent and child together at that key */
    [T_1, compose_1(mergeDeepWith_1(simpleMergeStrategy, parentOptions[key]), prop_1(key))]])(childDuckOptions);
});

/**
 * Merges the reducer from an existing duck with a reducer inside of a set of configuration options (for a new duck).
 * If a reducer doesn't exist in both sources, no merging is needed, otherwise the merged reducer
 * will always invoke the parent's reducer and feed that result into the child's reducer as its 'state' argument.
 * This way a chain of reducers will always be executed.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck An existing duck, which contains its original
 * configuration options at its 'options' prop.
 * @param {Object} options A set of configuration options for a new duck
 * @returns {Object} A reducer function which is either a chain of reducers (parent first, then child)
 * or one of the reducers (if a reducer wasn't present in both sets of config options)
 */
var extendReducer = function extendReducer(_ref2) {
    var parentOptions = _ref2.options;
    return compose_1(ifElse_1(all_1(has_1('reducer')), compose_1(makeExtendedReducer(parentOptions), last_1), compose_1(find_1(has_1('reducer')), pair_1(parentOptions))), pair_1(parentOptions));
};

/**
 * Merges a set of configuration options (for a new duck) with those of an
 * existing duck (along with defaults for any unsupplied, but required values).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck An existing duck, which contains its original
 * configuration options at its 'options' prop.
 * @param {Object} options A set of configuration options for a new duck
 * @returns {Object} A set of configuration options where schema defaults are
 * merged with the existing duck's configuration options and the new
 * (to-be-built) duck's configuration options
 */
var extendOptionsForDuck = function extendOptionsForDuck(duck) {
    return compose_1(converge_1(mergeDeepRight_1, [identity_1, compose_1(objOf_1('reducer'), extendReducer(duck))]), reduce_1(mergeDeepRight_1, {}), concat_1([duxDefaults, duck.options]), of_1, applyTo_1(duck), coerceToFn);
};

/**
 * Applies a set of evolvers, which are just transformations that are run against a given prop inside of an object.
 * In this case the config options for a new duck are transformed and then
 * those are merged with the same prop(s) inside of the config 'options' of the existing duck.
 *
 * @func
 * @sig [[k: (a -> b)] -> {k: v} -> {k: v} -> {k: v}
 * @param {Array} evolvers An array of arrays (which are key/value pairs whose
 * key is a prop name on an options object and whose value is a function to
 * execute against the corresponding value on the options object)
 * @param {Object} duck An existing duck, which contains its original
 * configuration options at its 'options' prop.
 * @param {Object} options A set of configuration options for a new duck
 * @returns {Object} The options from the original duck extended with the
 * configuration options for a new (to-be-built) duck, according to the
 * transformations defined in the evolvers.
 */
var createExtendedOptions = curry_1(function (evolvers, duck, options) {
    return compose_1(mergeDeepRight_1(options), mergeAll_1, last_1, mapAccum_1(function (mergedDuck, _ref3) {
        var key = _ref3[0];
        var builder = _ref3[1];

        var option = compose_1(objOf_1(key), when_1(is_1(Function), applyTo_1(key)), builder)(mergedDuck);
        return [mergeDeepRight_1(mergedDuck, option), option];
    }, duck))(evolvers);
});

/**
 * Takes an already built duck and the options for building a new one and
 * extends the new options onto the options and simple props of the already built duck.
 * The combined object can be passed as options when creating a new duck.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck An existing duck from which a new duck will be based
 * @param {Object} options A set of options to be merged with those of an
 * existing duck (to be used together whenn creating a new duck)
 * @returns {Object} a merged set of options to be fed into a new duck
 */
var createDuckExtender = function createDuckExtender(duck) {
    var extendOptions = extendOptionsForDuck(duck);
    return function (options) {
        var childOptions = extendOptions(options);
        var optionBuilders = [['consts', compose_1(mergeDeepWith_1(concatOrReplace, __, childOptions.consts), path_1(['options', 'consts']))], ['types', compose_1(uniq_1, concat_1(childOptions.types), path_1(['options', 'types']))], ['initialState', createOptionsExtender(childOptions)], ['machines', createOptionsExtender(childOptions)], ['creators', createOptionsExtender(childOptions)], ['selectors', createOptionsExtender(childOptions)], ['queries', createOptionsExtender(childOptions)], ['enhancers', createOptionsExtender(childOptions)], ['multipliers', createOptionsExtender(childOptions)], ['throttling', createOptionsExtender(childOptions)], ['debouncing', createOptionsExtender(childOptions)], ['effects', createOptionsExtender(childOptions)], ['validators', createOptionsExtender(childOptions)], ['workers', createOptionsExtender(childOptions)]];
        return createExtendedOptions(optionBuilders, duck, childOptions);
    };
};

/**
 * Builds a duck from a set of options (whose keys are validated against
 * a whitelist and corresponding schema), including selectors, action creators,
 * state machines, validators, a reducer and basic metadata
 * (including the required 'namespace' and 'store' name).
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} options A set of functions, objects (of various schema types)
 * and basic metadata which will go into the finished duck.
 * @returns {Object} A fully built duck, complete with types, action creators,
 * a reducer, selectors, state machines, validators and some basic metadata.
 */
var createDuck = compose_1(Object.freeze, createDuckReducer, createDuckEffects, createDuckWorkers, createValidationMiddlewareHelpers, createDuckActionMultipliers, createDuckActionEnhancers, createDuckActionCreators, createDuckDebouncers, createDuckThrottlers, createDuckSelectors, createDuckInitialState, createDuckMachines, createDuckValidators, createDuckQueries, createDuckMetadata, mergeDeepRight_1(duxDefaults));

/**
 * Builds a function that will extend a duck with options to be supplied at a later time.
 * Those options (supplied later) may include selectors, action creators, state machines,
 * validators, a reducer and basic metadata (including the required 'namespace' and 'store' name).
 *
 * @func
 * @sig {k: v} -> ({k: v} -> {k: v})
 * @param {Object} duck A 'parent' duck whose own options will be merged with those supplied for the new (child) duck.
 * @returns {Object} An extender function, ready to be supplied a set of options to extend the original duck.
 */
var createExtenderForDuck = converge_1(compose_1, [always_1(createDuck), createDuckExtender]);

/**
 * Builds a duck from an existing duck _and_ a set of options for a new duck.
 * This function is meant to be invoked when both the duck and the child options
 * are available together (not a curried function).
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} duck A 'parent' duck whose own options will be merged with
 * those supplied for the new (child) duck.
 * @param {Object} options A set of functions, objects (of various schema types)
 * and basic metadata which will go into the finished duck, merged with the parent duck.
 * @returns {Object} A fully built duck, complete with types, action creators,
 * a reducer, selectors, state machines, validators and some basic metadata,
 * merged from the parent duck along with the options supplied for the new duck.
 */
var extendDuck = uncurryN_1(2, createExtenderForDuck);

function _forceReduced(x) {
  return {
    '@@transducer/value': x,
    '@@transducer/reduced': true
  };
}
var _forceReduced_1 = _forceReduced;

var preservingReduced = function (xf) {
  return {
    '@@transducer/init': _xfBase.init,
    '@@transducer/result': function (result) {
      return xf['@@transducer/result'](result);
    },
    '@@transducer/step': function (result, input) {
      var ret = xf['@@transducer/step'](result, input);
      return ret['@@transducer/reduced'] ? _forceReduced_1(ret) : ret;
    }
  };
};

var _flatCat = function _xcat(xf) {
  var rxf = preservingReduced(xf);
  return {
    '@@transducer/init': _xfBase.init,
    '@@transducer/result': function (result) {
      return rxf['@@transducer/result'](result);
    },
    '@@transducer/step': function (result, input) {
      return !_isArrayLike_1(input) ? _reduce_1(rxf, result, [input]) : _reduce_1(rxf, result, input);
    }
  };
};

var _flatCat_1 = _flatCat;

var _xchain = /*#__PURE__*/_curry2_1(function _xchain(f, xf) {
  return map_1(f, _flatCat_1(xf));
});
var _xchain_1 = _xchain;

/**
 * `chain` maps a function over a list and concatenates the results. `chain`
 * is also known as `flatMap` in some libraries
 *
 * Dispatches to the `chain` method of the second argument, if present,
 * according to the [FantasyLand Chain spec](https://github.com/fantasyland/fantasy-land#chain).
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category List
 * @sig Chain m => (a -> m b) -> m a -> m b
 * @param {Function} fn The function to map with
 * @param {Array} list The list to map over
 * @return {Array} The result of flat-mapping `list` with `fn`
 * @example
 *
 *      var duplicate = n => [n, n];
 *      R.chain(duplicate, [1, 2, 3]); //=> [1, 1, 2, 2, 3, 3]
 *
 *      R.chain(R.append, R.head)([1, 2, 3]); //=> [1, 2, 3, 1]
 */


var chain = /*#__PURE__*/_curry2_1( /*#__PURE__*/_dispatchable_1(['fantasy-land/chain', 'chain'], _xchain_1, function chain(fn, monad) {
  if (typeof monad === 'function') {
    return function (x) {
      return fn(monad(x))(x);
    };
  }
  return _makeFlat_1(false)(map_1(fn, monad));
}));
var chain_1 = chain;

/**
 * Shorthand for `R.chain(R.identity)`, which removes one level of nesting from
 * any [Chain](https://github.com/fantasyland/fantasy-land#chain).
 *
 * @func
 * @memberOf R
 * @since v0.3.0
 * @category List
 * @sig Chain c => c (c a) -> c a
 * @param {*} list
 * @return {*}
 * @see R.flatten, R.chain
 * @example
 *
 *      R.unnest([1, [2], [[3]]]); //=> [1, 2, [3]]
 *      R.unnest([[1, 2], [3, 4], [5, 6]]); //=> [1, 2, 3, 4, 5, 6]
 */


var unnest = /*#__PURE__*/chain_1(_identity_1);
var unnest_1 = unnest;

var resultHandler = curry_1(function (dispatch, action, result) {
    if (isAction(result) && action.type !== result.type) {
        dispatch(result);
    }
});

var createEffectMiddleware = curry_1(function (logger, effects, dispatch, action) {
    var handleResult = resultHandler(dispatch, action);
    effects.forEach(function (effect) {
        var result = effect(action);
        if (isPromise(result)) {
            result.then(handleResult);
        } else {
            handleResult(result);
        }
    });
    return action;
});

var createEnhancerMiddleware = curry_1(function (logger, dispatch, enhancers, action) {
    var enhance = enhancers[action.type];
    if (is_1(Function, enhance)) {
        try {
            var nextAction = enhance(action);
            if (isAction(nextAction)) {
                if (nextAction.type !== action.type) {
                    dispatch(nextAction);
                } else {
                    return nextAction;
                }
            }
        } catch (error) {
            logger.error('Unable to enhance action: "' + action.type + '"', error);
        }
    }
    return action;
});

var createLimiterMiddleware = curry_1(function (throttling, debouncing, action) {
    if (isEmpty_1(throttling) && isEmpty_1(debouncing)) {
        return Promise.resolve(action);
    }
    var limiters = {};
    return new Promise(function (resolve) {
        if (!limiters[action.type]) {
            var ms = 0;
            if (throttling.some(function (_ref) {
                var throttleBy = _ref[0];
                var milliseconds = _ref[1];

                if (is_1(String, throttleBy) && throttleBy === action.type || is_1(RegExp, throttleBy) && throttleBy.test(action.type) || is_1(Function, throttleBy) && throttleBy(action)) {
                    ms = milliseconds;
                    return true;
                }
                return false;
            })) {
                resolve(action);
                limiters[action.type] = setTimeout(function () {
                    delete limiters[action.type];
                }, ms);
            } else if (debouncing.some(function (_ref2) {
                var debounceBy = _ref2[0];
                var milliseconds = _ref2[1];

                if (is_1(String, debounceBy) && debounceBy === action.type || is_1(RegExp, debounceBy) && debounceBy.test(action.type) || is_1(Function, debounceBy) && debounceBy(action)) {
                    ms = milliseconds;
                    return true;
                }
                return false;
            })) {
                limiters[action.type] = setTimeout(function () {
                    resolve(action);
                    delete limiters[action.type];
                }, ms);
            } else {
                resolve(action);
            }
        }
        resolve(action);
    });
});

var createMultiplierMiddleware = curry_1(function (logger, multipliers, dispatch, action) {
    multipliers.filter(function (multiplierMap) {
        return multiplierMap[action.type];
    }).forEach(function (multiplierMap) {
        try {
            var fanout = multiplierMap[action.type];
            var nextActions = fanout(action);
            if (is_1(Array, nextActions)) {
                unnest_1(nextActions).filter(function (na) {
                    return isAction(na) && na.type !== action.type;
                }).forEach(function (nextAction) {
                    try {
                        dispatch(nextAction);
                    } catch (err) {
                        logger.error('Unable to dispatch action: "' + nextAction.type + '", multiplied from "' + action.type + '"', err);
                    }
                });
            }
        } catch (error) {
            logger.error('Unable to multiply action: "' + action.type + '"', error);
        }
    });
    return action;
});

var createValidatorMiddleware = curry_1(function (logger, getDuckMatchingAction, getState, action) {
    if (isAction(action)) {
        try {
            var _validatorsByLevel;

            var _getDuckMatchingActio = getDuckMatchingAction(action);

            var machines = _getDuckMatchingActio.machines;
            var _getDuckMatchingActio2 = _getDuckMatchingActio.isPayloadValid;
            var isPayloadValid = _getDuckMatchingActio2 === undefined ? T_1 : _getDuckMatchingActio2;
            var _getDuckMatchingActio3 = _getDuckMatchingActio.getValidationErrors;
            var getValidationErrors = _getDuckMatchingActio3 === undefined ? always_1(null) : _getDuckMatchingActio3;
            var _getDuckMatchingActio4 = _getDuckMatchingActio.pruneInvalidFields;
            var pruneInvalidFields = _getDuckMatchingActio4 === undefined ? always_1(action) : _getDuckMatchingActio4;
            var _getDuckMatchingActio5 = _getDuckMatchingActio.stateMachinesPropName;
            var stateMachinesPropName = _getDuckMatchingActio5 === undefined ? 'states' : _getDuckMatchingActio5;
            var _getDuckMatchingActio6 = _getDuckMatchingActio.validationLevel;
            var validationLevel = _getDuckMatchingActio6 === undefined ? VALIDATION_LEVELS.CANCEL : _getDuckMatchingActio6;


            if (noMachines(machines)) {
                return action;
            }

            var validatorsByLevel = (_validatorsByLevel = {}, _validatorsByLevel[VALIDATION_LEVELS.CANCEL] = unless_1(isPayloadValid, F_1), _validatorsByLevel[VALIDATION_LEVELS.LOG] = converge_1(merge_1, [identity_1, compose_1(ifElse_1(isNil_1, always_1({}), objOf_1('validationErrors')), getValidationErrors)]), _validatorsByLevel[VALIDATION_LEVELS.PRUNE] = pruneInvalidFields, _validatorsByLevel[VALIDATION_LEVELS.STRICT] = unless_1(converge_1(isActionTypeInCurrentState, [getState, identity_1, always_1({ machines: machines, stateMachinesPropName: stateMachinesPropName })]), F_1), _validatorsByLevel);

            /* Get the validator matching the validationLevel setting and run it on the action payload */
            var validatedAction = call_1(compose_1(defaultTo_1(identity_1), prop_1(validationLevel))(validatorsByLevel), action);

            /* either pass the validated action forward through the middleware chain or stop it right here */
            if (isAction(validatedAction)) {
                return validatedAction;
            }
        } catch (error) {
            logger.error('Unable to validate action: ' + action.type, error);
        }
    }

    return false;
});

var stubLogger = pipe_1(toPairs_1, map_1(function (_ref) {
  var key = _ref[0];
  return [key, identity_1];
}), fromPairs_1)(console);

var createLogger = (function () {
  var loggingEnabled = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];
  return loggingEnabled ? console : stubLogger;
});

var flattenList = curry_1(function (duckPropName, row) {
    return compose_1(reject_1(either_1(isNil_1, isEmpty_1)), unnest_1, values_1, map_1(pathOr_1([], [duckPropName])), filter_1(has_1(duckPropName)))(row);
});

var index = (function (row) {
    var loggingEnabled = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    var validationErrors = getRowValidationErrors(row);
    var logger = createLogger(loggingEnabled);

    if (validationErrors) {
        logger.error('Error validating your row of ducks', validationErrors);
        throw new Error(validationErrors);
    }

    var getDuckMatchingAction = createDuckLookup(row);

    var effects = flattenList('effects', row);
    var debouncing = flattenList('debouncing', row);
    var throttling = flattenList('throttling', row);
    var multipliers = compose_1(values_1, reject_1(either_1(isNil_1, isEmpty_1)), map_1(pathOr_1({}, ['multipliers'])))(row);

    var curriedEnhancers = createEnhancerMiddleware(logger);
    var curriedEffects = createEffectMiddleware(logger, effects);
    var limitAction = createLimiterMiddleware(throttling, debouncing);
    var curriedMultipliers = createMultiplierMiddleware(logger, multipliers);
    var curriedValidators = createValidatorMiddleware(logger, getDuckMatchingAction);

    return function (_ref) {
        var dispatch = _ref.dispatch;
        var getState = _ref.getState;

        var enhanceAction = curriedEnhancers(dispatch);
        var validateAction = curriedValidators(getState);
        var multiplyAction = curriedMultipliers(dispatch);
        var handleAnyActionEffects = curriedEffects(dispatch);
        return function (next) {
            return function (action) {
                if (isAction(action)) {
                    limitAction(action)['catch'](function (err) {
                        return logger.error('There was a problem throttling or debouncing: "' + action.type + '"', err);
                    }).then(function (limitedAction) {
                        var validatedAction = validateAction(limitedAction);
                        return isAction(validatedAction) ? validatedAction : false;
                    })['catch'](function (err) {
                        logger.error('There was a problem validating: "' + action.type + '"', err);
                        next(action);
                    }).then(function (validatedAction) {
                        if (validatedAction) {
                            var _getDuckMatchingActio = getDuckMatchingAction(validatedAction);

                            var _getDuckMatchingActio2 = _getDuckMatchingActio.enhancers;
                            var enhancers = _getDuckMatchingActio2 === undefined ? {} : _getDuckMatchingActio2;

                            return pipe_1(multiplyAction, enhanceAction(enhancers))(validatedAction);
                        }
                        return false;
                    })['catch'](function (err) {
                        logger.error('A multiplier and/or enhancer failed for: "' + action.type + '"', err);
                        next(action);
                    }).then(function (enhancedAction) {
                        if (isAction(enhancedAction)) {
                            next(enhancedAction);
                            handleAnyActionEffects(enhancedAction);
                        }
                    })['catch'](function (err) {
                        return logger.error('There was a problem executing the effect for: "' + action.type + '"', err);
                    });
                } else {
                    logger.error('Looks like something was dispatched that is not a valid action:', action);
                }
            };
        };
    };
});

exports.createRow = createRow;
exports.createDuck = createDuck;
exports.extendDuck = extendDuck;
exports.createExtenderForDuck = createExtenderForDuck;
exports.createSelector = createSelector;
exports.createDuckSelector = createDuckSelector;
exports.createMiddleware = index;

Object.defineProperty(exports, '__esModule', { value: true });

})));
